#!/usr/bin/perl -w
## @file
# Implementation of SLEPOS LDAP-XML interface.
# @par Authors
# - Ondrej Holecek
# @note
# Contain POS::Admin::Ldap2XML and POS::Admin::Ldap2XML::SAXParser packages
# @copy 2012, SUSE

## @class
# POS::Admin::Ldap2XML class.
# SLEPOS LDAP-XML interface implementation
# @par Provides
# - LDAP DB export into XML
# - XML conversion into Net::LDAP::Entry list
# - RelaxNG validation template generation from SLEPOS LDAP schema (POS::Admin::LdapStruct)
# - XML validation against SLEPOS XML RelaxNG template
# @par Requires
# - POS::Syslog
# - POS::Admin::LdapStruct
# - POS::Config
package POS::Admin::Ldap2XML;

use strict;
use warnings;

use POS::Syslog;

# do not need to call import
require POS::Admin::LdapStruct;
require POS::Config;

# package version
our $VERSION = 0.2;

## @var const XMLVERSION
# current SLEPOS XML version of exported XML.
our constant $XMLVERSION = "0.2";
# version 0.1 as in SLEPOS11 SP2, version 0.2 since SLEPOS11 SP2 FP2
my @KNOWNVERSIONS = ("0.1", "0.2");

# SLEPOS Namespace

## @var const posnsURI
# SLEPOS XML namespace URI
our constant $posnsURI = 'http://www.suse.com/SLEPOS';
## @var const xincnsURI
# XInclude XML namespace URI
our constant $xincnsURI = 'http://www.w3.org/2001/XInclude';
## @var const xincnsPX
# XInclude XML namespace prefix
our constant $xincnsPX = 'xi';
## @var const POSXMLTEMLPATEDIR
our constant $POSXMLTEMPLATEDIR = '/etc/SLEPOS/template/XML';

## @cmethod LdapXML new($config)
# Create new Ldap2XML object
# @param config optional POS::Config object
# @return A new object ref
sub new
{
    my $self = shift;
    my $type = ref($self) || $self;
    my $this = bless {}, $type;

    # get constructor parameters
    while (@_) {
        my $field = shift;
        if (@_) {
            $this->{$field} = shift;
            if (!$this->{$field}) {
                $this->{$field} = "";
            }
        }
    }

    $this->{'CONFIG'} = POS::Config->new("/etc/SLEPOS/adminserver.conf") if (not defined $this->{'CONFIG'});
    $this->{'ENTITIES'} = \%POS::Admin::LdapStruct::LDAP_STRUCT;

    return $this;
} ## end sub new

################################################################################
# XML generator, LDAP db export

## @method bool getXML($outfile, $rangeOrSearch, $dn)
# Exports LDAP into XML file
# @param outfile output file name or undef for STDOUT
# @param rangeOrSearch Net::LDAP::Search ref or range string (all, branch, dn)
# @param dn base LDAP DN which subtree to export. Required only when rangeOrSearch is branch or dn
# @par usage:
# getXML takes two kind of parameters
# - Net::LDAP::Search object reference
# - string value, these values are allowed
#   - all - export whole LDAP DB (limited by LDAP acls)
#   - branch - export what given branch server see (3rd parameter dn is required)
#   - dn - export subtree of given dn (3rd parameter dn is required)
# @par aditional reguires:
# - POS::LDAP
# - XML::Writer
# - IO::File
# - File::Copy
# @return true if success or undef
sub getXML
{
    use POS::LDAP;
    use XML::Writer;
    use IO::File;
    require File::Copy;

    my ($self, $outfile, $rangeOrSearch, %opts) = @_;
    my $ldapData       = undef;
    my $base           = undef;
    my $output         = undef;
    my $allowOverwrite = $opts{'allow_overwrite'};
    my $dn             = $opts{'base'};
    if (defined $outfile) {
        if (-e $outfile && !$allowOverwrite) {
            log_error_print('Specified export file already exists. Aborting export');
            return;
        }
        $self->{'OUTPUT'} = $outfile;
        $output = IO::File->new($outfile, O_WRONLY | O_CREAT | O_TRUNC);
    }
    #is range Net::LDAP::Search?
    if (ref($rangeOrSearch) eq 'Net::LDAP::Search') {
        use POS::LDAP::Cache;
        $ldapData =
          POS::LDAP::Cache->new(CONFIG => $self->{'CONFIG'}, LDAP => 'none_needed', 'LDAP_DATA' => $rangeOrSearch);
        $base = $ldapData->getRoot();
    }
    elsif (ref($rangeOrSearch) eq 'POS::Admin') {
        $ldapData = $rangeOrSearch->{'CACHE'};
        $base     = $rangeOrSearch->{'CACHE'}->getRoot();
    }
    elsif (ref($rangeOrSearch) eq 'POS::LDAP::Cache') {
        $ldapData = $rangeOrSearch;
        $base     = $ldapData->getRoot();
    }
    else {
        # define defaults
        $base = $self->{'CONFIG'}{'ALL_LDAPBASE'};
        my $ldap = POS::LDAP->new(CONFIG => $self->{'CONFIG'});
        if (not defined $ldap) {
            return;
        }
        $ldap->async(1);
        # parse params
        my $op = {};
        $op->{'range'} = $rangeOrSearch;
        $op->{'dn'}    = $dn;

        if (not defined $op->{'range'}) {
            log_error_print("Mandatory parameter 'range' is missing");
            return;
        }
        elsif ($op->{'range'} eq "all") {
            log_debug_print("XML::All");
            #base remain default
        }
        elsif ($op->{'range'} eq "branch") {
            log_debug_print("XML::branch");
            if (not defined $op->{'dn'}) {
                log_error_print("Mandatory parameter 'dn' is missing");
                return;
            }
            else {
                #base remain default, we just relogin
                my $res = $ldap->search(
                    base   => $op->{'dn'},
                    filter => '(objectClass=scLocation)',
                    attrs  => ['userPassword'],
                    scope  => 'base'
                );
                if ($res->is_error()) {
                    log_error_printf("LDAP error %s", $res->error);
                    return;
                }
                elsif ($res->count != 1) {
                    log_crit_print("Cannot find proper branch or database is corrupted!");
                    return;
                }
                my $temppass = $res->shift_entry()->get_value("userPassword");
                #$ldap->unbind();
                $res = $ldap->bind($op->{'dn'}, $temppass);
                if ($res->is_error()) {
                    log_error_printf("LDAP error %s", $res->error());
                    return;
                }
                log_debug_printf("Successfuly logged in as %1", $op->{'dn'});
            }
        }
        elsif ($op->{'range'} eq "dn") {
            log_print_debug("XML::dn");
            if (not defined $op->{'dn'}) {
                log_error_print("Mandatory parameter 'dn' is missing");
                return;
            }
            else {
                $base = $op->{'dn'};
            }
        }
        else {
            log_error_print("Unrecognized range option");
            return;
        }
        use POS::LDAP::Cache;
        $ldapData = POS::LDAP::Cache->new(CONFIG => $self->{'CONFIG'}, BASE => $base, LDAP => $ldap);
        if (not $ldapData) {
            log_error_print("LDAP error %s", $ldapData->error());
            return;
        }
    }
    log_debug_print("Init completed");
    # create XML object

    my $writer = new XML::Writer(
        DATA_MODE       => 1,
        DATA_INDENT     => 4,
        ENCODING        => 'utf-8',
        NAMESPACES      => 1,
        PREFIX_MAP      => { $posnsURI => '', $xincnsURI => $xincnsPX },
        FORCED_NS_DECLS => [$xincnsURI],
        OUTPUT          => $output
    );
    $writer->xmlDecl("utf-8", "no");
    # get ldap data

    # parse ldap and write xml
    $writer->startTag([$posnsURI, 'slepos'], "xmlversion" => $XMLVERSION);
    $self->_parseEntry($writer, $ldapData, $base);
    $writer->endTag([$posnsURI, 'slepos']);
    #done
    $writer->end();
    $output->close() if $output;
    return 1;
} ## end sub getXML
#this, writer, entries, lastDN
#returns dn

## @fn private string _parseEntry($writer, $entries, $lastDN)
# translate Net::LDAP::Entry into XML element
# @param writer XML::Writer object ref
# @param entries List of Net::LDAP::Entry objects to process
# @param lastDN DN of parent entry which subtree is to be parsed
# @note
# _parseEntry is recursive function
# @return DN of last parsed entry
sub _parseEntry
{
    my ($self, $writer, $entries, $entryDN) = @_;
    if ($entryDN) {
        my $entry = $entries->getLDAPentry($entryDN);
        #parse entry
        #get entry objectClass
        my $entity = "entry";
        foreach ($entry->get_value('objectClass')) {
            if ($self->{'ENTITIES'}{$_}) {
                $entity = $_;
                last;
            }
        }
        # myDN  is eq lastDN only first call and myDN is then topmost DN. Others are derived from it
        $writer->startTag($entity, 'dn' => $entryDN eq $entries->getRoot() ? $entryDN : (split(/,/, $entryDN, 2))[0]);
        #parse attributes
        $writer->startTag('attributes');
        foreach my $aname ($entry->attributes) {
            #skip object classes -> they are added in import according to LdapStruct
            next if ($aname eq 'objectClass');
            #skip MD5 sums, they are recreated during import
            next if ($aname eq 'scConfigMd5');
            #skip scLocked, will be recreated during import
            next if ($aname eq 'scLocked');
            #skip dn attribute, it will be recreated by LDAP itself
            next if ($aname eq (split(/=/, $entryDN, 2))[0]);
            # array, because of multival attributes (bnc#764946)
            foreach my $data ($entry->get_value($aname)) {
                #dump data to separate file and insert local path to it instead
                if ($aname eq 'scConfigFileData') {
                    #if output is not stdout (undef)
                    if ($self->{'OUTPUT'}) {
                        #construct filename from output filename + '-' + cn part of dn
                        #in case of existing file, add number between filename and cn part
                        my $prefix   = $self->{'OUTPUT'} . '-';
                        my $suffix   = $entry->get_value('cn') . '.conf';
                        my $filename = $prefix . $suffix;
                        if (-e $filename) {
                            my $x = 1;
                            $x += 1 while (-e $prefix . $x . '-' . $suffix);
                            $filename = $prefix . $x . '-' . $suffix;
                            log_info_printf('%s file exists, using %s for export', $prefix . $suffix, $filename);
                        }
                        my $fh = IO::File->new($filename, O_WRONLY | O_CREAT | O_TRUNC);
                        if (not defined $fh) {
                            log_warn_printf('Unable to open file for configuration export. Storing directly into XML');
                        }
                        else {
                            $fh->print($entry->get_value($aname));
                            if ($fh->error()) {
                                log_error_printf('Unable to store data into external file. Leaving in XML. Error: %s',
                                    $fh->error());
                            }
                            else {
                                #replace $data with $filename path
                                $data = $filename;
                            }
                            $fh->close();
                        }
                    }
                }
                $writer->startTag($aname);
                $writer->characters($data);
                $writer->endTag($aname);
            }
        }
        $writer->endTag('attributes');
        #parse children
        while (1) {
            my $_next = $entries->getUnprocessedEntry();
            last unless ($_next);
            $self->_parseEntry($writer, $entries, $_next);
        }
        #close tag and mark as process entry
        $entries->setFlag($entryDN, 'checked');
        $writer->endTag($entity);
    }
    return;
} ## end sub _parseEntry

################################################################################
# XML validation scheme generator

## @method void getValidationXML($outfile)
# Creates RelaxNG validation schema from POS::Admin::LdapStruct::LDAP_STRUCT structure
# @param outfile output file name or string ref for output to string
# @par usage:
# getValidationXML takes one parameter $outfile which type affects methods behaviour
# - scalar filename is then expected to be string containing filename to store the output
# - ref filename is then expected to be reference to string variable where to store the output
# @par aditional reguires:
# - XML::Writer
# - IO::File
sub getValidationXML
{
    my ($self, $outfile, $version) = @_;
    unless (defined($version)) {
        $version = $XMLVERSION;
    }
    use XML::Writer;
    use IO::File;
    my $output;
    if (ref $outfile) {
        if ($version eq '0.1') {
            #old templates are considered stable and not regenerated
            $$outfile = $POSXMLTEMPLATEDIR . '/validateSLEPOSXMLv0.1.rng';
            return;
        }
        # else another version
        # ...
        # else current version
        $output = $outfile;
    }
    elsif (defined $outfile) {
        if ($version eq '0.1') {
            system("cp ${POSXMLTEMPLATEDIR}/validateSLEPOSXMLv0.1.rng $outfile");
            return;
        }
        # else another version
        # ...
        # else current version
        $output = IO::File->new($outfile, O_WRONLY | O_TRUNC | O_CREAT);
        if (not defined $output) {
            log_error_printf('Unable to create file "%s". Check rights, free disk space', $outfile);
            return;
        }
    }
    else {
        die 'Missing parameter "outfile"';
    }
    my $writer =
      new XML::Writer(DATA_MODE => 1, DATA_INDENT => 4, ENCODING => 'utf-8', NAMESPACES => 0, OUTPUT => $output);
    $writer->xmlDecl("utf-8", "no");
    $writer->pi("xml-stylesheet", 'href="/etc/SLEPOS/template/XML/LDAPschema.xsl" type="text/xsl" media="screen"');
    $writer->startTag(
        'grammar',
        'xmlns'           => 'http://relaxng.org/ns/structure/1.0',
        'xmlns:sch'       => "http://www.ascc.net/xml/schematron",
        'datatypeLibrary' => "http://www.w3.org/2001/XMLSchema-datatypes"
    );
    #processing queue
    #layout
    $writer->startTag('start');
    $writer->startTag(
        'element',
        'name' => 'slepos',
        'ns'   => $posnsURI
    );
    # check allowed SLEPOS XML version
    $writer->startTag('attribute', 'name' => 'xmlversion');
    $writer->startTag('data',      'type' => 'string');
    $writer->startTag('param',     'name' => 'pattern');
    $writer->characters($version);
    $writer->endTag('param');
    $writer->endTag('data');
    $writer->endTag('attribute');
    # add baseDN optional attribute to globaly specify dn suffix
    $writer->startTag('optional');
    $writer->startTag('attribute', 'name' => 'baseDN');
    $writer->startTag('data',      'type' => 'string');
    $writer->startTag('param',     'name' => 'pattern');
    $writer->characters('[^=]*=.+');
    $writer->endTag('param');
    $writer->endTag('data');
    $writer->endTag('attribute');
    $writer->endTag('optional');
    # exported XML can be arbitrary subtree, allow it
    $writer->startTag('interleave');
    foreach my $entry (keys(%{ $self->{'ENTITIES'} })) {
        $writer->startTag('zeroOrMore');
        $writer->emptyTag('ref', 'name' => $entry);
        $writer->endTag('zeroOrMore');
    }
    $writer->endTag('interleave');
    $writer->endTag('element');
    $writer->endTag('start');
    # named classes for individual elements
    foreach my $entry (keys(%{ $self->{'ENTITIES'} })) {
        $writer->startTag('define', 'name' => $entry, 'ns' => $posnsURI);
        $self->_parseStructure($writer, $entry);
        $writer->endTag('define');
    }
    # add common SLEPOS/LDAP data types
    ## dn attribute
#        <attribute name="dn">
#            <data type="string">
#                <param name="pattern">[^=]*=.+</param>
#            </data>
#        </attribute>
    $writer->startTag('define',    'name' => 'dnPartAttribute');
    $writer->startTag('attribute', 'name' => 'dn');
    $writer->startTag('data',      'type' => 'string');
    $writer->startTag('param',     'name' => 'pattern');
    $writer->characters('[^=]*=.+');
    $writer->endTag('param');
    $writer->endTag('data');
    $writer->endTag('attribute');
    $writer->endTag('define');
    $writer->endTag('grammar');
} ## end sub getValidationXML

## @fn private void _parseStructure($writer, $objClass)
# generates RelaxNG validation schema (SLEPOS XML v0.1) for given SLEPOS object class
# @param writer XML::Writer object ref
# @param objClass String containing SLEPOS object class to parse
# @note
# _parseStructure generates RelaxNG schema for validation of mandatory and optional attributes of given objectClass
# then calls _parseSubentries for validation of allowed subelements
sub _parseStructure
{
    my ($self, $writer, $objClass) = @_;
    my $pq = $self->{'ENTITIES'}{$objClass};
    #open entry element
    log_debug_printf("Parsing structure entry %s", $objClass);
    $writer->startTag('element', 'name' => $objClass);
    #there is always dnPart attribute
    $writer->emptyTag('ref', 'name' => 'dnPartAttribute');
    $writer->startTag('optional');
    $writer->startTag('attribute', 'name' => 'modifyType');
    $writer->startTag('choice');
    $writer->startTag('value');
    $writer->characters('insert');
    $writer->endTag('value');
    $writer->startTag('value');
    $writer->characters('replace');
    $writer->endTag('value');
    $writer->startTag('value');
    $writer->characters('update');
    $writer->endTag('value');
    $writer->startTag('value');
    $writer->characters('delete');
    $writer->endTag('value');
    $writer->endTag('choice');
    $writer->endTag('attribute');
    $writer->endTag('optional');
    # optional xml:base when xincluding xmls from different directory
    $writer->startTag('optional');
    $writer->emptyTag('attribute', 'name' => 'xml:base');
    $writer->endTag('optional');
    # remove first must attributes, it is part of DN. Store must copy so we do not destroy ENTITIES
    my @must = @{ $pq->{'attributes'}{'must'} };
    shift @must;
    my $must = scalar @must;
    my $may  = scalar @{ $pq->{'attributes'}{'may'} };
    #parse attributes
    $writer->startTag('element', 'name' => 'attributes');
    # subentries checking was eased for fate#313414 - XML modification
    # modifyType update need to be usable without whole attribute list present
    # currently only allowed attributes, max number is checked and data type/format is checked
    if ($must + $may) {
        my $multivalue;
        # interleave so it is position independent
        $writer->startTag('interleave');
        # multivalues have zeroOrMore
        foreach my $attr (@must, @{ $pq->{'attributes'}{'may'} }) {
            if (defined $POS::Admin::LdapStruct::LDAP_ATTR_STRUCT{$attr}{'SYNTAX'} and
                $POS::Admin::LdapStruct::LDAP_ATTR_STRUCT{$attr}{'SYNTAX'} =~ /SINGLE-VALUE/) {
                $multivalue = 0;
            }
            else {
                $multivalue = 1;
            }
            if ($multivalue) {
                $writer->startTag('zeroOrMore');
            }
            else {
                $writer->startTag('optional');
            }
            $writer->startTag('element', 'name' => $attr);
            _getAttrTypeElement($writer, $attr);
            $writer->endTag('element');
            if ($multivalue) {
                $writer->endTag('zeroOrMore');
            }
            else {
                $writer->endTag('optional');
            }
        }
        $writer->endTag('interleave');
    }
    else {
        $writer->emptyTag('empty');
    }
    $writer->endTag('element');
    #end attribute parsing
    #parse subentries
    my $subentries = $pq->{'subentries'};
    if (defined $subentries) {
        if (scalar @{$subentries} > 1) {
            $writer->startTag('interleave');
            _parseSubentries($writer, $subentries);
            $writer->endTag('interleave');
        }
        elsif (scalar @{$subentries} == 1) {
            _parseSubentries($writer, $subentries);
        }
    }
    #close entry element
    $writer->endTag('element');
} ## end sub _parseStructure

# @fn private void _getAttrTypeElement($writer, $attributeName)
# helper function for RelaxNG validation template to generate proper $attributeName type checking
# @param writer XML::Writer object ref
# @param attributeName string  attribute name
# @note
# Writer must be set in correct position.
# @return void
sub _getAttrTypeElement
{
    my ($writer, $attributeName) = @_;
    my %attrDefs = %POS::Admin::LdapStruct::LDAP_ATTR_STRUCT;
    my ($type, $param);
    if (not defined($attrDefs{$attributeName})) {
        # if we don't know given attribute (common LDAP attributes not specified in SLEPOS Schema)
        $type = 'string';
    }
    elsif (defined($attrDefs{$attributeName}{'data_type'})) {
        # if we know attribute and have specified data type
        $type = $attrDefs{$attributeName}{'data_type'};
    }
    else {
        # if we know attribute but don't specify special type
        $type = 'string';
    }
    #always allow empty type - needed for update modification type and to allow defaults
    $type .= '|empty';
    # used build-in patterns
    our $ip_pattern  = '((1?[0-9]?[0-9]|2[0-4][0-9]|25[0-5]).){3}(1?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])';
    our $mac_pattern = '([0-9a-zA-Z]{2}:){5}[0-9a-zA-Z]{2}';
    our $dn_pattern  = '[^=]*=.+';
    # parse $type for | (ORs)
    my @types = split(/\|/, $type);
    if ($#types > 0) {
        $writer->startTag('choice');
    }
    foreach (@types) {
        # parse $type to type and params
        ($type, undef, $param) = /^([^(]*)(\((.*)\))?$/;
        # ifswitch
        if ($type eq 'boolean') {
            $writer->startTag('choice');
            $writer->startTag('value');
            $writer->characters('TRUE');
            $writer->endTag('value');
            $writer->startTag('value');
            $writer->characters('FALSE');
            $writer->endTag('value');
            $writer->endTag('choice');
        }
        elsif ($type =~ /ip|ip,ip|mac|dn/) {
            $writer->startTag('data',  'type' => 'string');
            $writer->startTag('param', 'name' => 'pattern');
            if ($type eq 'ip') {
                $writer->characters($ip_pattern);
            }
            elsif ($type eq 'ip,ip') {
                $writer->characters($ip_pattern . ',' . $ip_pattern);
            }
            elsif ($type eq 'dn') {
                $writer->characters($dn_pattern);
            }
            elsif ($type eq 'mac') {
                $writer->characters($mac_pattern);
            }
            $writer->endTag('param');
            $writer->endTag('data');
        }
        elsif ($type =~ /unsignedInt|unsignedShort/) {
            if ($param) {
                my ($min, $max) = split(/,/, $param);
                $writer->startTag('data',  'type' => $type);
                $writer->startTag('param', 'name' => 'minInclusive');
                $writer->characters($min);
                $writer->endTag('param');
                $writer->startTag('param', 'name' => 'maxInclusive');
                $writer->characters($max);
                $writer->endTag('param');
                $writer->endTag('data', 'type' => $type);
            }
            else {
                $writer->emptyTag('data', 'type' => $type);
            }
        }
        elsif ($type =~ /enum/) {
            my @enums = split(/,/, $param);
            $writer->startTag('choice');
            foreach (@enums) {
                $writer->startTag('value');
                $writer->characters($_);
                $writer->endTag('value');
            }
            $writer->endTag('choice');
        }
        elsif ($type =~ /string/) {
            if ($param) {
                $writer->startTag('data',  'type' => 'string');
                $writer->startTag('param', 'name' => 'pattern');
                $writer->characters($param);
                $writer->endTag('param');
                $writer->endTag('data');
            }
            else {
                $writer->emptyTag('data', 'type' => 'string');
            }
        }
        elsif ($type eq 'empty') {
            $writer->emptyTag('empty');
        }
    }
    if ($#types > 0) {
        $writer->endTag('choice');
    }
} ## end sub _getAttrTypeElement

## @fn private bool _parseSubentries($writer, $subentries)
# helper function for _parsegenerates RelaxNG validation template
# @param writer XML::Writer object ref
# @param subentries array ref containing SLEPOS object class structure
# @note
# format and data of SLEPOS object class structure are defined in POS::Admin::LdapStruct::LDAP_STRUCT
# @return true on success else undef
sub _parseSubentries
{
    my ($writer, $subentries) = @_;
    # subentries checking was eased for fate#313414 - XML modification
    # modifyType update need to be usable without whole subtree present
    # currently only allowed objectClass and max number is checked
    foreach my $entry (@{$subentries}) {
        my $tag = 'optional';
        #parse quantifiers
        if (@$entry[1] eq '?') {
            $tag = 'optional';
        }
        elsif (@$entry[1] eq '+') {
            $tag = 'zeroOrMore';
        }
        elsif (@$entry[1] eq '*') {
            $tag = 'zeroOrMore';
        }
        elsif (@$entry[1] ne '1') {
            log_warn_print("Unsupported quantifier in LdapStruct");
            return;
        }
        $writer->startTag($tag) if ($tag ne "");
        #parse objectClasses
        if (ref(@$entry[0]) ne 'ARRAY') {
            #simple objectClass
            $writer->emptyTag('ref', 'name' => @$entry[0]);
        }
        else {
            #choice or group
            if ($tag eq 'optional') {
                $writer->startTag('choice');
            }
            else {
                $writer->startTag('group');
            }
            _parseSubentries($writer, @$entry[0]);
            if ($tag eq 'optional') {
                $writer->endTag('choice');
            }
            else {
                $writer->endTag('group');
            }
        }
        $writer->endTag($tag) if ($tag ne "");
    }
    return 1;
} ## end sub _parseSubentries

################################################################################
# XML validator

## @method bool validateXML($xmlfile, $validfile)
# Validates XML file specified by $xmlfile againts validation schema specified by $validfile.
# @param xmlfile filename of XML file to be validated
# @param validfile filename of RelaxNG XML file or undef to autogenerate validation XML from POS::Admin::LdapStruct::LDAP_STRUCT
# @par usage:
# validateXML has two mode of operation
# - validate against custom RelaxNG validation schema if $validfile is provided
# - validate against SLEPOS RelaxNG validation schema generated on-the-fly from POS::Admin::LdapStruct::LDAP_STRUCT
# @note
# Validator understands and expands XML XInclude tags
# @par aditional reguires:
# - XML::LibXML
# - XML::LibXML::RelaxNG
# @return true if success else undef
sub validateXML
{
    my ($self, $xmlfile, $validfile) = @_;
    use XML::LibXML;

    my $parser = XML::LibXML->new();
    # disable network access - should not be needed
    $parser->no_network(1);
    # follow local XIncludes
    $parser->expand_xinclude(1);
    #check and parse xml file
    my $xml;
    if (ref($xmlfile) eq 'XML::LibXML::Document') {
        $xml = $xmlfile;
    }
    elsif (-f $xmlfile) {
        eval { $xml = $parser->parse_file($xmlfile); };
        if ($@) {
            log_error_printf('XML file "%s" is malformed. Parser output:%s%s', $xmlfile, "\n", $@);
            return;
        }
    }
    else {
        log_error_printf('Unable fo find input file "%s"', $xmlfile);
        return;
    }
    #get SLEPOS XML version
    my $version = $xml->findnodes('/*[local-name()="slepos"]/@xmlversion');
    if (!$version and !$version->[0]) {
        log_error_print('Can not determine SLEPOS XML version');
        return;
    }
    $version = $version->[0]->getValue();
    #check and parse validation file
    if (not defined $self->{'XMLVALIDATOR'}) {
        $self->setXMLValidationTemplate($version, $validfile);
    }

    eval { $self->{'XMLVALIDATOR'}->validate($xml); };
    if ($@) {
        log_error_printf('XML document "%s" is not valid SLEPOS XML document. Validator output:%s%s',
            $xmlfile, "\n", $@);
        return;
        #FIXME: this code does not work on SLES11 perl-XML-LibXML (v1.66)
        if (wantarray()) {
            # return element where error occured and dn of entry (or undef when dn is not possible to obtain)
            # parse validator error output to find out where the problem is
            # find last occurence of 'Element %s failed to validate content' or 'Invalid sequence in interleave'
            my $xmlerror = $@;
            my $nerrors  = $xmlerror->__prev_depth();
            # throw all except last 2 errors away
            $nerrors -= 2;
            for (my $i = 0; $i <= $nerrors; $i++) {
                $xmlerror = $xmlerror->_prev();
            }
            # relevant RelaxNG validator error codes (they are not exported into perl module):
            our $XML_RELAXNG_ERR_INTERSEQ     = 11;
            our $XML_RELAXNG_ERR_CONTENTVALID = 25;
            our $XML_RELAXNG_ERR_DATATYPE     = 31;
     # if second last error is code 25, next will be 31 -> need element name (str1) from code 25 which is second to last
     # if second last error is code 11, next will be 25 -> need element name (str1) from code 25 which is last
            if ($xmlerror->code() == $XML_RELAXNG_ERR_INTERSEQ) {
                $xmlerror = $xmlerror->_prev();
            }
            # in str1 validator place element name, in our case it is object class
            my $objClass = $xmlerror->str1();
            # now try to extract dn?
            return ($objClass, "");
        }
        else {
            log_error_printf('XML document "%s" is not valid SLEPOS XML document. Validator output:%s%s',
                $xmlfile, "\n", $@);
            return;
        }
    }
    return 1;
} ## end sub validateXML

## @method bool setXMLValidationTemplate($template)
# Sets validation template to be used in future XML validation
# @param template filename or string containg RelaxNG validation template.
# @note
# template parameter may be ommited, in this case template is generated from LdapStruct
# @par aditional reguires:
# - XML::LibXML
# - XML::LibXML::RelaxNG
# @return false if validation template failed parsing, else true
sub setXMLValidationTemplate
{
    my ($self, $version, $template) = @_;
    my $isFile = 0;
    if (not defined $template) {
        #generate validation schema into string
        $self->getValidationXML(\$template, $version);
    }
# always check if $template is file after getValidationXML, old XML versions are passed as link to file, current versions are strings
# check only first line to avoid perl warning if $template is string
    if (-f ($template =~ /^([^\n]*).*/)[0]) {
        $isFile = 1;
    }
    #validate xml
    my $rng;
    eval {
        if ($isFile) {
            # by supplied relaxng schema
            $rng = XML::LibXML::RelaxNG->new(location => $template);
        }
        else {
            # by generated string
            $rng = XML::LibXML::RelaxNG->new(string => $template);
        }
    };
    if ($@) {
        log_error_printf('Validation schema is malformed. Parser output:%s%s', "\n", $@);
        return;
    }
    $self->{'XMLVALIDATOR'} = $rng;
    return 1;
} ## end sub setXMLValidationTemplate

################################################################################
# XML parser

## @method array parseXML($file, $no_ldap)
# Parse XML file into list of Net::LDAP::Entry objects.
# @param file filename of XML file to be validated or undef to use STDIN
# @param no_ldap optional boolean. If true, POS::LDAP object is created by XML imported. Defaults to true
# @par usage:
# Parser perform 3 tasks
# -# either read XML file specified by $file parameter or open STDIN handle if $file is undef.
# -# validates input against SLEPOS RelaxNG validation schema
# -# parse input using LibXML::SAX and POS::Admin::Ldap2XML::SAXParser parser
# @note
# Parser understands and expands XML XInclude tags
# @note
# During XML parsing is computed nearest common ancestor of imported entries. This specify root element DN even if root
# element is not present in imported data. This DN is then used by POS::LDAP::Cache module for data optimizations.
# @note
# No LDAP mode is useful when converted XML is not supposed to be directly applied to LDAP. POS::Admin::GUI use this mode.
# @par aditional reguires:
# - XML::LibXML
# - XML::LibXML::SAX
# - XML::LibXML::SAX::Parser
# - IO::Handle
# - IO::File
# - POS::Admin::Ldap2XML::SAXParser
# @return array of 2 values
# - first is list of Net::LDAP::Entry objects
# - second is nearest common ancestor (root) of imported entries
sub parseXML
{
    my ($self, $file, $no_ldap) = @_;
    # common init
    use XML::LibXML;
    use XML::LibXML::SAX;
    use XML::LibXML::SAX::Parser;
    use POS::Admin;
    use IO::Handle;
    use IO::File;

    my $sax = POS::Admin::Ldap2XML::SAXParser->new($self->{'CONFIG'}, $no_ldap);
    if (not defined $sax) {
        log_error_printf('Unable to create SAX parser. Probably LDAP connection error');
        return;
    }
    my $parser = XML::LibXML->new();
    # disable network access - should not be needed
    $parser->no_network(1);
    # follow local XIncludes
    $parser->expand_xinclude(1);
    my $doc;
    # try to open input file or use STDIN
    if (defined $file) {
        if (-e $file) {
            eval { $doc = $parser->parse_file($file); };
            if ($@) {
                log_error_printf('XML file "%s" is malformed. Parser output:%s%s', $file, "\n", $@);
                return;
            }
        }
        else {
            log_error_printf('Unable to find input file "%s"', $file);
            return;
        }
    }
    else {
        my $fh = IO::Handle->new();
        if ($fh->fdopen(fileno(STDIN), 'r')) {
            eval { $doc = $parser->parse_fh($fh); };
            if ($@) {
                log_error_printf('XML input is malformed. Parser output:%s%s', "\n", $@);
                return;
            }
            $fh->close();
        }
        else {
            log_error_print('Unable to open input file');
            return;
        }
    }
    if (not defined $doc) {
        log_error_print('Unable to parse input');
        return;
    }
    # validate input XML file with RelaxNG
    return unless ($self->validateXML($doc));
    # create SAX object
    my $gen = XML::LibXML::SAX::Parser->new(Handler => $sax);
    # convert XML to entries array
    eval { $gen->generate($doc); };
    if ($@ or $sax->{'ERROR'}) {
        #errors are reported via log_* functions, just return undef
        return;
    }

    #return all entries and their nearest common ancestor
    return ($sax->{'ADMIN'}, $sax->{'BASE'});
} ## end sub parseXML

## @class POS::Admin::Ldap2XML::SAXParser
# POS::Admin::Ldap2XML::SAXParser class.
# helper class for LibXML SAX parser
# @par Provides
# - callbacks for LibXML::SAX parser
# @par Requires
# - POS::Syslog
# - POS::Admin::LdapStruct
# - Net::LDAP::Entry
# - IO::File
# - Digest::MD5
package POS::Admin::Ldap2XML::SAXParser;

use Net::LDAP::Entry;
use POS::Admin;
use POS::Admin::Constant qw(POS_NO_ERROR);
use POS::Admin::LdapStruct qw(%LDAP_STRUCT);
use POS::Syslog;
use IO::File;
require Digest::MD5;

## @cmethod SAXParser new()
# Create new Ldap2XML::SAXParser object
# @return A new object ref
sub new
{
    my $class = shift;
    $class = ref($class) || $class;
    my $this = bless {}, $class;
    $this->{'ENTRY'}      = undef;
    $this->{'MODIFYTYPE'} = [{ 'SINCE' => '', 'TYPE' => 'insert' }];
    $this->{'ANAME'}      = undef;
    $this->{'AVALUE'}     = undef;
    $this->{'ERROR'}      = undef;
    $this->{'CONFIG'}     = shift;
    $this->{'NO_LDAP'}    = shift;
    #BASE is changing only when new root DN is found. At the end represents the topmost DN
    $this->{'BASE'} = undef;
    #FORCEBASE is never changing, can be specified only in slepos:slepos element. Automatically added to any DN
    $this->{'FORCEBASE'} = '';
    #DN is changing for each entry, reconstructing full entry DN
    $this->{'DN'} = undef;
    my $admin = POS::Admin->new(CONFIG => $this->{'CONFIG'}, LDAP => $this->{'NO_LDAP'} ? 'no_ldap' : undef);
    if (defined $admin and $admin->{'ERROR'} == POS_NO_ERROR) {
        $this->{'ADMIN'} = $admin;
    }
    else {
        return;
    }
    return $this;
} ## end sub new

## @method void start_element($el)
# called by LibXML::SAX when new element is opened.
# @param el XML element struct
# @par Purpose
# if SLEPOS objectClass element is opened
# - new DN is computed and new Net::LDAP::Entry object is created
# - SLEPOS objectClass attributes are added to new Net::LDAP::Entry object
# &lt;attributes&gt; tags are skipped
# all else tags are assumed to be representing attribute names
sub start_element
{
    my ($self, $el) = @_;
    if (defined $LDAP_STRUCT{ $el->{'LocalName'} }) {
        #reconstruct entry DN
        my $dn = $el->{'Attributes'}{'{}dn'}{'Value'};
        if (defined $self->{'DN'}) {
            $dn = $dn . ',' . $self->{'DN'};
        }
        else {
            # this happen only on top level entries
            $dn = $dn . $self->{'FORCEBASE'};
        }
        $self->{'DN'} = $dn;
        #check xml:base - if xinluding xml in other directory, this is set
        if (defined $el->{'Attributes'}{'{http://www.w3.org/XML/1998/namespace}base'}) {
            ($self->{'XMLBASE'}) =
              $el->{'Attributes'}{'{http://www.w3.org/XML/1998/namespace}base'}{'Value'} =~ m@^(.*)\/[^\/]*$@;
        }
        #update BASE
        if (not $self->{'BASE'}) {
            # check if $dn is under ALL_LDAPBASE
            if ($dn !~ $self->{'CONFIG'}{'ALL_LDAPBASE'}) {
                log_error_printf('DN from importing file (%s) is not valid for configured database suffix (%s)',
                    $dn, $self->{'CONFIG'}{'ALL_LDAPBASE'});
                die;
            }
            $self->{'BASE'} = $dn;
        }
        my $base = $self->{'BASE'};
        #check nearest common ancestor
        if ($dn !~ $base) {
            #find common ancestor (supremum)
            while ($base && $dn !~ $base) {
                $base = (split(/,/, $base, 2))[1];
            }
            unless (defined $base) {
                log_error_printf('Incompatible DNs detected. Cannot find common base for %s and %s DNs!',
                    $self->{'BASE'}, $self->{'DN'});
                die;
            }
            $self->{'BASE'} = $base;
        }
        #check modifyType, default is 'insert'
        if ($el->{'Attributes'}{'{}modifyType'}) {
# check previous modify type if new is allowed
# delete and replace do not allow further change of modify types (after delete all are deletes, after replace all are replaces)
            if ($self->{'MODIFYTYPE'}->[0]->{'TYPE'} =~ /delete|replace/ and
                $el->{'Attributes'}{'{}modifyType'}{'Value'} ne $self->{'MODIFYTYPE'}->[0]->{'TYPE'}) {
                log_error_printf(
                    'Invalid modifyType change! Modify type %s cannot be changed to %s. Check entry %s',
                    $self->{'MODIFYTYPE'}->[0]->{'TYPE'},
                    $el->{'Attributes'}{'{}modifyType'}{'Value'},
                    $self->{'DN'}
                );
                die;
            }
            my $modify = { 'TYPE' => $el->{'Attributes'}{'{}modifyType'}{'Value'}, 'SINCE' => $dn };
            unshift(@{ $self->{'MODIFYTYPE'} }, $modify);
        }

        #create entry object
        log_debug_printf('Parsing entry %s, modify type %s', $dn, $self->{'MODIFYTYPE'}->[0]->{'TYPE'});
        $self->{'ENTRY'} = { 'DN' => $dn, 'objectClass' => $el->{'LocalName'} };
    }
    elsif ($el->{'LocalName'} eq 'attributes') {
        #skip attributes tag
        #and clear all remains, bnc#733346
        $self->{'ANAME'} = undef;
        return;
    }
    elsif ($el->{'LocalName'} eq 'slepos') {
        # need this in separate if as we must not set $self->{'ANAME}.
        # ANAME set to slepos can interfere with modityType update and deleting parameters
        if (defined $el->{'Attributes'}{"{}baseDN"}) {
            $self->{'FORCEBASE'} = ',' . $el->{'Attributes'}{"{}baseDN"}{'Value'};
        }
    }
    else {
        #store attribute name
        $self->{'ANAME'} = $el->{'LocalName'};
    }
} ## end sub start_element

## @method void end_element($el)
# called by LibXML::SAX when new element is closed.
# @param el XML element struct
# @par Purpose:
# - adds loaded attribute to currently opened Net::LDAP::Entry object (ref in $self->{'ENTRY'}), in case of
# scConfigFileData or scConfigFileSyncTemplate attributes, computed MD5 sum of their value is added as 'scConfigMd5'
# - store finalized $self->{'ENTRY'} if &lt;/attributes&gt; tag is encountered
# - transition to parent DN if closing SLEPOS objectClass element
sub end_element
{
    my ($self, $el) = @_;
    if ($self->{'ANAME'}) {
        # store attribute only when value is present
        if (defined $self->{'AVALUE'}) {
            if ($self->{'ANAME'} =~ /scConfigFileData/) {
                $self->{'ENTRY'}{'DATA'}{'scConfigMd5'} = Digest::MD5::md5_base64($self->{'AVALUE'});
            }
            elsif ($self->{'ANAME'} =~ /scConfigFileLocalPath/) {
                my $fh;
                if (not open($fh, '<', $self->{'AVALUE'})) {
                    log_error_printf("Can't open configuration file '$self->{'AVALUE'}': $!");
                    die;
                }
                binmode($fh);
                $self->{'ENTRY'}{'DATA'}{'scConfigMd5'} = Digest::MD5->new->addfile(*$fh)->b64digest;
                close $fh;
            }
            # multivalue support
            if (exists $self->{'ENTRY'}{'DATA'}{ $self->{'ANAME'} }) {
                if (ref $self->{'ENTRY'}{'DATA'}{ $self->{'ANAME'} } eq 'ARRAY') {
                    # already converted to multivalue
                    push(@{ $self->{'ENTRY'}{'DATA'}{ $self->{'ANAME'} } }, $self->{'AVALUE'});
                }
                else {
                    # must convert to multivalue
                    $self->{'ENTRY'}{'DATA'}{ $self->{'ANAME'} } =
                      [$self->{'ENTRY'}{'DATA'}{ $self->{'ANAME'} }, $self->{'AVALUE'}];
                }
            }
            else {
                $self->{'ENTRY'}{'DATA'}{ $self->{'ANAME'} } = $self->{'AVALUE'};
            }
            $self->{'AVALUE'} = undef;
        }
        elsif ($self->{'MODIFYTYPE'}->[0]->{'TYPE'} eq 'update') {
            # in case of modifyType update, store undef value as attribute deletion
            $self->{'ENTRY'}{'DATA'}{ $self->{'ANAME'} } = undef;
        }
        # always undef ANAME when closing tag
        $self->{'ANAME'} = undef;
    }
    elsif ($el->{'LocalName'} eq 'attributes') {
        # check if dn-part attribute was specified and add if not
        my ($dnpart, $dnvalue) = $self->{'ENTRY'}{'DN'} =~ /([^=]*)=([^,]*).*/;
        unless ($self->{'ENTRY'}{'DATA'}{$dnpart}) {
            $self->{'ENTRY'}{'DATA'}{$dnpart} = $dnvalue;
        }
        # apply entry
        my $modifyType = $self->{'MODIFYTYPE'}->[0]->{'TYPE'};
        if ($modifyType eq 'insert') {
            $self->{'ADMIN'}->add($self->{'ENTRY'}{'DN'}, $self->{'ENTRY'}{'objectClass'}, $self->{'ENTRY'}{'DATA'}) and
              die;
        }
        elsif ($modifyType eq 'replace') {
            $self->{'ADMIN'}->remove($self->{'ENTRY'}{'DN'}, 0, 0, $self->{'ENTRY'}{'objectClass'}) or die;
            $self->{'ADMIN'}->add($self->{'ENTRY'}{'DN'}, $self->{'ENTRY'}{'objectClass'}, $self->{'ENTRY'}{'DATA'}) and
              die;
        }
        elsif ($modifyType eq 'update') {
            $self->{'ADMIN'}->modify($self->{'ENTRY'}{'DN'}, $self->{'ENTRY'}{'objectClass'}, $self->{'ENTRY'}{'DATA'})
              and
              die;
        }
        elsif ($modifyType eq 'delete') {
            $self->{'ADMIN'}->remove($self->{'ENTRY'}{'DN'}, 0, 0, $self->{'ENTRY'}{'objectClass'}) and die;
        }
        else {
            log_crit_printf('Assert error: bad modifyType %s', $modifyType);
            die;
        }
        # cleanup and set modifyType to insert as default
        undef $self->{'ENTRY'};
    }
    elsif (defined $LDAP_STRUCT{ $el->{'LocalName'} }) {
        if ($self->{'DN'} eq $self->{'MODIFYTYPE'}->[0]->{'SINCE'}) {
            shift @{ $self->{'MODIFYTYPE'} };
        }
        $self->{'DN'} = (split(/,/, $self->{'DN'}, 2))[1];
        unless (defined $self->{'DN'}) {
            log_error_printf('Incomplete DN detected, check dn attribute of top level %s element', $el->{'LocalName'});
            die;
        }
        #remove dynamic suffix on top level to avoid messing with different subtree
        if ($self->{'DN'} !~ $self->{'BASE'}) {
            undef $self->{'DN'};
        }
    }
} ## end sub end_element

## @method void characters($chars)
# called by LibXML::SAX when data between tags are read.
# @param chars string of read data
# @par Purpose:
# stores read data as attribute data. Exception is attribute "scConfigFileData" when data from external file are
# loaded and stores as attribute data
sub characters
{
    my ($self, $chars) = @_;
    if ($self->{'ANAME'}) {
        if ($self->{'ANAME'} eq 'scConfigFileData') {
            # get proper filename
            my $filename = $chars->{'Data'};
            if (defined $self->{'XMLBASE'} and -f $self->{'XMLBASE'} . "/" . $chars->{'Data'}) {
                $filename = $self->{'XMLBASE'} . "/" . $chars->{'Data'};
            }
            $self->{'AVALUE'} = $filename;
        }
        else {
            $self->{'AVALUE'} = $chars->{'Data'};
        }
    }
} ## end sub characters
1;
