#!/usr/bin/perl -w
## @file
# Implementation of branch administration commands

## @class
package POS::FriendlyAdmin;

use strict;

use Net::LDAP::Util qw(escape_filter_value);
use POS::Syslog;
use POS::LdapUtil;
use POS::SyncImages;
use POS::posCrConfig;
use POS::posDataStatic;
use POS::posDataCache;
use POS::Admin;
use POSIX qw(strftime);

# version of the package
our $VERSION = 0.1;


sub get_ws_by_desc
{
    my ($cache, $base, $desc) = @_;

    # let's assume that desc is mac or cn
    my $edesc = escape_filter_value($desc);

    my $res = POS::LdapUtil::ldap_gen_entrycache_successive(
        'BASE'   => $base,
        'CACHE'  => $cache,
        'FILTER' => "(&(objectClass=scWorkstation)(|(macAddress=$edesc)(cn=$edesc)))"
    );

    my @entries = values(%{$res});

    if (@entries > 0) {
        if (@entries > 1) {
            log_warn_printf("Multiple scWorkStation objects for '$desc");
        }

        return $entries[0]->dn;
    }

} ## end sub get_ws_by_desc

sub get_cr_by_desc
{
    my ($cache, $base, $desc) = @_;

    # let's assume that desc is hwtype
    my $edesc = escape_filter_value($desc);

    my $res = POS::LdapUtil::ldap_gen_entrycache_successive(
        'BASE'   => $base,
        'CACHE'  => $cache,
        'FILTER' => "(&(objectClass=scCashRegister)(|(scCashRegisterName=$edesc)(cn=$edesc)))"
    );

    my @entries = values(%{$res});

    if (@entries > 0) {
        if (@entries > 1) {
            log_warn_printf("Multiple scCashRegister objects for '$desc");
        }

        return $entries[0]->dn;
    }
} ## end sub get_cr_by_desc

sub get_role_by_desc
{
    my ($cache, $base, $uptobase, $desc) = @_;


    my $roles = get_roles_from_ldap('CACHE' => $cache, 'UPTOBASE' => $uptobase, 'BASE' => $base, 'SCLOCDN' => $base);

    my $found_dn;
    my $found_index;

    foreach my $dn (keys %$roles) {
        if (($roles->{$dn}{'cn'} eq $desc) ||
            ($roles->{$dn}{'scRoleName'} eq $desc)) {
            if ($found_dn) {
                log_warn_print("multiple roles found for '$desc'");
            }
            if (!defined $found_index || $found_index > $roles->{$dn}{'INDEX'}) {
                $found_dn    = $dn;
                $found_index = $roles->{$dn}{'INDEX'};
            }
        }
    }
    return $found_dn;
} ## end sub get_role_by_desc

sub get_image_by_desc
{
    my ($ldap, $base, $desc) = @_;

    # let's assume that desc is image name, file, or cn
    my $edesc = escape_filter_value($desc);

    my $res = $ldap->search(
        base   => $base,
        scope  => 'sub',
        filter => "(&(objectClass=scPosImage)(|(|(scImageName=$edesc)(scImageFile=$edesc))(cn=$edesc)))"
    );
    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return;
    }
    if ($res->count > 0) {
        if ($res->count > 1) {
            log_warn_printf("Multiple scPosImageName objects for '$desc");
        }

        my $entry = $res->pop_entry;
        return $entry->dn;
    }
} ## end sub get_image_by_desc


## @cmethod object new()
# @return A new object
sub new
{
    my $package = shift;
    my $self    = {
        #init data
        'STATIC' => undef,    ## config data
        'CACHE'  => undef,    ## cached ldap data
    };

    %$self = (%$self, @_);

    return if !defined $self->{'STATIC'};
    return if !defined $self->{'CACHE'};
    return if !defined $self->{'ON_ADMIN'};

    return bless($self);
} ## end sub new


# collects results of get_ids_from_ldap,
# optionally filters them (does not include ids from $exc_id_hash), and
# creates array to apply directly into ldap

sub id_list
{
    my ($self) = @_;
    my ($idinv_list, $id_hash) =
      get_ids_from_ldap('CACHE' => $self->{'CACHE'}, 'BASE' => $self->{'STATIC'}{'BRANCHBASE'});
    foreach my $id (sort (keys(%{$id_hash}))) {
        print "$id\n";
    }
    return 0;
} ## end sub id_list

#empty return means no change
sub _standardize_ldap_ids
{
    my ($self, $id_hash) = @_;
    my $illegal_regexp     = '\;';          #todo get from static?
    my $rewrite_all        = 0;
    my $multiids           = 0;
    my %ldap_clean_id_hash = %{$id_hash};
    foreach my $ldapid (keys %ldap_clean_id_hash) {
        if ($ldap_clean_id_hash{$ldapid}->{'COUNT'} > 1) {    #multiplicities
            log_warn_printf("LDAP id '$ldapid' has multiplicities");
            $rewrite_all = 1;
        }
        if ($ldap_clean_id_hash{$ldapid}->{'MULTILINEIDX'} != $ldap_clean_id_hash{$ldapid}->{'INDEX'}) {   # sepp by ';'
            log_warn_print("LDAP id list has multiple ids on one line (';' separated), will be split") unless $multiids;
            $rewrite_all = 1;
            $multiids    = 1;
        }
        if ($ldapid =~ /$illegal_regexp/) {
            log_warn_print("LDAP id '$ldapid' has illegal name, will be removed");
            $rewrite_all = 1;
            delete $ldap_clean_id_hash{$ldapid};
        }
    }
    return \%ldap_clean_id_hash if $rewrite_all;
    return undef;

} ## end sub _standardize_ldap_ids

sub id_add
{
    my ($self, $ids) = @_;
    my $illegal_regexp = '\;';
    my @in_id_list     = @{$ids};
    my $retval         = 0;

    my ($idinv_list, $id_hash) =
      get_ids_from_ldap('CACHE' => $self->{'CACHE'}, 'BASE' => $self->{'STATIC'}{'BRANCHBASE'});

    #check problems in LDAP data
    my $ldap_clean_id_hr = $self->_standardize_ldap_ids($id_hash);
    log_warn_printf("=> LDAP id list will be reformated") if defined $ldap_clean_id_hr;

    #check problems in input data
    my %in_clean_id_hash;
    my %in_id_hash;
    foreach my $id (@in_id_list) {
        if (exists $in_id_hash{$id}) {
            log_warn_printf("ignoring multiplicities of '$id'");
            $retval = 2;
            next;
        }
        $in_id_hash{$id}++;
        if ($id =~ /$illegal_regexp/) {
            log_warn_print("ID '$id' name is illegal (must not eval to regexp '$illegal_regexp'), will be ignored");
            $retval = 2;
            next;
        }
        if (exists $id_hash->{$id}) {
            log_warn_print("ID '$id' already exists in LDAP, will not be added");
            $retval = 2;
            next;
        }
        $in_clean_id_hash{$id} = 1;
    }

    my @writearray;
    my $reswrite;
    if (defined $ldap_clean_id_hr) {
        @writearray = (keys(%in_clean_id_hash), keys(%{$ldap_clean_id_hr}));
        log_debug_print("replace write: @writearray \n");
        $reswrite =
          $self->{'CACHE'}{'LDAPCACHE'}
          ->modify($self->{'STATIC'}{'BRANCHBASE'}, replace => { scIdPool => [@writearray] })
          if @writearray;
    }
    else {
        @writearray = (keys(%in_clean_id_hash));
        log_debug_print("add write: @writearray \n");
        $reswrite =
          $self->{'CACHE'}{'LDAPCACHE'}->modify($self->{'STATIC'}{'BRANCHBASE'}, add => { scIdPool => [@writearray] })
          if @writearray;
    }
    if ($reswrite && $reswrite->code) {
        log_error_print $reswrite->code, ":", $reswrite->error;
        return 1;
    }

    return $retval;
} ## end sub id_add

sub id_delete
{
    my ($self, $ids, $force) = @_;
    my @in_id_list = @{$ids};
    my ($idinv_list, $id_hash) =
      get_ids_from_ldap('CACHE' => $self->{'CACHE'}, 'BASE' => $self->{'STATIC'}{'BRANCHBASE'});

    #check problems in LDAP data
    my $ldap_clean_id_hr = $self->_standardize_ldap_ids($id_hash);
    log_warn_printf("=> LDAP id list will be reformated") if defined $ldap_clean_id_hr;

    #check problems in input data
    my %in_clean_id_hash;
    my %in_id_hash;
    foreach my $id (@in_id_list) {
        if (exists $in_id_hash{$id}) {
            log_warn_printf("ignoring multiplicities of '$id'");
            next;
        }
        $in_id_hash{$id}++;
        if (!defined $id_hash->{$id}) {
            log_warn_print("ID '$id' does not exists in LDAP, will be ignored...");
            next;
        }
        $in_clean_id_hash{$id} = 1;
    }

    my %res;
    #check if command can proceed
    foreach my $id (keys %in_clean_id_hash) {
        my $eid = escape_filter_value($id);
        my $res = $self->{'CACHE'}{'LDAP'}->search(
            base   => $self->{'STATIC'}{'BRANCHBASE'},
            scope  => 'sub',
            filter => "(&(objectClass=scWorkstation)(scId=$eid))",
        );
        if ($res->code) {
            log_error_printf('LDAP-Error: Can\'t read scWorkstation: Error=%s ', $res->error);
        }
        elsif ($res->count) {
            if (!$force) {
                log_error_printf("ID '$id' is still in use, use --force");
                return 1;
            }
        }
        $res{$id} = $res;
    }
    #apply
    #update ws
    foreach my $id (keys %in_clean_id_hash) {
        foreach my $entry ($res{$id}->entries()) {
            my $mac = $entry->get_value('macAddress');
            log_warn_printf("removing ID '$id' from $mac");
            $self->id_set('', $mac);
        }
    }
    #update pool
    my @writearray;
    my $reswrite;
    if (defined $ldap_clean_id_hr) {
        foreach my $id (keys %{$ldap_clean_id_hr}) {
            @writearray = (@writearray, $id) unless exists $in_clean_id_hash{$id};
        }
        log_debug_print("replace write: @writearray \n");
        $reswrite =
          $self->{'CACHE'}{'LDAPCACHE'}
          ->modify($self->{'STATIC'}{'BRANCHBASE'}, replace => { scIdPool => [@writearray] })
          if @writearray;
    }
    else {
        @writearray = (keys(%in_clean_id_hash));
        log_debug_print("delete: @writearray \n");
        $reswrite =
          $self->{'CACHE'}{'LDAPCACHE'}
          ->modify($self->{'STATIC'}{'BRANCHBASE'}, delete => { scIdPool => [@writearray] })
          if @writearray;
    }
    if ($reswrite && $reswrite->code) {
        log_error_print $reswrite->code, ":", $reswrite->error;
        return 1;
    }

    return 0;
} ## end sub id_delete

sub id_set
{
    my ($self, $id, $mac) = @_;

    my $id_list = get_ids_from_ldap('CACHE' => $self->{'CACHE'}, 'BASE' => $self->{'STATIC'}{'BRANCHBASE'});

    $id = '' unless defined $id;

    if ($id ne '' && !defined $id_list->{$id}) {
        log_error_print("ID '$id' does not exist.");
        return 1;
    }

    my $ws_dn = get_ws_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $mac);

    if (!$ws_dn) {
        log_error_printf("Workstation '$mac' not found");
        return 1;
    }

    my $res;
    if ($id ne '') {
        $res = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => { scId => $id });
    }
    else {
        $res = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, delete => ['scId']);
    }
    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return 1;
    }

    return 0;
} ## end sub id_set

sub cr_list
{
    my ($self) = @_;

    my $base = "cn=global," . $self->{'STATIC'}{'LDAPBASE'};

    my $res = $self->{'CACHE'}{'LDAP'}->search(base => $base, scope => 'one', filter => "(objectClass=scCashRegister)");
    if ($res->code) {
        log_error_printf('LDAP-Error: Can\'t read scCashRegister: Error=%s ', $res->error);
    }
    printf "%-15s %-20s  %s\n", "CN", "HWTYPE", "IMAGE";
    foreach my $entry ($res->entries()) {
        my $image_dn = $entry->get_value('scPosImageDn');
        my $image    = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($image_dn);
        printf "%-15s %-20s  %s\n", $entry->get_value('cn'), $entry->get_value('scCashRegisterName'),
          defined $image ? $image->get_value('scImageName') : '';
    }
    return 0;
} ## end sub cr_list

sub image_list
{
    my ($self) = @_;

    my $sync_data;
    my $image_list;
    my $image_dir;

    $image_list = ldap_gen_entrycache(
        'LDAP'        => $self->{'CACHE'}{'LDAP'},
        'BASE'        => "cn=global," . $self->{'STATIC'}{'LDAPBASE'},
        'OBJECTCLASS' => 'scPosImage'
    );

    if ($self->{'ON_ADMIN'}) {
        $image_dir = $self->{'STATIC'}{'SYNC_IMAGE_DIR'};
    }
    else {
        $image_dir = $self->{'STATIC'}{'TFTP_IMAGE_DIR'};
    }

    print("IMAGE\n");
    printf("   %-10s %-25s %s\n", 'VERSION', 'FILE', 'STATUS');

    foreach my $entry (values %$image_list) {
        print $entry->get_value('scImageName'), "\n";
        my $file = $entry->get_value('scImageFile');
        my %versions;

        foreach my $version ($entry->get_value('scPosImageVersion')) {
            if ($version =~ /(.*);active/) {
                my $ver_str = $1;
                $versions{$ver_str} = 1;
            }
        }

        foreach my $version (
            values %{
                ldap_gen_entrycache(
                    'LDAP'        => $self->{'CACHE'}{'LDAP'},
                    'BASE'        => $entry->dn(),
                    'OBJECTCLASS' => 'scImageVersion'
                )
            }
          ) {
            if ($version->get_value('scDisabled') ne 'TRUE') {
                $versions{ $version->get_value('scVersion') } = 1;
            }
        }

        foreach my $ver_str (sort keys %versions) {
            my $image = POS::Image->new();
            $image->lookup($image_dir, $file, $ver_str);
            printf("   %-10s %-25s %s\n", $ver_str, "$file-$ver_str", $image->valid() ? "OK" : $image->get_error());
        }
    }
    return 0;

} ## end sub image_list

sub _printform
{
    my ($p1, $p2, $p3, $p4, $p5, $p6) = @_;
    printf("%-8s %-17s %-15s %-15s %-10s %-10s\n", $p1, $p2, $p3, $p4, $p5, $p6);
} ## end sub _printform

sub ws_list
{
    my ($self) = @_;

    my $ws_list = ldap_gen_entrycache_successive(
        'CACHE'  => $self->{'CACHE'},
        'BASE'   => $self->{'STATIC'}{'BRANCHBASE'},
        'FILTER' => 'objectClass=scWorkstation'
    );

    _printform('CN', 'MAC', 'IP', 'HWTYPE', 'ID', 'Role');
    foreach my $entry (values %$ws_list) {
        next unless $entry->get_value('macAddress');
        my @mac     = $entry->get_value('macAddress');
        my $role_dn = $entry->get_value('scRoleDn');
        my $role_entry;
        $role_entry = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($role_dn) if $role_dn;
        my $role;
        if (defined $role_entry) {
            $role = "'" . $role_entry->get_value('scRoleName') . "'";
        }
        else {
            $role = "<No role>";
        }
        my $id = $entry->get_value('scId');
        if (defined $id) {
            $id = "'" . $id . "'";
        }
        else {
            $id = "<No ID>";
        }

        my $mac_s = pop @mac;
        my $ip_s  = $entry->get_value('ipHostNumber');
        $ip_s = '<No ip>' unless defined $ip_s;
        my $hwtype_s = $entry->get_value('scPosRegisterType');
        $hwtype_s = '<No type>' unless defined $hwtype_s;
        _printform($entry->get_value('cn'), $mac_s, $ip_s, $hwtype_s, $id, $role);
        foreach $mac_s (@mac) {
            _printform('', $mac_s, '"', '"', '"', '"');
        }
    }
    return 0;

} ## end sub ws_list


sub role_list
{
    my ($self, $hide_global, $hide_local) = @_;

    my $roles = get_roles_from_ldap(
        'CACHE'    => $self->{'CACHE'},
        'UPTOBASE' => $self->{'STATIC'}{'LDAPBASE'},
        'BASE'     => $self->{'STATIC'}{'BRANCHBASE'},
        'SCLOCDN'  => $self->{'STATIC'}{'BRANCHBASE'}
    );

    foreach my $dn (keys %$roles) {
        next if $roles->{$dn}{'GLOBAL'}  && $hide_global;
        next if !$roles->{$dn}{'GLOBAL'} && $hide_local;

        printf("%-20s %-8s %s\n",
            $roles->{$dn}{'scRoleName'},
            $roles->{$dn}{'GLOBAL'} ? "global" : "local",
            $roles->{$dn}{'scRoleDescription'});

        my $res =
          $self->{'CACHE'}{'LDAP'}->search(base => $dn, scope => 'one', filter => "(objectClass=scCashRegister)");
        if ($res->code) {
            log_error_printf('LDAP-Error: Can\'t read scCashRegister: Error=%s ', $res->error);
        }
        elsif (!$res->count) {
            print "   Empty role\n";
        }

        foreach my $entry ($res->entries()) {
            my $image_dn = $entry->get_value('scPosImageDn');
            my $image    = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($image_dn);
            printf "   %-20s: %s\n", $entry->get_value('scCashRegisterName'),
              defined $image ? $image->get_value('scImageName') : '';
        }
    }
    return 0;
} ## end sub role_list

sub copy_cr
{
    my ($self, $source, $dest_role, $force) = @_;

    my $cn = ldap_get_cn_from_dn($source);

    my $res =
      $self->{'CACHE'}{'LDAP'}
      ->search(base => $dest_role, scope => 'sub', filter => "(&(objectClass=scCashRegister)(cn=$cn))");
    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return;
    }
    if ($res->count > 0) {
        if (!$force) {
            log_error_print("Cash Register $cn already exists");
            return;
        }

        foreach my $entry ($res->entries()) {
            $res = $self->{'CACHE'}{'LDAP'}->subtreeDelete($entry->dn);
            if ($res->code) {
                log_error_print "Can't delete '" . $entry->dn . "'";
                $self->{'CACHE'}->update();
                return;
            }
        }
    }

    $res = $self->{'CACHE'}{'LDAP'}->subtreeCopy($source, $dest_role);
    $self->{'CACHE'}->update();

    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return;
    }

    my $dest = "cn=$cn,$dest_role";
    return $dest;
} ## end sub copy_cr

sub role_add
{
    my ($self, $cr, $image, $role, $force) = @_;

    my $source_cr_dn = get_cr_by_desc($self->{'CACHE'}, "cn=global," . $self->{'STATIC'}{'LDAPBASE'}, $cr);

    if (!$source_cr_dn) {
        log_error_print "Unknown Cash Register '$cr'";
        return 1;
    }

    my $image_dn = get_image_by_desc($self->{'CACHE'}{'LDAP'}, "cn=global," . $self->{'STATIC'}{'LDAPBASE'}, $image);

    if (!$image_dn) {
        log_error_print "Unknown Image '$image'";
        return 1;
    }


    my $role_dn =
      get_role_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $self->{'STATIC'}{'LDAPBASE'}, $role);

    if (!$role_dn) {
        log_error_print "Unknown Role '$role'";
        return 1;
    }


    my $target_cr = $self->copy_cr($source_cr_dn, $role_dn, $force);

    return 1 unless $target_cr;

    my $res = $self->{'CACHE'}{'LDAP'}->modify($target_cr, replace => { scPosImageDn => $image_dn });
    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return 1;
    }
    return 0;

} ## end sub role_add

sub role_container_create
{
    my ($self, $base) = @_;


    my $res =
      $self->{'CACHE'}{'LDAP'}->search(base => $base, scope => 'sub', filter => "(objectClass=scRefObjectContainer)");
    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return;
    }
    if ($res->count > 0) {
        if ($res->count > 1) {
            log_warn_printf("Multiple scRefObjectContainer objects under '$base");
        }

        my $entry = $res->pop_entry;
        return $entry->dn;
    }

    my $dn = "cn=default,$base";
    $res =
      $self->{'CACHE'}{'LDAPCACHE'}
      ->add($dn, attr => [objectClass => ['scRefObjectContainer', 'top'], cn => 'default',]);

    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return;
    }
    return $dn;
} ## end sub role_container_create

sub role_create
{
    my ($self, $cn, $role_name, $role_desc) = @_;

    my $base = $self->{'STATIC'}{'BRANCHBASE'};

    my $container = $self->role_container_create($base);

    my $dn = "cn=$cn,$container";

    my $res = $self->{'CACHE'}{'LDAPCACHE'}->add($dn,
        attr => [objectClass => ['scRole', 'top'], cn => $cn, scRoleName => $role_name, scRoleDescription => $role_desc]
    );

    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return 1;
    }
    return 0;
} ## end sub role_create


sub role_for_ws
{
    my ($self, $ws_list, $roles) = @_;

    my $ret = 0;

    foreach my $ws (@$ws_list) {

        my $ws_dn = get_ws_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $ws);

        if (!$ws_dn) {
            log_error_printf("Workstation '$ws' not found");
            $ret = 1;
            next;
        }

        my @role_dns;
        foreach my $role (split /,/, $roles) {
            my $role_dn =
              get_role_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $self->{'STATIC'}{'LDAPBASE'}, $role);

            if (!$role_dn) {
                log_error_print "Unknown Role '$role', skipping";
                next;
            }
            push @role_dns, $role_dn;
        }

        my $res = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => { scAllowedRolesDn => \@role_dns });

        if ($res->code) {
            log_error_print $res->code, ":", $res->error;
            $ret = 1;
        }
    }
    return $ret;

} ## end sub role_for_ws

sub role_set
{
    my ($self, $ws_list, $role) = @_;

    my $role_dn =
      get_role_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $self->{'STATIC'}{'LDAPBASE'}, $role);

    if (!$role_dn) {
        log_error_print "Unknown Role '$role'";
        return 1;
    }

    my $ret = 0;
    foreach my $ws (@$ws_list) {

        my $ws_dn = get_ws_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $ws);

        if (!$ws_dn) {
            log_error_printf("Workstation '$ws' not found");
            $ret = 1;
            next;
        }

        my $res = $self->{'CACHE'}{'LDAPCACHE'}->modify($ws_dn, replace => { scRoleDn => $role_dn });
        if ($res->code) {
            log_error_print $res->code, ":", $res->error;
            $ret = 1;
        }
    }
    return $ret;

} ## end sub role_set

sub role_delete
{
    my ($self, $role, $force) = @_;

    my $role_dn =
      get_role_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $self->{'STATIC'}{'LDAPBASE'}, $role);

    if (!$role_dn) {
        log_error_print "unknown Role '$role'";
        return 1;
    }

    my $res = $self->{'CACHE'}{'LDAP'}->search(base => $role_dn, scope => 'sub', filter => "(objectClass=*)");
    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return 1;
    }
    if ($res->count > 1 && !$force) {    #1 entry is for the role itself
        log_error_printf("Role '$role' is not empty, use --force");
        return 1;
    }

    $res = $self->{'CACHE'}{'LDAP'}->subtreeDelete($role_dn);
    $self->{'CACHE'}->update();
    if ($res->code) {
        log_error_print "Can't delete '$role_dn'";
        return 1;
    }
    return 0;

} ## end sub role_delete

sub role_cr_remove
{
    my ($self, $role, $cr) = @_;

    my $role_dn =
      get_role_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $self->{'STATIC'}{'LDAPBASE'}, $role);

    if (!$role_dn) {
        log_error_print "unknown Role '$role'";
        return 1;
    }

    my $cr_dn = get_cr_by_desc($self->{'CACHE'}, $role_dn, $cr);

    if (!$cr_dn) {
        log_error_print "Unknown Cash Register '$cr' in role '$role'";
        return 1;
    }

    my $res = $self->{'CACHE'}{'LDAP'}->subtreeDelete($cr_dn);
    $self->{'CACHE'}->update();
    if ($res->code) {
        log_error_print "Can't delete '$cr_dn'";
        return 1;
    }

    return 0;
} ## end sub role_cr_remove

sub role_refresh
{
    my ($self) = @_;
} ## end sub role_refresh

sub dump_ws
{
    my ($self, $ws_list) = @_;

    if ($self->{'ON_ADMIN'}) {
        log_error_print("dump works only on Branch Server");
        return 1;
    }

    my $ret = 0;
    foreach my $ws (@$ws_list) {
        my $ws_dn = get_ws_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $ws);

        if (!$ws_dn) {
            log_error_printf("Workstation '$ws' not found");
            $ret = 1;
            next;
        }

        my $crconfig = new POS::posCrConfig('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'});

        $ret = $ret + $crconfig->dumpWsDn($ws_dn, "dumped by pos dump-ws");
    }
    return $ret;
} ## end sub dump_ws

sub dump_all
{
    my ($self) = @_;
    if ($self->{'ON_ADMIN'}) {
        log_error_print("dump works only on Branch Server");
        return 1;
    }
    my $crconfig = new POS::posCrConfig('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'});

    return $crconfig->dumpAll("dumped by pos dump-all");

} ## end sub dump_all

sub remap_all
{
    my ($self) = @_;
    if ($self->{'ON_ADMIN'}) {
        log_error_print("remap-all works only on Branch Server");
        return 1;
    }
    my $crconfig = new POS::posCrConfig('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'});

    return $crconfig->regupAll();

} ## end sub remap_all

sub dump_lists
{
    my ($self) = @_;

    if ($self->{'ON_ADMIN'}) {
        log_error_print("dump works only on Branch Server");
        return 1;
    }
    my $crconfig = new POS::posCrConfig('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'});

    return $crconfig->dumpListsOnly();

} ## end sub dump_lists


sub _read_files
{
    my ($path, $filter) = @_;
    opendir(DIR, $path);
    my %files;
    my @all = readdir(DIR);
    closedir(DIR);
    foreach my $file (@all) {
        if ($file =~ /$filter/ && !-d "$path/$file") {
            $files{$file} += 1;
        }
    }

    return \%files;

} ## end sub _read_files

sub _read_pxe_files
{
    my ($self, $path) = @_;
    my $posbk       = $self->{'STATIC'}{'CONFIG'}{'POS_BACKUP_SUFFIX'};
    my $pxemac_filt = '([0-9a-f]{2}-){6}[0-9a-f]{2}';
    my %pxe_h;
    $pxe_h{'USED'}   = _read_files($path, "^(default|$pxemac_filt)\$");
    $pxe_h{'BACKED'} = _read_files($path, "^(default|$pxemac_filt)\.$posbk\$");
    foreach my $use_d (keys %{ $pxe_h{'USED'} }) {
        if (defined $pxe_h{'BACKED'}{ $use_d . '.' . $posbk }) {
            $pxe_h{'COLLISION'}{$use_d} = 1;
        }
        else {
            $pxe_h{'UNBACKED'}{$use_d} = 1;
        }
    }


    return \%pxe_h;

} ## end sub _read_pxe_files

sub _bck_pxe_files
{
    my ($self, $path, $files_hr) = @_;
    my $posbk = $self->{'STATIC'}{'CONFIG'}{'POS_BACKUP_SUFFIX'};
    foreach my $file (keys %{$files_hr}) {
        log_info_print("disabling $file");
        system('cp', "$path/$file", "$path/$file.$posbk");
    }

} ## end sub _bck_pxe_files

sub _restore_pxe_files
{
    my ($self, $path, $files_hr) = @_;
    my $posbk = $self->{'STATIC'}{'CONFIG'}{'POS_BACKUP_SUFFIX'};
    foreach my $file (keys %{$files_hr}) {
        my $rfile = $file;
        $rfile =~ s/\.$posbk$//;
        log_info_print("restoring $file");
        system('mv', '-f' ,"$path/$file", "$path/$rfile");
    }


} ## end sub _restore_pxe_files

sub _create_pxemenu
{
    my ($self, $path, $files_hr, $imgs_hr) = @_;
    my $posbk  = $self->{'STATIC'}{'CONFIG'}{'POS_BACKUP_SUFFIX'};
    my $postmp = $self->{'STATIC'}{'CONFIG'}{'POS_TMP_PREFIX'};
    my $extdir = $self->{'STATIC'}{'CONFIG'}{'TFTP_CUSTOM_BOOT_SUBDIR'};
    foreach my $filename (keys %{$files_hr}) {
        my $tmppxe = "$path/$postmp$filename";
        log_info_print("enabling and appending menu to  $path/$filename");
        system('cp', "$path/$filename.$posbk", $tmppxe);
        open PXE, ">>", $tmppxe or die $!;
        print PXE "\n";
        print PXE "DEFAULT menu.c32\n";
        print PXE "PROMPT 0 \n";
        print PXE "MENU ENDROW 24 \n";
        print PXE "MENU TITLE Select boot mechanism:\n\n";
        print PXE "MENU BEGIN\n";
        print PXE "MENU LABEL firmware change\n";
        print PXE "MENU TITLE available BIOS updates list:\n";
        print PXE "MENU START\n";
        print PXE "MENU TABMSG select correct BIOS image to boot\n\n";

        foreach my $img (sort keys %{$imgs_hr}) {
            print PXE "LABEL $img\n";
            print PXE "KERNEL memdisk initrd=$extdir/$img\n\n";
        }

        print PXE "LABEL SLEPOS boot\n";
        print PXE "TEXT HELP\n";
        print PXE "  continue with standard SLEPOS KIWI netboot or localboot \n";
        print PXE "ENDTEXT\n";
        print PXE "MENU EXIT\n";
        print PXE "MENU END\n";
        close(PXE);
        system('mv', $tmppxe, "$path/$filename");
    }

} ## end sub _create_pxemenu


sub pxe_menu_set
{
    my ($self, $imgpath, $force) = @_;
    # backup all pxe files, error if exist (and no change on error), add pxe menu
    my $tftp_dir = $self->{'STATIC'}{'TFTP_PXE_DIR'};
    #read pxe dir
    my $pxe_hr = $self->_read_pxe_files($tftp_dir);
    if (!defined $pxe_hr->{'USED'}{'default'}) {
        log_error_print("cannot continue, default pxe file not found");
        return 1;
    }
    my @bckpxes = keys %{ $pxe_hr->{'BACKED'} };
    my @colpxes = keys %{ $pxe_hr->{'COLLISION'} };
    if (@bckpxes && !$force) {
        log_error_printf(
"cannot continue, backed up pxe files already exists: (%s) of which (%s) will be overwritten (use --force to override)",
            join(', ', @bckpxes),
            join(', ', @colpxes)
        );
        return 1;
    }
    #copy images to tftp structure
    my $extboot_dir = $self->{'STATIC'}{'TFTP_CUSTOM_BOOT_DIR'};
    system('install', '-d', $extboot_dir) if !-d $extboot_dir;
    if (defined $imgpath) {
        my $imgs_hr = _read_files($imgpath, '');
        foreach my $img (keys %{$imgs_hr}) {
            system('cp', "$imgpath/$img",  $extboot_dir);
        }
    }
    my $imgs_hr = _read_files($extboot_dir, '');
    #bakup pxe files
    $self->_bck_pxe_files($tftp_dir, $pxe_hr->{'USED'});
    #create pxe menu
    $self->_create_pxemenu($tftp_dir, $pxe_hr->{'USED'}, $imgs_hr);
    return 0;
} ## end sub pxe_menu_set

sub pxe_menu_undo
{
    my ($self, $force) = @_;
    # restore all pxe files, error if spec pxe exists, error if no backup default pxe exists  (and no change on error)
    my $posbk  = $self->{'STATIC'}{'CONFIG'}{'POS_BACKUP_SUFFIX'};
    my $pxe_hr = $self->_read_pxe_files($self->{'STATIC'}{'TFTP_PXE_DIR'});


    if (!defined $pxe_hr->{'BACKED'}{ 'default.' . $posbk } && !$force) {
        log_error_printf(
"backup of 'default' pxe not found, either already restored or lost, cannot continue, use --force to override"
        );
        return 1;
    }
    my @unbpxes = keys %{ $pxe_hr->{'UNBACKED'} };
    if (@unbpxes && !$force) {
        log_error_printf(
"cannot continue, specific pxe files without corresponding backup exists, undo would mix them with restored ones: (%s)  (use --force to override)",
            join(', ', @unbpxes)
        );
        return 1;
    }
    $self->_restore_pxe_files($self->{'STATIC'}{'TFTP_PXE_DIR'}, $pxe_hr->{'BACKED'});
    return 0;


} ## end sub pxe_menu_undo


#delete all older thatn sincedate
sub ws_remove_bytime
{
    my ($self, $datetime, $type, $force, $dryrun) = @_;
    my @cn_a;
    my %ws_h;
    my $linuxtime;
    $linuxtime = $self->_checkdate($datetime) if $type eq 'since';
    $linuxtime = $self->_checkdate()          if $type eq 'inlast';
    return 1 if !defined $linuxtime;
    foreach my $cn (keys %{ $self->{'CACHE'}{'WS'}{'ALL'} }) {
        my ($lwsdate, $wsdate) = $self->_getwsdate($cn);
        #my $lwsdate = $self->_checkdate($wsdate);
        $ws_h{$cn} = $wsdate;
        if (defined $lwsdate) {
            @cn_a = (@cn_a, $cn)
              if (($type eq 'since' && $linuxtime > $lwsdate) ||
                ($type eq 'inlast' && $linuxtime - $datetime > $lwsdate));
        }
    }
    if ($dryrun) {    #print dates of all workstation if dryrun
        foreach my $wscn (sort { $ws_h{$a} cmp $ws_h{$b} } keys(%ws_h)) {
            printf("%-15s  %s\n", $wscn, $ws_h{$wscn}) if $dryrun;
        }
    }
    return $self->_ws_remove(\@cn_a, $force, $dryrun);
} ## end sub ws_remove_bytime


sub ws_remove_list
{
    my ($self, $ws_list, $force, $dryrun) = @_;

    my $ret = 0;
    foreach my $ws (@$ws_list) {
        my $ws_dn = get_ws_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $ws);

        if (!$ws_dn) {
            log_error_printf("Workstation '$ws' not found");
            $ret = 1;
            next;
        }
        my $entry = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($ws_dn);

        my @cn_a = ($entry->get_value('cn'));
        $ret = $ret + $self->_ws_remove(\@cn_a, $force, $dryrun);
    }
    return $ret;
} ## end sub ws_remove_list


sub _checkdate
{
    my ($self, $date) = @_;
    my $mainarg = '';
    $mainarg = "-d \"$date\"" if $date;
    my $ldate = `date $mainarg +%s 2>&1`;
    $ldate =~ s/\n$//;
    if ($?) {
        log_error_printf("Problem parsing date '%s' : %s, aborting...", $date, $ldate);
        return undef;
    }
    return $ldate;
} ## end sub _checkdate

#returns workstatio date in 'date' compatible format
sub _getwsdate
{
    my ($self, $wscn) = @_;
    my $lwsdate = $self->{'CACHE'}{'WS'}{'ALL'}{$wscn}->get_value('scLastBootTime');    #linux time
    return (undef, "no boot time yet") if !defined $lwsdate;
    my $wsdate = POSIX::strftime('%x %X %Z', localtime($lwsdate));
    return ($lwsdate, $wsdate);
} ## end sub _getwsdate


sub _test_ws
{
    my ($self, $cn_ar) = @_;
    my $cmd = 'nmap --system-dns -sP -oG - ' . join(' ', @{$cn_ar}) . ' 2>&1';
    my $res = `$cmd`;
    my %isonline;
    foreach my $cn (@{$cn_ar}) {
        if ($res =~ /Host:.*\($cn(\..*|)\).*Up\n/) {
            $isonline{$cn} = 1;
        }
    }
    return \%isonline;
} ## end sub _test_ws

sub _ws_remove
{
    my ($self, $cn_ar, $force, $dryrun) = @_;
    my $isonline_hr;
    unless ($force) {
        log_info_printf("Testing if some workstations are online:");
        $isonline_hr = $self->_test_ws($cn_ar) unless $force;
    }
    my $n_online = keys %{$isonline_hr};
    foreach my $cn (@{$cn_ar}) {
        if (!$force && $isonline_hr->{$cn}) {
            log_warn_printf("Workstation '%s' is online, will not be deleted.", $cn);
            next;
        }
        my $ws = $self->{'CACHE'}{'WS'}{'ALL'}{$cn};
        if (defined $ws) {
            if ($dryrun) {
                printf("Workstation %-15s would be deleted \n", "'" . $cn . "'");
            }
            else {
                log_info_printf("Deleting %-15s", "'" . $cn . "'");
                my $res = $self->{'CACHE'}{'LDAP'}->subtreeDelete($ws->dn);
                if ($res->code) {
                    log_error_printf("Can't delete %s, aborting...", $ws->dn);
                    $self->{'CACHE'}->update();
                    return 1;
                }
            }
        }
        else {
            log_warn_printf("Workstation '$cn' not found in LDAP, skipping action.");
        }
    }
    return 0;

} ## end sub _ws_remove

sub role_mode
{
    my ($self, $attrib, $value) = @_;
    my $branch_dn = $self->{'STATIC'}{'BRANCHBASE'};
    my $res = $self->{'CACHE'}{'LDAPCACHE'}->modify($branch_dn, replace => { $attrib => $value });
    return $res;
} ## end sub role_mode


sub role_status
{
    my ($self) = @_;
    printf("rolebased mode enabled:    %s\n", $self->{'CACHE'}{'BRANCH'}{'ALLOWROLEBASED'}   ? 'TRUE' : 'FALSE');
    printf("global roles enabled  :    %s\n", $self->{'CACHE'}{'BRANCH'}{'ALLOWGLOBALROLES'} ? 'TRUE' : 'FALSE');
    return 0;
} ## end sub role_status


sub cfg_regenerate_md5
{
    my ($self, $ws, $role, $cr, $cn) = @_;
    my $res;
    my $branch_dn = $self->{'STATIC'}{'BRANCHBASE'};
    if (defined $ws) {
        my $ws_dn = get_ws_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $ws);
        if (!$ws_dn) {
            log_error_printf("Workstation '$ws' not found");
            return 1;
        }
        $res = $self->cfg_regenerate_md5_dn('cn=' . $cn . ',' . $ws_dn);
    }
    else {
        my $dn;
        #get role $dn
        $dn = get_role_by_desc($self->{'CACHE'}, $self->{'STATIC'}{'BRANCHBASE'}, $self->{'STATIC'}{'LDAPBASE'}, $role);
        if (!$dn) {
            log_error_print "unknown Role '$role'";
            return 1;
        }
        if ($cr) {
            $dn = get_cr_by_desc($self->{'CACHE'}, $dn, $cr);
            if (!$dn) {
                log_error_print "Unknown Cash Register '$cr' in role '$role'";
                return 1;
            }
        }
        $res = $self->cfg_regenerate_md5_dn('cn=' . $cn . ',' . $dn);
    }

    return $res;

} ## end sub cfg_regenerate_md5


sub cfg_regenerate_md5_dn
{
    my ($self, $cfg_dn) = @_;
    my $md5;
    my $entry = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($cfg_dn);
    if (!defined $entry) {
        log_error_print "Unknown object DN: $cfg_dn";
        return 1;
    }

    my $objclass = $entry->get_value('objectClass');
    if ($objclass eq 'scConfigFileTemplate') {
        $md5 = POS::Admin::_createMd5FromEntry($self, $entry, 'scConfigFileData');
    }
    elsif ($objclass eq 'scConfigFileSyncTemplate') {
        my $file = $entry->get_value('scConfigFileLocalPath');
        $md5 = POS::Admin::_createMd5FromFile($self, $file);
        return 1 if ($md5 == POS::Admin::Constant::POS_FILE_NOT_FOUND);
    }
    my $res = $self->{'CACHE'}{'LDAPCACHE'}->modify($cfg_dn, replace => { scConfigMd5 => $md5 });
    return 0;

} ## end sub cfg_regenerate_md5_dn


1;
