#!/usr/bin/perl -w

package POS::LDAP;

use strict;
use warnings;
use Net::LDAP;
use Net::LDAPS;
use Net::LDAP::Constant qw(
  LDAP_SUCCESS
  LDAP_COMPARE_TRUE
  LDAP_COMPARE_FALSE
  LDAP_OPERATIONS_ERROR
  LDAP_UNAVAILABLE
  LDAP_ADMIN_LIMIT_EXCEEDED
  LDAP_NO_SUCH_OBJECT
  LDAP_NO_SUCH_ATTRIBUTE
  LDAP_ALREADY_EXISTS
  LDAP_INSUFFICIENT_ACCESS);
use POS::Syslog;

# version  of the package
our @ISA     = qw(Net::LDAP);
our $VERSION = '0.3';

# Constructor - either parameter-less or with POS::Config object as parameter
sub new
{
    my $package = shift;    ## packagename
    $package = ref($package) || $package;

    # Definition of defaults
    my $defaults = {
        'TIMEOUT'     => 12,       ## timeout for LDAP operations
        'LOCKTIMEOUT' => 300,      ## Locking timeout
        'LDAPBINDDN'  => undef,    ## LDAP bind DN for modifying operations
        'LDAPBINDPW'  => undef,    ## LDAP bind password for modifying operations
        'LDAPHOST'    => undef,    ## actual ldap host
        'MAXRETRIES'  => 2,        ## max. number of retries
        'LDAPCONN'    => undef,    ## LDAP connection (Net::LDAP object)
        'NEEDCAP'     => "ro",     ## needed capability ("ro"|"rw")
        'CAFILE'      => undef,    ## path to CA certificate
        'ONLINE'      => 0,        ## AS is accessible (1) or not (0)
        'BASE'        => undef     ## default BASE (LDAP_BASE on AS, BRANCH_BASE on BS)
    };

    #bless to an object
    my $self = bless($defaults, $package);
    # get constructor parameters
    my %opts = @_;
    #parse options
    return unless _parseOptions($self, \%opts);
    #return undef in failed connect
    return unless _connect($self);
    # return the object
    return $self;
} ## end sub new

sub _parseOptions
{

    my ($self, $opts) = @_;

    # Check if config is provided, else load default
    use POS::Config;
    if (ref($opts->{'CONFIG'}) ne 'POS::Config') {
        log_debug_print('POS::Config not defined, creating new');
        $opts->{'CONFIG'} = new POS::Config();
    }
    # load bind dn and password, use CONFIG as default
    $self->{'LDAPBINDDN'} = $opts->{'LDAPBINDDN'} || $opts->{'CONFIG'}{'POS_ADMIN_DN'};
    $self->{'LDAPBINDPW'} = $opts->{'LDAPBINDPW'} || $opts->{'CONFIG'}{'POS_ADMIN_PASSWORD'};
    # load ldap host, try opts param as first
    $self->{'LDAPHOST'} = $opts->{'LDAPHOST'};
    # now check if was specified and use CONFIG default if not
    if (!defined $self->{'LDAPHOST'}) {
   # check if BS, then connect ALL_BRANCHLDAP, else connect ALL_LDAPHOST, needed capability is determined by proxy chain
   # do not change this else synchronized (s_add, s_modify, ..) functions may stop working correctly
        if (defined $opts->{'CONFIG'}{'POS_BRANCH_DEPLOYED'} && !defined $opts->{'CONFIG'}{'POS_ADMIN_DEPLOYED'}) {
            if (not defined $opts->{'CONFIG'}{'ALL_BRANCHLDAP'}) {
                log_alert_print('ALL_BRANCHLDAP is not defined in configuration file. Configuration error.');
                return;
            }
            $self->{'LDAPHOST'} = $opts->{'CONFIG'}{'ALL_BRANCHLDAP'};
        }
        else {
            if (not defined $opts->{'CONFIG'}{'ALL_LDAPHOST'}) {
                log_alert_print('ALL_LDAPHOST is not defined in configuration File. Configuration error.');
                return;
            }
            $self->{'LDAPHOST'} = $opts->{'CONFIG'}{'ALL_LDAPHOST'};
        }
    }
    # now set proper LDAP URI prefix
    # check if address is in ldapi format
    if ($self->{'LDAPHOST'} !~ m/^ldap[s]{0,1}:\/\//) {    # eclipse EPIC syntax highlite fix ]]
                                                           #adress is not an URI -> convert
        if ($opts->{'CONFIG'}{'ALL_LDAPSSL'} == 0) {
            $self->{'LDAPHOST'} = "ldap://" . $self->{'LDAPHOST'};
        }
        else {
            $self->{'LDAPHOST'} = "ldaps://" . $self->{'LDAPHOST'};
        }
    }
    # now copy CA path from CONFIG, so POS::Config object is not necessary to keep
    $self->{'CAFILE'} = $opts->{'CONFIG'}{'CERTPATH'} . '/ca.crt';
    # set default LDAP base
    if (defined $opts->{'CONFIG'}{'BRANCH_LDAPBASE'}) {
        $self->{'BASE'} = $opts->{'CONFIG'}{'BRANCH_LDAPBASE'};
    }
    else {
        $self->{'BASE'} = $opts->{'CONFIG'}{'ALL_LDAPBASE'};
    }
    #we are done here
    log_debug_printf('using BIND DN "%s" on LDAP HOST "%s"', $self->{'LDAPBINDDN'}, $self->{'LDAPHOST'});
    return 1;
} ## end sub _parseOptions

sub _connect
{
    my ($self) = @_;
    # make LDAP connection
    log_debug_printf('connecting to LDAP "%s"', $self->{'LDAPHOST'});
    my $ldap = Net::LDAP->new(
        $self->{'LDAPHOST'},
        version => '3',
        verify  => 'require',           #this can remain.. if not using ssl, this is ignored
        cafile  => $self->{'CAFILE'}    #same here
    );
    # and check it
    if ($ldap) {
        $self->{'LDAPCONN'} = $ldap;
        if ($self->{'LDAPBINDDN'}) {
            my $mesg = $ldap->bind($self->{'LDAPBINDDN'}, password => $self->{'LDAPBINDPW'});
            if ($mesg->code) {
                log_error_printf('error binding to LDAP Server %s : %s', $self->{'LDAPHOST'}, $mesg->error);
                return;
            }
        }
    }
    else {
        log_error_printf('error connecting to LDAP Server %s', $self->{'LDAPHOST'});
        return;
    }
    #all ok
    log_debug_print('LDAP connection successful');
    return 1;
} ## end sub _connect

sub bind
{
    my ($self, $dn, $pw) = @_;
    my $ldap = $self->{'LDAPCONN'};
    my $res = $ldap->bind($dn, password => $pw);
    return $res;
} ## end sub bind


sub search
{
    my $self        = shift;
    my @search_args = @_;
    my $ldap        = $self->{'LDAPCONN'};
    my $result      = $ldap->search(@search_args);
    return $result;
} ## end sub search


sub add
{
    my $self    = shift;
    my @args    = @_;
    my $timeout = $self->{'TIMEOUT'};
    my $result  = $self->{'LDAPCONN'}->add(@args);
    while ($result->code != LDAP_SUCCESS and
        $result->code != LDAP_ALREADY_EXISTS      and
        $result->code != LDAP_INSUFFICIENT_ACCESS and
        $timeout > 0) {
        if ($result->code == LDAP_OPERATIONS_ERROR) {
            log_info_print('connection to LDAP server expired, creating new connection');
            $self->_connect();
        }
        else {
            log_warn_printf("AdminServer LDAP unavailable, waiting 10s for next try... (remaining tries: %s)",
                $timeout);
        }
        sleep(10);
        $timeout -= 1;
        if (defined $self->{'LDAPCONN'}) {
            $result = $self->{'LDAPCONN'}->add(@args);
        }
    }
    return $result;
} ## end sub add

## $ @fn public LDAPresult s_add($self, $dn, $attributes)
# Synchronized LDAP ADD function. Performs POS::LDAP::add and then wait and check for added entry to appear on local LDAP
# Note: connection to LDAP is provided by POS::LDAP module
# @param $self  Blessed POS::LDAP hash object
# @param $dn LDAP DN of adding entry
# @param $attributes attributes of adding entry
# @return Net::LDAP::Message object
sub s_add
{
    my $self   = shift;
    my @args   = @_;
    my $result = $self->add(@args);
    if ($result->code == LDAP_UNAVAILABLE || $result->code == LDAP_ADMIN_LIMIT_EXCEEDED) {
        log_warn_printf("LDAP ADD failed due to LDAP server unavailibility!");
    }
    elsif ($result->code == LDAP_SUCCESS) {
        # sleep for 0.25s, should be enought to sync change on broadband
        sleep(0.25);
        my $iter = 0;
        while ($self->compare(@args) != LDAP_COMPARE_TRUE && $iter < 9) {
            $iter++;
            sleep(1);
        }
        # if we timedout and LDAP still not synced -> error (waiting time is > 10s)
        if ($iter == 9) {
            log_warn_print('LDAP DBs are not yet synchronized, continuing with warning');
        }
    }
    return $result;
} ## end sub s_add

sub moddn
{
    my $self = shift;
    # my $dn = shift;
    my @args    = @_;
    my $timeout = $self->{'TIMEOUT'};
    my $result  = $self->{'LDAPCONN'}->moddn(@args);
    while ($result->code != LDAP_SUCCESS and
        $result->code != LDAP_ALREADY_EXISTS      and
        $result->code != LDAP_INSUFFICIENT_ACCESS and
        $timeout > 0) {
        if ($result->code == LDAP_OPERATIONS_ERROR) {
            log_info_print('connection to LDAP server expired, creating new connection');
            $self->_connect();
        }
        else {
            log_warn_printf("AdminServer LDAP unavailable, waiting 10s for next try... (remaining tries: %s)",
                $timeout);
        }
        sleep(10);
        $timeout -= 1;
        if (defined $self->{'LDAPCONN'}) {
            $result = $self->{'LDAPCONN'}->moddn(@args);
        }
    }
    return $result;

} ## end sub moddn

sub modify
{
    my $self = shift;
    # my $dn = shift;
    my @args    = @_;
    my $timeout = $self->{'TIMEOUT'};
    my $result  = $self->{'LDAPCONN'}->modify(@args);
    while ($result->code != LDAP_SUCCESS and $result->code != LDAP_INSUFFICIENT_ACCESS and $timeout > 0) {
        if ($result->code == LDAP_OPERATIONS_ERROR) {
            log_info_print('connection to LDAP server expired, creating new connection');
            $self->_connect();
        }
        else {
            log_warn_printf("AdminServer LDAP unavailable, waiting 10s for next try... (remaining tries: %s)",
                $timeout);
        }
        sleep(10);
        $timeout -= 1;
        if (defined $self->{'LDAPCONN'}) {
            $result = $self->{'LDAPCONN'}->modify(@args);
        }
    }
    return $result;
} ## end sub modify

## $ @fn public LDAPresult s_modify($self, $dn, $param)
# Synchronized LDAP MODIFY function. Performs POS::LDAP::modify and then wait and check for modified entry to appear on local LDAP
# Note: connection to LDAP is provided by POS::LDAP module
# @param $self  Blessed POS::LDAP hash object
# @param $dn LDAP DN of modifying entry
# @param $param instructions to what and how to modify given entry
# @return Net::LDAP::Message object
sub s_modify
{
    my $self   = shift;
    my @args   = @_;
    my $result = $self->modify(@args);
    if ($result->code == LDAP_UNAVAILABLE || $result->code == LDAP_ADMIN_LIMIT_EXCEEDED) {
        log_warn_printf("LDAP MODIFY failed due to LDAP server unavailibility!");
    }
    elsif ($result->code == LDAP_SUCCESS) {
        # sleep for 0.25s, should be enought to sync change on broadband
        sleep(0.25);
        my $iter = 0;
        while ($self->compare(@args) != LDAP_COMPARE_TRUE && $iter < 9) {
            $iter++;
            sleep(1);
        }
        # if we timedout and LDAP still not synced -> error (waiting time is > 10s)
        if ($iter == 9) {
            log_warn_print('LDAP DBs are not yet synchronized, continuing with warning');
        }
    }
    return $result;
} ## end sub s_modify

sub delete
{
    my $self    = shift;
    my $dn      = shift;
    my $timeout = $self->{'TIMEOUT'};
    my $result  = $self->{'LDAPCONN'}->delete($dn);
    while ($result->code != LDAP_SUCCESS and
        $result->code != LDAP_INSUFFICIENT_ACCESS and
        $result->code != LDAP_NO_SUCH_OBJECT      and
        $timeout > 0) {
        if ($result->code == LDAP_OPERATIONS_ERROR) {
            log_info_print('connection to LDAP server expired, creating new connection');
            $self->_connect();
        }
        else {
            log_warn_printf("AdminServer LDAP unavailable, waiting 10s for next try... (remaining tries: %s)",
                $timeout);
        }
        sleep(10);
        $timeout -= 1;
        if (defined $self->{'LDAPCONN'}) {
            $result = $self->{'LDAPCONN'}->delete($dn);
        }
    }
    return $result;
} ## end sub delete

## $ @fn public LDAPresult s_delete($dn)
# Synchronized LDAP DELETE operation. Performs POS::LDAP::delete and then wait and check until entry disappears on local LDAP
# Note: connection to LDAP is provided by POS::LDAP module
# @param $dn LDAP DN of modifying entry
# @return Net::LDAP::Message object
sub s_delete
{
    my $self   = shift;
    my $dn     = shift;
    my $result = $self->delete($dn);
    if ($result->code == LDAP_UNAVAILABLE || $result->code == LDAP_ADMIN_LIMIT_EXCEEDED) {
        log_warn_printf("LDAP DELETE failed due to LDAP server unavailibility!");
    }
    elsif ($result->code == LDAP_SUCCESS) {
        # sleep for 0.25s, should be enought to sync change on broadband
        sleep(0.25);
        my $iter = 0;
        while ($self->search(base => $dn, scope => 'base', filter => '(objectClass=*)')->code != LDAP_NO_SUCH_OBJECT &&
            $iter < 9) {
            $iter++;
            sleep(1);
        }
        # if we timedout and LDAP still not synced -> error (waiting time is > 10s)
        if ($iter == 9) {
            log_warn_print('LDAP DBs are not yet synchronized, continuing with warning');
        }
    }
    return $result;
} ## end sub s_delete

sub unbind
{
    my $self   = shift;
    my $ldap   = $self->{'LDAPCONN'};
    my $result = $ldap->unbind;
    return $result;
} ## end sub unbind

sub root_dse
{
    my $self   = shift;
    my $ldap   = $self->{'LDAPCONN'};
    my $result = $ldap->root_dse();
    return $result;
} ## end sub root_dse

sub check
{
    my $self   = shift;
    my $result = $self->root_dse();
    return (defined($result));
} ## end sub check

sub max
{
    my @list = @_;
    my $max  = $list[0];
    foreach my $elem (@list) {
        if ($elem > $max) {
            $max = $elem;
        }
    }
    return $max;
} ## end sub max

sub subtreeMove
{
    my $self     = shift;
    my $dn       = shift;
    my $location = shift;

    my $result = $self->search(base => $dn, scope => 'base', filter => '(objectClass=top)');
    if ($result->count == 1) {
        my $entry = $result->pop_entry;
        my @orig = split(/,/, $dn);
        $entry->dn($orig[0] . "," . $location);
        my $imageResult = $self->add($entry);
        if ($imageResult->code == 0) {
            my $childResult = $self->search(base => $dn, scope => 'one', filter => '(objectClass=top)');
            while ($childResult->count > 0) {
                my $child      = $childResult->pop_entry;
                my $childDn    = $child->dn;
                my $moveResult = $self->subtreeMove($childDn, $orig[0] . "," . $location);
                if ($moveResult->code) {
                    return $moveResult;
                }
            }
            return $self->delete($dn);
        }
        else {
            return $imageResult;
        }
    }
    else {
        # Original entry not found
        return $result;
    }
} ## end sub subtreeMove

sub subtreeCopy
{
    my $self     = shift;
    my $dn       = shift;
    my $location = shift;

    my $result = $self->search(base => $dn, scope => 'base', filter => '(objectClass=top)');
    if ($result->count == 1) {
        my $entry = $result->pop_entry;
        my @orig = split(/,/, $dn);
        $entry->dn($orig[0] . "," . $location);
        my $imageResult = $self->add($entry);
        if ($imageResult->code == 0) {
            my $childResult = $self->search(base => $dn, scope => 'one', filter => '(objectClass=top)');
            while ($childResult->count > 0) {
                my $child      = $childResult->pop_entry;
                my $childDn    = $child->dn;
                my $moveResult = $self->subtreeCopy($childDn, $orig[0] . "," . $location);
                if ($moveResult->code) {
                    return $moveResult;
                }
            }
            return $imageResult;
        }
        else {
            return $imageResult;
        }
    }
    else {
        # Original entry not found
        return $result;
    }
} ## end sub subtreeCopy

sub subtreeDelete
{
    my $self = shift;
    my $dn   = shift;

    my $result = $self->search(base => $dn, scope => 'base', filter => '(objectClass=top)');
    if ($result->count == 1) {
        my $childResult = $self->search(base => $dn, scope => 'one', filter => '(objectClass=top)');
        while ($childResult->count > 0) {
            my $child     = $childResult->pop_entry;
            my $childDn   = $child->dn;
            my $delResult = $self->subtreeDelete($childDn);
            if ($delResult->code) {
                return $delResult;
            }
        }
        return $self->delete($dn);
    }
    else {
        # Original entry not found
        return $result;
    }
} ## end sub subtreeDelete

## $ @fn public LDAPresult compare($self, @params)
# This function compare entry given in @params to BS LDAP
# @param $self  Blessed LDAP hash object
# @param @params various params depending on LDAP action called
# @return LDAP_COMPARE_TRUE if equal, LDAP_COMPARE_FALSE if not equal, LDAP_OPERATIONS_ERROR otherwise
sub compare
{
    my $self = shift;
    my $dn;
    my %newData;
    my %deleteData;
    #parse params
    #first param is always dn, either scalar or as dn => 'dn'
    $dn = shift;
    if ($dn eq "dn") {
        $dn = shift;
    }
    while (@_) {
        my $action = shift;
        #add second param is "attr"
        if ($action eq "attr") {
            #parse add operants
            # add uses ($dn, $attrname, $attrvalue, $attrname, $attrvalue ...)
            log_debug_print('POS::LDAP::compare - parsing add attributes');
            my $attrs = shift;
            %newData = (%newData, (ref $attrs eq 'HASH') ? %$attrs : @$attrs);
        }
        #modify second param may be "add", "replace",  "delete", "increment" and "changes"
        #parse "add" and "replace"
        elsif ($action eq 'add' || $action eq 'replace') {
            log_debug_print('POS::LDAP::compare - parsing "add/replace" modify');
            my $attrs = shift;
            %newData = (%newData, (ref $attrs eq 'HASH') ? %$attrs : @$attrs);
        }
        elsif ($action eq 'delete') {
            log_debug_print('POS::LDAP::compare - parsing "delete" modify');
            my $attrs = shift;
            if (ref $attrs eq 'ARRAY') {
                foreach (@$attrs) {
                    $deleteData{$_} = undef;
                }
            }
            elsif (ref $attrs eq 'HASH') {
                %deleteData = (%deleteData, %$attrs);
            }
        }
        #parse "changes"
        elsif ($action eq 'changes') {
            log_debug_print('POS::LDAP::compare - parsing "change" modify');
            #just shift to next, under changes its the same as add/replace/delete
        }
        #info about other unsupported
        else {
            #unsupported features like increment
            log_info_printf('POS::LDAP::compare - action "%s" is not supported, skipping', $action);
        }
    }
    # check mandatory DN
    if ($dn eq '') {
        log_error_print('POS::LDAP::compare - DN argument missing!');
        return LDAP_OPERATIONS_ERROR;
    }
    log_info_print('POS::LDAP::compare - missing attributes to compare! Checking only DN existence')
      if (not %newData and not %deleteData);
    # compare attributes
    foreach my $aname (keys %newData) {
        my @avalue;
        if (ref $newData{$aname} eq 'ARRAY') {
            @avalue = @{ $newData{$aname} };
        }
        else {
            @avalue = ($newData{$aname});
        }
        foreach (@avalue) {
            my $result = $self->{'LDAPCONN'}->compare($dn, attr => $aname, value => $_);
            if ($result->code != LDAP_COMPARE_TRUE) {
                log_debug_printf('POS::LDAP::compare - attr %s: %s is not equal', $aname, $_);
                return LDAP_COMPARE_FALSE;
            }
        }
    }
    foreach my $aname (keys %deleteData) {
        my @avalue;
        if (ref $deleteData{$aname} eq 'ARRAY') {
            @avalue = @{ $deleteData{$aname} };
        }
        else {
            @avalue = ($deleteData{$aname});
        }
        foreach (@avalue) {
            my $result;
            if (defined $_) {
                $result = $self->{'LDAPCONN'}->compare($dn, attr => $aname, value => $_);
            }
            else {
                $result =
                  $self->{'LDAPCONN'}->search(base => $dn, scope => 'base', filter => "(&(objectClass=*)($aname=*))");
            }
            if ($result->code() == LDAP_COMPARE_TRUE ||
                ($result->code() == LDAP_SUCCESS && $result->count() != 0)) {
                log_debug_printf('POS::LDAP::compare - attr %s is still present', $aname);
                return LDAP_COMPARE_FALSE;
            }
        }
    }
    return LDAP_COMPARE_TRUE;
} ## end sub compare

1;
