package POS::Manager;

use strict;
use warnings;

# XML-RPC client for SM comm
use RPC::XML::Client;
# for various data struct as boolean
use RPC::XML;

use POS::Config;
use POS::Syslog;
use POS::LDAP::Cache;
use POS::LDAP;

sub new
{
    my $package = shift;
    $package = ref($package) || $package;
    my $self = bless({}, $package);
    my %opts = @_;
    $self->{'CONFIG'} = $opts{'CONFIG'};
    $self->{'CONFIG'} = POS::Config->new() unless (defined $self->{'CONFIG'});
    # determine from where to read data
    my $useBranch;
    if ($self->{'CONFIG'}{'POS_ADMIN_DEPLOYED'} and $self->{'CONFIG'}{'POS_BRANCH_DEPLOYED'}) {
        #combo, by detault admin, use FORCE_BRANCH to use Branch data
        if (defined $opts{'FORCE_BRANCH'}) {
            $useBranch = 1;
        }
        else {
            $useBranch = 0;
        }
    }
    elsif ($self->{'CONFIG'}{'POS_BRANCH_DEPLOYED'}) {
        $useBranch = 1;
    }
    elsif ($self->{'CONFIG'}{'POS_ADMIN_DEPLOYED'}) {
        $useBranch = 0;
    }

    # determine whether SM is enabled, read SM host
    $self->{'CACHE'} = $opts{'CACHE'};
    my $standardsDN = 'cn=standards,' . $self->{'CONFIG'}{'ALL_LDAPBASE'};
    $self->{'CACHE'} = POS::LDAP::Cache->new(CONFIG => $self->{'CONFIG'}, BASE => $standardsDN)
      unless defined $self->{'CACHE'};
    my $entry = $self->{'CACHE'}->getLDAPentry($standardsDN);
    unless ($entry) {
        log_warn_print('scHardware object not found, check LDAP structure. SUSE Manager support disabled.');
        return $self;
    }
    my $host = $entry->get_value('scSUSEManager');
    # connect to SM, obtain sessionKey
    unless ($host) {
        log_debug_print('SUSE Manager address not defined, support disabled');
        return $self;
    }

    my ($user, $pass);
    if (not $useBranch) {
        #on AS use only CONFIG
        $user = $self->{'CONFIG'}{'POS_SM_USER'};
        $pass = $self->{'CONFIG'}{'POS_SM_PASSWORD'};
        log_debug_print('using config from AdminServer');
    }
    elsif ($useBranch) {
        #on BS use only LDAP
        my $entry = $self->{'CACHE'}{'LDAP'}->search(
            base   => $self->{'CONFIG'}{'BRANCH_LDAPBASE'},
            scope  => 'base',
            attrs  => ['userPassword'],
            filter => '(objectClass=scLocation)'
        );
        if ($entry->is_error()) {
            log_crit_print('Cannot obtain branch server password!');
            return;
        }
        $entry = $entry->pop_entry();
        $user  = $self->{'CONFIG'}{'BRANCH_LDAPBASE'};
        $user =~ s/cn=//;
        $user =~ s/,cn=|,ou=|,o=|,c=/./g;
        $pass = $entry->get_value('userPassword');
        log_debug_print('using config from BranchServer LDAP');
    }
    else {
        log_crit_print('Unknown SLEPOS configuration');
        return;
    }

    unless (defined $user) {
        log_error_print('SUSE Manager account not defined, unable to authenticate! Support disabled');
        return $self;
    }
    unless (defined $pass) {
        log_error_print('SUSE Manager account password not defined, unable to authenticate! Support disabled');
        return $self;
    }
    # this never fail, no connection is made
    $self->{'SM'} = new RPC::XML::Client('http://' . $host . '/rpc/api');
    # initial connection
    $self->{'SESSION'} = $self->{'SM'}->send_request('auth.login', $user, $pass);
    unless (ref($self->{'SESSION'})) {
        log_error_printf($self->{'SESSION'});
        delete $self->{'SM'};
        # return undef so we will try again later
        return;
    }
    #store Manager address, needed for CRs registration
    $self->{'SMHOST'} = $host;
    return $self;
} ## end sub new

sub call
{
    my ($self, $action, @options) = @_;
    return unless ($self->{'SM'});
    return $self->{'SM'}->simple_request($action, $self->{'SESSION'}, @options);
} ## end sub call

sub logout
{
    my ($self) = @_;
    return unless (defined $self->{'SM'});
    $self->call('auth.logout');
} ## end sub logout

sub getRemoteAddress
{
    my ($self) = @_;
    return $self->{'SMHOST'};
} ## end sub getRemoteAddress

sub bool
{
    my ($self, $value) = @_;
    return RPC::XML::boolean->new($value);
} ## end sub bool

sub DESTROY
{
    my ($self) = @_;
    $self->logout();
} ## end sub DESTROY

1;
