#!/usr/bin/perl -w
## @file
# Handle downloading of images from remote Image Server or Studio

## @class
package POS::RemoteImages;

use strict;

use File::Path qw(rmtree);
use File::Temp qw(tempdir);
use File::Find;
use Cwd qw(abs_path);

#use LWP::Simple qw(getstore getprint);
use POS::Image;
use POS::Syslog;
#use POS::LdapUtil;

# version of the package
our $VERSION = 0.1;


## @cmethod object new()
# @return A new object
sub new
{
    my $package = shift;
    my $self    = {};

    %$self = (%$self, @_);

    return if !defined $self->{'STATIC'};

    $self->{'TMPDIR'} = $self->{'STATIC'}{'SYNC_DIR'};

    return bless($self);
} ## end sub new

sub is_curl
{
    my ($src) = @_;
    return 1 if $src =~ /^http:\/\//;
    return 1 if $src =~ /^ftp:\/\//;
    return 0;
} ## end sub is_curl

sub is_ssh
{
    my ($src) = @_;
    return 0 if is_curl($src);
    $src =~ s/scp:\/*//;
    return 1 if $src =~ /^[^\/]+:/;
    return 0;
} ## end sub is_ssh

sub is_remote
{
    my ($src) = @_;
    return is_curl($src) || is_ssh($src);
} ## end sub is_remote

sub create_temp
{
    my ($self, $prefix, $dir) = @_;
    $dir = "/tmp" unless defined $dir;
    my $temp = tempdir("${prefix}_XXXXXX", DIR => $dir);
    push @{ $self->{'TEMPDIRS'} }, $temp;
    return $temp;
} ## end sub create_temp

sub delete_temps
{
    my ($self) = @_;
    log_debug_print("deleting temp dirs");
    foreach my $d (@{ $self->{'TEMPDIRS'} }) {
        rmtree($d);
    }
    $self->{'TEMPDIRS'} = undef;
} ## end sub delete_temps

sub get_untar_pipe
{
    my ($name) = @_;
    my $pipe = '';
    if ($name =~ /\.tar\.gz$/) {
        $pipe .= "zcat |";
        $name =~ s/\.gz$//;
    }
    if ($name =~ /\.tar\.bz2$/) {
        $pipe .= "bzcat |";
        $name =~ s/\.bz2$//;
    }
    if ($name =~ /\.tgz$/) {
        $pipe .= "zcat |";
        $name =~ s/\.tgz$/.tar/;
    }
    if ($name =~ /\.tar$/) {
        $pipe .= "tar x";
    }
    return $pipe;
} ## end sub get_untar_pipe

sub extract_local_tarball
{
    my ($self, $src, $target_dir) = @_;
    $src = abs_path($src);
    my $pipe = get_untar_pipe($src);
    return if (!$pipe);

    log_debug_print("cd '$target_dir'; ( $pipe ) < '$src'");
    return system("cd '$target_dir'; ( $pipe ) < '$src'") == 0;
} ## end sub extract_local_tarball

sub download_http_tarball
{
    my ($self, $src, $target_dir) = @_;
    my $pipe = get_untar_pipe($src);
    return if (!$pipe);

    #    open(OUTPUT, "| (cd '$target_dir'; $pipe)");
    #    my $oldfh = select(OUTPUT);
    #    getprint($src);
    #    select($oldfh);
    #    close OUTPUT;
    log_debug_print("cd '$target_dir'; curl -L '$src' | $pipe");
    return system("cd '$target_dir'; curl -L '$src' | $pipe") == 0;
} ## end sub download_http_tarball

sub download_http_file
{
    my ($self, $src, $target_dir) = @_;

    my $name = $src;
    $name =~ s/.*\///;
    log_debug_print("curl -L '$src' > '$target_dir/$name'");
    return system("curl -L '$src' > '$target_dir/$name'") == 0;
} ## end sub download_http_file

sub prepare_shared_ssh_connection
{
    my ($self, $host) = @_;

    return unless defined $host;

    if (!$self->{'SSH_TEMP_DIR'}) {
        $self->{'SSH_TEMP_DIR'} = create_temp($self, 'pos_ssh');
    }

    $self->{'SSH_MASTERS'} = {} unless defined $self->{'SSH_MASTERS'};
    my $ctl_path = $self->{'SSH_MASTERS'}{$host};

    return $ctl_path if ($ctl_path && -S $ctl_path);

    $ctl_path = $self->{'SSH_TEMP_DIR'} . "/$host";

    log_debug_print("ssh -M -N -S '$ctl_path' -f '$host'");
    system("ssh -M -N -S '$ctl_path' -f '$host'");
    $self->{'SSH_MASTERS'}{$host} = $ctl_path;

    for (my $i = 0; $i < 10; $i++) {
        return $ctl_path if (-S $ctl_path);
        sleep(1);
    }
    log_error_print("Can't get ssh controll path");
    return;
} ## end sub prepare_shared_ssh_connection

sub close_ssh_connections
{
    my ($self) = @_;
    log_debug_print("closing ssh connections");
    foreach my $host (keys %{ $self->{'SSH_MASTERS'} }) {
        my $ctl_path = $self->{'SSH_MASTERS'}{$host};
        log_debug_print("ssh -q -S '$ctl_path' -O exit '$host' 2>/dev/null");
        system("ssh -q -S '$ctl_path' -O exit '$host' 2>/dev/null");
    }
    $self->{'SSH_MASTERS'} = undef;
} ## end sub close_ssh_connections

sub get_ssh_host
{
    my ($src) = @_;
    return unless is_ssh($src);
    $src =~ s/^scp:\/*//;
    $src =~ s/:.*$//;
    return $src;
} ## end sub get_ssh_host

sub download_ssh_tarball
{
    my ($self, $src, $target_dir) = @_;

    my $pipe = get_untar_pipe($src);
    return if (!$pipe);

    my $host = get_ssh_host($src);
    my $ctl_path = prepare_shared_ssh_connection($self, $host);
    return unless $ctl_path;
    $src =~ s/^scp:\/*//;
    $src =~ s/^[^:]*://;

    log_debug_print("cd '$target_dir'; ssh -S '$ctl_path' '$host' cat '$src' | $pipe");
    return system("cd '$target_dir'; ssh -S '$ctl_path' '$host' cat '$src' | $pipe") == 0;
} ## end sub download_ssh_tarball

sub download_ssh_dir_or_file
{
    my ($self, $src, $target_dir) = @_;

    my $name = $src;
    $name =~ s/.*\///;
    $name =~ s/\.gz$//;

    my $host = get_ssh_host($src);
    my $ctl_path = prepare_shared_ssh_connection($self, $host);
    return unless $ctl_path;

    $src =~ s/^scp:\/*//;

    my $src_path = $src;
    $src_path =~ s/[^\/]*$//;

    my @command = ('rsync', '-axzvr', '-e', "ssh -S '$ctl_path'");
    push @command, "--include=$name", "--include=$name.gz", "--include=$name.md5", "--include=$name/*" if ($name ne '');
    # for the symlink '*.kernel' add the target '*.kernel.version', bnc#835428
    push @command, "--include=$name.*" if ($name =~ /\.kernel$/);
    push @command, '--exclude=*' if ($name ne '');
    push @command, $src_path, $target_dir;
    log_debug_print join(' ', @command);
    return system(@command) == 0;
} ## end sub download_ssh_dir_or_file

sub download_remote
{
    my ($self, $src, $target_dir) = @_;

    if (is_ssh($src)) {
        return 1 if download_ssh_tarball($self, $src, $target_dir);
        return download_ssh_dir_or_file($self, $src, $target_dir);
    }
    elsif (is_curl($src)) {
        return 1 if download_http_tarball($self, $src, $target_dir);
        my $ret = download_http_file($self, $src, $target_dir);
        $src =~ s/\.gz$//;
        $src .= ".md5";
        download_http_file($self, $src, $target_dir) if $ret;
        return $ret;
    }
} ## end sub download_remote

sub download_remote_file
{
    my ($self, $src, $target_dir) = @_;
    if (is_ssh($src)) {
        return unless download_ssh_dir_or_file($self, $src, $target_dir);
    }
    elsif (is_curl($src)) {
        return unless download_http_file($self, $src, $target_dir);
    }
    my $local = $src;
    $local =~ s/^.*\///;
    $local = "$target_dir/$local";
    return $local if -f $local;
    return;
} ## end sub download_remote_file

sub find_image
{
    my ($dir) = @_;
    my $res_image;
    my $image = POS::Image->new();
    find(
        sub {
            $image->set($File::Find::name);
            if ($image->valid() && 
                $image->{'PATH'} !~ /\/initrd-[^\/]*$/ &&
                (!defined $res_image || $image->{'PATH'} =~ /.gz$/)) {
                $res_image = $image->{'PATH'};
                }
        },
        abs_path($dir)
    );
    log_info_print("Using image: $res_image") if defined $res_image;
    return $res_image;
} ## end sub find_image

sub find_kernel
{
    my ($dir) = @_;
    my $res_kernel;
    my $res_initrd;
    find(
        sub {
            if ($File::Find::name =~ /\/initrd-[^\/]*\.kernel[^\/]*$/) {
                if ($File::Find::name !~ /\.md5$/ && (!defined $res_kernel || $File::Find::name =~ /kernel\.[^\/]*$/))
                {    # prefer longer name
                    $res_kernel = $File::Find::name;
                }
            }
            elsif ($File::Find::name =~ /\/initrd-[^\/]*\.splash\.gz$/) {
                $res_initrd = $File::Find::name;
            }
            elsif ($File::Find::name =~ /\/initrd-[^\/]*\.gz$/ && !$res_initrd) {
                $res_initrd = $File::Find::name;
            }
        },
        abs_path($dir)
    );
    return unless $res_kernel && $res_initrd;
    log_info_print("Using kernel: $res_kernel");
    log_info_print("Using initrd: $res_initrd");
    return ($res_kernel, $res_initrd);
} ## end sub find_kernel

sub download
{
    my ($self, $include_boot, $kernel_ref, $initrd_ref, $images_ref) = @_;

    if (defined ${$kernel_ref}) {
        my $tmp_dir;

        if (-e ${$kernel_ref}) {
            if (! -f ${$kernel_ref}) {
                log_error_print("Kernel is not a regular file: ${$kernel_ref}");
                return;
            }
            log_debug_print("found local kernel file: ${$kernel_ref}");
            ${$kernel_ref} = abs_path(${$kernel_ref});
        }
        elsif (is_remote(${$kernel_ref})) {
            $tmp_dir = create_temp($self, ".pos_kernel", $self->{'TMPDIR'});
            my $local_kernel = download_remote_file($self, ${$kernel_ref}, $tmp_dir);
            download_remote_file($self, ${$kernel_ref} . ".md5", $tmp_dir);
            if (!$local_kernel) {
                log_error_print("Download failed: ${$kernel_ref}");
                return;
            }
            log_debug_print("remote kernel downloaded to: $local_kernel");
            ${$kernel_ref} = $local_kernel;
        }
        else {
            log_error_print("File/URL not found: ${$kernel_ref}");
            return;
        }

        if (!defined ${$initrd_ref}) {
            return;
        }

        if (-e ${$initrd_ref}) {
            if (! -f ${$initrd_ref}) {
                log_error_print("Initrd is not a regular file: ${$initrd_ref}");
                return;
            }
            log_debug_print("found local initrd file: ${$initrd_ref}");
            ${$initrd_ref} = abs_path(${$initrd_ref});
        }
        elsif (is_remote(${$initrd_ref})) {
            $tmp_dir = create_temp($self, ".pos_kernel", $self->{'TMPDIR'}) unless defined $tmp_dir;
            my $local_initrd = download_remote_file($self, ${$initrd_ref}, $tmp_dir);
            my $initrd_md5 = ${$initrd_ref};
            $initrd_md5 =~ s/\.gz$//;
            $initrd_md5 .= ".md5";
            download_remote_file($self, $initrd_md5, $tmp_dir);

            if (!$local_initrd) {
                log_error_print("Download failed: ${$initrd_ref}");
                return;
            }
            log_debug_print("remote initrd downloaded to: $local_initrd");
            ${$initrd_ref} = $local_initrd;
        }
        else {
            log_error_print("File/URL not found: ${$initrd_ref}");
            return;
        }
    }

    my $first = 1;

    foreach my $image (@{$images_ref}) {
        if (-f $image) {
            if ($image =~ /\.tar\.gz$/ || $image =~ /\.tar\.bz2$/) {
                my $tmp_dir = create_temp($self, ".pos_image", $self->{'TMPDIR'});
                if (!extract_local_tarball($self, $image, $tmp_dir)) {
                    log_error_print("Extracting failed: $image");
                    return;
                }
                if ($first && $include_boot) {
                    (${$kernel_ref}, ${$initrd_ref}) = find_kernel($tmp_dir);
                }
                $image = find_image($tmp_dir);
                if (! defined($image)) {
                    log_error_print("Can't find a valid image in the tarball");
                    return;
                }
            }
            else {
                $image = abs_path($image);
                if ($first && $include_boot) {
                    log_error_print("Can't extract kernel and initrd from: $image");
                    return;
                }
            }
        }
        elsif (-d $image) {
            my $dir = $image;
            if ($first && $include_boot) {
                (${$kernel_ref}, ${$initrd_ref}) = find_kernel($dir);
                if (! defined(${$kernel_ref})) {
                    log_error_print("Can't find a valid boot image (kernel/initrd) in $dir");
                    return;
                }
            }

            $image = find_image($dir);
            if (! defined($image)) {
                log_error_print("Can't find a valid image in $dir");
                return;
            }
        }
        elsif (is_remote($image)) {
            my $tmp_dir = create_temp($self, ".pos_image", $self->{'TMPDIR'});
            if (!download_remote($self, $image, $tmp_dir)) {
                log_error_print("Download failed: $image");
                return;
            }
            if ($first && $include_boot) {
                (${$kernel_ref}, ${$initrd_ref}) = find_kernel($tmp_dir);
                if (! defined(${$kernel_ref})) {
                    log_error_print("Can't find a valid boot image (kernel/initrd) in $tmp_dir");
                    return;
                }
            }

            $image = find_image($tmp_dir);
            if (! defined($image)) {
                log_error_print("Can't find a valid image in $tmp_dir");
                return;
            }
        }
        else {
            log_error_print("Can't determine file location: $image");
            return;
        }
        $first = 0;

    }
    return 1;
} ## end sub download

sub DESTROY
{
    local $?;
    my $self = shift;
    close_ssh_connections($self);
    delete_temps($self);
} ## end sub DESTROY

1;
