#!/bin/bash

#==========================================================================
# FILE        : nlpos9_backup_data.sh
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : (c) 2009 Novell inc, USA. All rights reserved
#
# Author      : Petr Ostadal <postadal@suse.cz>
#
# DESCRIPTION : Backuping nlpos9 configurations, images for slepos11 migration.
#
#   backup data structure:
#     NLPOS9-migration/rpm.list         - `rpm -qa`
#     NLPOS9-migration/RPM/POS_*        - installed POS_* packages
#     NLPOS9-migration/nlpos9-ldap.ldif - `slapcat`
#     etc/SuSE-release
#     etc/opt/SLES/POS
#     opt/SLES/POS
#     tftpboot
#
#==========================================================================

appName=`basename ${0}`
appVersion="0.1"

SLAPCAT_BIN="/usr/sbin/slapcat"
SLAPTEST_BIN="/usr/sbin/slaptest"
RCLDAP_BIN="/usr/sbin/rcldap"
SUSERELEASE="etc/SuSE-release"
NLPOSMIGRATIONDIR="NLPOS9-migration"

# Usage and Help text

usage () {
  cat << EOF_USAGE
${appName} v${appVersion}

Usage:    ${appName} [OPTIONS] -f|--file=<FILE.tar.gz|bzip2>
          ${appName} [OPTIONS] -d|--directory=<DIRECTORY>
          ${appName} -h|--help

Helper migration script for backuping NLPOS9 configurations, ldap DB and binary
images.
EOF_USAGE
}

helptext () {
  usage
  cat << EOF_HELP

Options:
 -f, --file=<FILE.tar.gz|bzip2>    save migration data to file
                                   (supported .tar, .tgz, .tar.gz, .tar.bz2)
 -d, --directory=<DIRECTORY>       save migration data to existing directory
 --force                           force product version checking
 -q, --quiet                       provide less detailed output
 -v, --verbose                     provide more detailed output
 -V, --version                     print the version of script being used
 -h, --help                           show this help message
 --usage                           display brief usage message
EOF_HELP
}

# migration file list
migration_files() {
  cat << EOF_MIGRATION_FILES
etc/SuSE-release
etc/opt/SLES/POS
opt/SLES/POS/image
opt/SLES/POS/rsync
opt/SLES/POS/system
tftpboot
EOF_MIGRATION_FILES
}

# migration file list 
existing_migration_files() {
  migration_files | while read f; do
    # all files are relative to '/'
    if [ -e "/$f" ]; then
      echo "$f"
    fi
  done
}

# print verbose message
#  params: $1 - message
log() {
  if [ "$verbose" = "1" ]; then
    echo $1
  fi
}

# We need TEMP as the `eval set --' would nuke the return value of getopt.
TEMP=`getopt -o f:d:hqvV -l file:,directory:,help,usage,quiet,verbose,version,force \
  -n "${appName}" -- "$@"`

if [ $? != 0 ] ; then usage 1>&2; exit 1; fi

# Note the quotes around `$TEMP': they are essential!
eval set -- "$TEMP"

force=0
quiet=0
verbose=0
tar_compression=""
tar_verbose=""

while true; do
  case $1 in
    -f|--file)
      BACKUPFILE=$2
      shift 2
      if [ -f "$BACKUPFILE" ]; then
        echo "error: target file '${BACKUPFILE}' exists" >&2
        exit 2
      fi
      case "$BACKUPFILE" in
        *.tgz|*.tar.gz)
          tar_compression="--gzip" ;;
        *.tar.bz2)
          tar_compression="--bzip2" ;;
        *.tar)
          tar_compression="" ;;
        *)
          echo "error: archive file '${BACKUPFILE}' is not supported" >&2
          echo "Use one of: .tar .tgz .tar.gz .tar.bz2"
          usage
          exit 2
          ;;
      esac
      ;;
    -d|--directory)
      BACKUPDIR=$2
      shift 2
      if [ ! -d "$BACKUPDIR" ]; then
        echo "error: target directory '${BACKUPDIR}' does not exist" >&2
        exit 2
      fi
      ;;
    --force) force=1; shift ;;
    -q|--quiet) quiet=1; shift ;;
    -v|--verbose) verbose=1; shift ;;
    -h|--help) helptext;  exit 0 ;;
    --usage) usage; exit 0 ;;
    -V|--version) echo "${appName} ${appVersion}"; exit 0 ;;
    --) shift; break ;;
    *)
      echo "error: internal error!" >&2
      exit 3
      ;;
    esac
done

if [ $# -gt 0 ]; then
  echo "error: odd addition arguments '$*'" >&2
  usage
  exit 1
fi

if ( [ "$BACKUPDIR" ] && [ "$BACKUPFILE" ] ) || 
   ( [ ! "$BACKUPDIR" ] && [ ! "$BACKUPFILE" ] ); then
  echo "error: specify one of target --file or --directory options" >&2
  usage
  exit 1
fi


VERSION=`grep "VERSION" "/${SUSERELEASE}" | awk '{ print $3 }'`
#TODO: enhace nlpos9 ssp3 service pack checking
if [ "$force" = "0" ] && [ "$VERSION" != "9" ]; then
  echo "error: this product version '$VERSION' ('/${SUSERELEASE}) is not supported" >&2
  exit 1
fi

exitcode=0
tmpdir=`mktemp -d /tmp/tmp.nlpos9-migration.XXXXXXXXXX`
tmpmigrationdir="${tmpdir}/${NLPOSMIGRATIONDIR}"

rsync_progress="--progress"

if [ "$quiet" = "1" ]; then
  rsync_progress=""
  rsync_verbose=""
  tar_verbose=""
  verbose=0
fi

if [ "$verbose" = "1" ]; then
  rsync_verbose="-v"
  tar_verbose="-v"
fi

mkdir "$tmpmigrationdir"

# create installed package lists 'rpm.list'
log "creating '${NLPOSMIGRATIONDIR}/rpm.list'"
rpmqpack |sort > "${tmpmigrationdir}/rpm.list"

# create RPM/POS* files if appropriate POS_* package is installed
log "noticing installed POS_ package names in '${NLPOSMIGRATIONDIR}'"
mkdir "${tmpmigrationdir}/RPM"
pushd "${tmpmigrationdir}/RPM" >/dev/null
rpmqpack | grep "^POS_" | xargs touch 2>/dev/null
if [ "$verbose" ]; then ls; fi
popd >/dev/null

# dump Ldap DB
if [ -x "$RCLDAP_BIN" ] && [ -x "$SLAPTEST_BIN" ] && [ -x "$SLAPCAT_BIN" ]; then
  # interrupt/stop LDAP service (save service state)
  ldapstatus=1
  if "$RCLDAP_BIN" status &>/dev/null; then
      ldapstatus=0
      "$RCLDAP_BIN" stop
  fi

  if "$SLAPTEST_BIN" &>/dev/null; then
    log "looking for LDAP database"
    if "$SLAPCAT_BIN" -d 0 -l "${tmpmigrationdir}/nlpos9-ldap.ldif"; then
      log "dumped LDAP DP to '${tmpmigrationdir}/nlpos9-ldap.ldif' file"
    else
      # can't dump LDAP DB
      rm -f "${tmpmigrationdir}/nlpos9-ldap.ldif"
    fi
  fi

  # rerun/start LDAP service
  if [ $ldapstatus -eq 0 ]; then
    "$RCLDAP_BIN" start
  fi
fi

log "backuping NLPOS9 configurations and images to '${BACKUPDIR}${BACKUPFILE}':"

if [ "$BACKUPDIR" ]; then
  existing_migration_files | rsync -arHRc $rsync_verbose $rsync_progress \
    --files-from=- / "${BACKUPDIR}"
  exitcode=$?
  rsync -arHcS $rsync_verbose $rsync_progress "${tmpmigrationdir}" "${BACKUPDIR}"
  exitcode=$((exitcode+$?))

elif [ "$BACKUPFILE" ]; then
  existing_migration_files | tar $tar_compression -C "${tmpdir}" \
    "${NLPOSMIGRATIONDIR}" -C / --files-from=- -cS $tar_verbose -f "${BACKUPFILE}"
  exitcode=$?
else
  echo "error: used --file and --directory in the same time" >&2
  usage
  exitcode=1
fi
log "cleaning up"
# clean up
test -d "$tmpdir" && rm -r "$tmpdir"
log "finished"
exit $exitcode
