#!/bin/bash

#==========================================================================
# FILE        : slepos_migration.sh
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : (c) 2013 SUSE LINUX Products GmbH, Nuernberg, All rights reserved
#               (c) 2009 Novell inc, USA. All rights reserved
#
# Author      : Ondrej Holecek <oholecek@suse.com>
#               Petr Ostadal <postadal@suse.com>
#
# DESCRIPTION : Migrating NLPOS9/SLEPOS10/SLEPOS11 to SLEPOS11 SP3
#
#==========================================================================

appName=$(basename $0)
appVersion="0.1"
appDir=$(dirname $0)

#source common
MIGRATION_COMMON="${appDir}/migration_common.sh"
test -r "$MIGRATION_COMMON" || exit 255
. "$MIGRATION_COMMON"

# Usage and Help text

usage () {
    cat << EOF_USAGE
${appName} v${appVersion}

Usage:    ${appName} [OPTIONS] -f|--file=<FILE.tar.gz|bzip2> -t|--deploy_type=<AS|BS|IS|NUM>
          ${appName} [OPTIONS] -d|--directory=<DIRECTORY> -t|--deploy_type=<AS|BS|IS|NUM>
          ${appName} -h|--help

NLPOS9/SLEPOS10/SLEPOS11 migration tool to SLEPOS11 SP3 version. 
EOF_USAGE
}

displayHelp () {
    usage
    cat << EOF_HELP

Options:
 -f, --file=<FILE.tar.gz|bzip2>    load migration data from file
                                   (supported .tar, .tgz, .tar.gz, .tar.bz2)
 -d, --directory=<DIRECTORY>       load migration data from existing directory
 -o. --old-release=<FILE>          path to stored old SuSE-release file before
                                   migration
 -t, --deploy_type=<AS|BS|IS|NUM>  type of SLEPOS server to migrate or combination (e.g. AS+IS)
                                   NUM can be 1 (AS), 2 (BS), 4 (IS) or combination (e.g. 5 (AS+IS))
 -q, --quiet                       provide less detailed output
 -v, --verbose                     provide more detailed output
 -V, --version                     print the version of script being used
 -h, --help                        show this help message
 --usage                           display brief usage message
EOF_HELP
}
# We need TEMP as the `eval set --' would nuke the return value of getopt.
TEMP=`getopt -o f:d:o:t:l:hqvV -l file:,directory:,old-release:,deploy-type:,log:,help,usage,quiet,verbose,version \
    -n "${appName}" -- "$@"`

if [ $? != 0 ] ; then usage 1>&2; exit 1; fi

# Note the quotes around `$TEMP': they are essential!
eval set -- "$TEMP"

verbose=0
quiet=0
tar_compression=""
tar_verbose=""
# default deploy type is no migration at all
DEPLOY_TYPE=0
readonly DEPLOY_TYPE_AS=1
readonly DEPLOY_TYPE_BS=2
readonly DEPLOY_TYPE_IS=4

while true; do
    case $1 in
      -f|--file)
          BACKUPFILE=$2
          shift 2
          if [ ! -f "$BACKUPFILE" ]; then
              err "Backup file '${BACKUPFILE}' not found!  (${appName})"
          fi
          ;;
     -d|--directory)
          BACKUPDIR=$2
          shift 2
          if [ ! -d "$BACKUPDIR" ]; then
              err "Backup directory '${BACKUPDIR}' not found!  (${appName})"
          fi
          ;;
      -o|--old-release)
          log "Loading old SuSE-release" $LOG_VERBOSE
          OLDRELEASE=$2
          shift 2
          if [ ! -f "$OLDRELEASE" ]; then
              err "Old SuSE-release file '${OLDRELEASE}' not found!  (${appName})"
          fi
          ;;
      -t|--deploy-type)
          DEPLOY_TYPE=`sed -n -e 's/ *//g;s/,/+/g'                        \
                              -e '/^\([ABI]S\)\(+[ABI]S\)*$\|^[1-7]$/!q'  \
                              -e 's/AS/1/;s/BS/2/;s/IS/4/;s/[ABI]S/0/g;p' \
                           <<< "$2" | bc`
          shift 2
          ;;
      -l|--log)
          setLogLevel $LOG_VERBOSE
          LOGFILENAME=$2
          SetFileOutput 1
          shift 2
          ;;
      -v|--verbose)
          setLogLevel $LOG_VERBOSE
          shift
          ;;
      -q|--quiet)
          setLogLevel $LOG_OFF
          shift
          ;;
      -h|--help)
          displayHelp; exit 0 ;;
      --usage) usage; exit 0 ;;
      -V|--version) echo "${appName} ${appVersion}"; exit 0 ;;
      --) shift; break ;;
      *)
          err "Internal error!  (${appName})"
          exit 254
          ;;
    esac
done

if [ $# -gt 0 ]; then
    err "Odd addition arguments '$*'\n"
fi

###############################################################################
# other migration code here

function cleanup
{
    rm -rf "$tmpdir"
}

tmpdir=`mktemp --tmpdir -d tmp.nlpos9-migration.XXXXXXXXXX`

#set value if is not set
if [ ! "$OLDRELEASE" ] && GetOldSuseRelease; then
    OLDRELEASE="$OLDSUSERELEASE"
fi

SUSEVERSION=$(GetSuseVersion "$OLDRELEASE")

# total number of progress bars
bars=0

case "${SUSEVERSION}" in
    11)
        OLDVER="v11"
        if (($DEPLOY_TYPE & $DEPLOY_TYPE_AS )); then
            ((bars=$bars + 3))
        fi
        if (($DEPLOY_TYPE & $DEPLOY_TYPE_BS )); then
            ((bars=$bars + 1))
        fi
        ;;
    10)
        OLDVER="v10"
        if (($DEPLOY_TYPE & $DEPLOY_TYPE_AS )); then
            ((bars=$bars + 3))
        fi
        if (($DEPLOY_TYPE & $DEPLOY_TYPE_BS )); then
            ((bars=$bars + 1))
        fi
        ;;
    9)
        OLDVER="v9"
        if (($DEPLOY_TYPE & $DEPLOY_TYPE_AS )); then
            ((bars=$bars + 2 + 3))
        fi
        if (($DEPLOY_TYPE & $DEPLOY_TYPE_BS )); then
            ((bars=$bars + 1 + 1))
        fi
        if (($DEPLOY_TYPE & $DEPLOY_TYPE_IS )); then
            ((bars=$bars + 4))
        fi
        if ( [ "$BACKUPDIR" ] && [ "$BACKUPFILE" ] ) ||
           ( [ ! "$BACKUPDIR" ] && [ ! "$BACKUPFILE" ] ); then
            err "Specify one of --file or --directory options" >&2
            usage
            exit 1
        fi
        ;;
    *)
        err "Unknown/unsupported previous SuSE product release version '${SUSEVERSION}'"
        ;;
esac

if [ $bars -gt 0 ]; then
    # start progress
    ProgressTitles $bars
fi

###############################################################################
# configuration migration starts here
migration_perfomed=0
# admin server
if (($DEPLOY_TYPE & $DEPLOY_TYPE_AS )); then
    log "Updating admin server configuration" $LOG_VERBOSE
    GetOldConfFile $DEPLOY_TYPE_AS
    if [ ! "$CONFFILE" ]; then
        err "old admin server configuration file is missing"
    fi

    # extract config file and certificates from backup file/dir
    GetLDAPfile
    if [ ! "$LDAPFILE" ]; then
        err "old LDAP backup is missing"
    fi

    # keys/certs
    if [ "$SUSEVERSION" = "9" ]; then
        dir="/etc/SLEPOS/keys"
        ProgressTitle "Migrating NLPOS9 keys to '${dir}'"
        test -d "$dir" || mkdir -p "$dir"
        CopyFileFromBackup opt/SLES/POS/keys/ "$dir" 1

        dir="/srv/SLEPOS/certs"
        ProgressTitle "Migrating NLPOS9 certs to '${dir}'"
        test -d "$dir" || mkdir -p "$dir"
        CopyFileFromBackup opt/SLES/POS/rsync/certs/ "$dir" 1
    fi

    # prepare ldap db if importing ldif
    ProgressTitle "Migrating LDAP DB"
    if grep ".ldif" <<< "$LDAPFILE"; then
        if ! "${appDir}/posLdapMigration.pl" --config-file="$CONFFILE" $OLDVER "$LDAPFILE" "${tmpdir}/migrateLDAP.ldif"; then
            err "posLdapMigration failed on ${LDAPFILE} file."
        fi
        LDAPFILE="${tmpdir}/migrateLDAP.ldif"
    fi
    ProgressTitle "Migrating Admin Server"
    # call posInitAdminserver.sh --noninteractive --config-file=<old_config_file> --ldap-backup=<ldif_of_xml_with_ldap_data>
    if ! posInitAdminserver.sh --noninteractive --reinitialize="$CONFFILE" --ldap-backup="$LDAPFILE"; then
        err "AdminServer migration failed on reinitialization"
    fi
    ProgressTitle "Refreshing LDAP DB"
    posAdmin.pl --refresh

    if [ "$SUSEVERSION" = "9" ]; then
        # migrate admind conf
        log "Migrating admind.conf file" $LOG_VERBOSE
        CopyFileFromBackup etc/opt/SLES/POS/admind.conf "/etc/SLEPOS" 0
    fi
    migration_perfomed=1
fi

# branch server
if (( $DEPLOY_TYPE & $DEPLOY_TYPE_BS )); then
    ProgressTitle "Migrating Branch Server configurations"
    GetOldConfFile $DEPLOY_TYPE_BS
    if [ ! "$CONFFILE" ]; then
        err "old branch server configuration file is missing"
    fi
    log "Updating branch server configuration" $LOG_VERBOSE
    # extract config gile and certificates from backup file/dir
    # call posInitBranchserver.sh --noninteractive --config-file=<old_config_file>
    posInitBranchserver.sh --noninteractive --reinitialize="$CONFFILE"

    if [ "$SUSEVERSION" = "9" ]; then
        # migration NLPOS9 tftp dir
        dir="/srv/tftpboot"
        ProgressTitle "Migrating NLPOS9 tftp dir to '${dir}'"
        test -d "$dir" || mkdir -p "$dir"
        CopyFileFromBackup tftpboot/ "$dir" 1
    fi
    migration_perfomed=1
fi

# image server
if (( $DEPLOY_TYPE & $DEPLOY_TYPE_IS )); then
    log "Updating image server configuration" $LOG_VERBOSE
    if [ "$SUSEVERSION" = "9" ]; then
        # migration image hw configurations
        dir="/srv/SLEPOS/config"
        test -d "$dir" || mkdir -p "$dir"
        CopyFileFromBackup opt/SLES/POS/config/ "$dir" 0

        # migration xsrc image configurations
        dir="/var/lib/SLEPOS/system-nlpos9/"
        ProgressTitle "Migrating xscr NLPOS9 image configurations to '${dir}'"
        test -d "$dir" || mkdir -p "$dir"
        CopyFileFromBackup opt/SLES/POS/system/ "$dir" 1

        # migration binary images
        dir="/var/lib/SLEPOS/system-nlpos9/image"
        ProgressTitle "Migrating NLPOS9 binary images to '${dir}'"
        test -d "$dir" || mkdir -p "$dir"
        CopyFileFromBackup opt/SLES/POS/image/ "$dir" 1

        # migration rsync/image
        dir="/srv/SLEPOS/image"
        ProgressTitle "Migrating NLPOS9 rsync images to '${dir}'"
        test -d "$dir" || mkdir -p "$dir"
        CopyFileFromBackup opt/SLES/POS/rsync/image/ "$dir" 1

        # migration rsync/boot
        dir="/srv/SLEPOS/boot"
        ProgressTitle "Migrating NLPOS9 rsync boot dir to '${dir}'"
        test -d "$dir" || mkdir -p "$dir"
        CopyFileFromBackup opt/SLES/POS/rsync/boot/ "$dir" 1
    elif [ "$SUSEVERSION" = "11" ]; then
       # migration image hw configurations
        dir="srv/SLEPOS/config"
        test -d "/$dir" || mkdir -p "/$dir"
        CopyFileFromBackup "$dir" "/$dir" 0

        # migration xsrc image configurations
        dir="var/lib/SLEPOS/"
        ProgressTitle "Migrating/restoring image configurations to '/${dir}'"
        test -d "/$dir" || mkdir -p "/$dir"
        CopyFileFromBackup "$dir" "/$dir" 1

        # migration binary images
        dir="var/lib/SLEPOS/image"
        ProgressTitle "Migrating/restoring binary images to '/${dir}'"
        test -d "/$dir" || mkdir -p "/$dir"
        CopyFileFromBackup "$dir" "/$dir" 1

        # migration rsync/image
        dir="srv/SLEPOS/image"
        ProgressTitle "Migrating/restoring rsync images to '/${dir}'"
        test -d "/$dir" || mkdir -p "/$dir"
        CopyFileFromBackup "$dir" "/$dir" 1

        # migration rsync/boot
        dir="srv/SLEPOS/boot"
        ProgressTitle "Migrating/restoring rsync boot dir to '/${dir}'"
        test -d "/$dir" || mkdir -p "/$dir"
        CopyFileFromBackup "$dir" "/$dir" 1
    fi
    migration_perfomed=1
fi

if [ $migration_perfomed -lt 1 ]; then
    err "unknown deploy type or deploy type not specified"
fi
cleanup
exit 0
