#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>

#define ANCHOR_SM "_SM_"
#define ANCHOR_DMI "_DMI_"

#define SMBIOSSearchSegment 0xf000

#define null NULL

char *memdevDefault = "/dev/mem";
char *memdev = null;

char *versionString = "SMBIOS Parser 1.05\nCopyright (C) IBM 2003";

char *programName = null;
#define nerr( aerrno ) fprintf( stderr, "%s:%d: %s\n", programName, __LINE__, strerror( aerrno ) );
#define nserr( astring, aerrno ) fprintf( stderr, "%s:%d:%s: %s\n", programName, __LINE__, astring, strerror( aerrno ) );
#define serr( astring ) fprintf( stderr, "%s:%d: %s\n", programName, __LINE__, astring );

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned long u32;
typedef unsigned long long u64;

#pragma pack( 1 )
struct SMBIOSEntryPoint {
	u8 anchor[ 4 ];
	u8 checksum;
	u8 length;
	u8 major;
	u8 minor;
	u16 maxStructureSize;
	u8 epRevision;
	u8 area[ 5 ];
	u8 dmiAnchor[ 5 ];
	u8 dmiChecksum;
	u16 structureTableLength;
	u32 structureTableAddress;
	u16 structureCount;
	u8 revision;
};        

struct SMBIOSStructureHeader {
	u8 type;
	u8 length;
	u16 handle;
}; 

struct SMBIOSStructure {
    struct SMBIOSStructureHeader header;
    u8 buffer[ 255 ];
};

struct SMBIOSHandle {
    int file; // stores the handle to the memory devnod
    long entryOffset;
    struct SMBIOSEntryPoint entry; // points to the contents of the SMBIOS entrypoint, after it is found
    u32 currentStructure;
    long currentStructureOffset;
};

// SMBIOS Standard structure types

struct SMBIOSStructure_BIOS { // Type 0
    struct SMBIOSStructureHeader header;
    u8 vendorString;
    u8 versionString;
    u16 biosSegment;
    u8 releaseDateString;
    u8 romSize;
    u64 biosCharacteristics;
};

struct SMBIOSStructure_System { // Type 1
    struct SMBIOSStructureHeader header;
    u8 manufacturerString;
    u8 productNameString;
    u8 versionString;
    u8 serialNumberString;
    u8 uuid[ 16 ];
    u8 wakeupType;
};

struct SMBIOSStructure_BaseBoard { // Type 2
    struct SMBIOSStructureHeader header;
    u8 manufacturerString;
    u8 productNameString;
    u8 versionString;
    u8 serialNumberString;
};

#pragma pack()


enum SMBErrors{
    MEMACCESS=1, /* Could not obtain access to memory */
    SMBIOSNOTOPEN=2, /* SMBIOSHandle passed does not has not be used with openSMBIOS */
    NOTFOUND=3, /* Could not find SMBIOS Entrypoint */
    INDEXOUTOFRANGE=4, /* Attempt to access non-existent SMBIOS structure */
    BUFFERTOOSMALL=5, /* Not enough buffer space to store SMBIOS structure string */
    MEMSEEK=6, /* Seek in memory dev node failed */
    READERROR=7 /* when calling read() failed */
};

void closeSMBIOS( struct SMBIOSHandle *handle ) {
    if ( handle->file != -1 ) {
        close( handle->file );
        handle->file = -1;
    }
}

/** 
 * Initializes a connection to the SMBIOS using the memory dev node filename provided
 * @param memdev The filename of the memory dev node (e.g. "/dev/mem")
 * @handle An SMBIOSHandle structure that will be used with other functions to retreive information from the SMBIOS
 * @return Returns 0 in the case of success
 *          Returns an SMBErrors element in the case of failure
 */
int openSMBIOS( const char *memdev, struct SMBIOSHandle *handle ) {
    u16 searches;
    int found = 0;            
    char check[ 0x10 ];
    
    if ( ( handle->file = open( memdev, O_RDONLY ) ) == -1 ) {
        nserr( memdev, errno );
        return MEMACCESS;
    }
   

    // Search from 0xf000:00000 to 0xf000:fff0 along 0x10 byte boundary for SMBIOS anchor
    if ( lseek( handle->file, (SMBIOSSearchSegment << 4) , SEEK_SET ) == -1 ) {
        nserr( "could not perform seek in memory", errno );
        return MEMACCESS;
    }
    
    for ( searches = 0; (searches < 0xfff) && (found == 0); searches++ ) {
        
        if ( read( handle->file, check, 0x10 ) == 0x10 && memcmp( check, ANCHOR_SM, 4 ) == 0 ) {  // if a possible anchor is found
            
            handle->entryOffset = lseek( handle->file, 0, SEEK_CUR ) - 0x10; // store start of matching anchor
            lseek( handle->file, handle->entryOffset, SEEK_SET ); // position at start of matching anchor

            if ( read( handle->file, &( handle->entry ), sizeof( struct SMBIOSEntryPoint ) ) == sizeof( struct SMBIOSEntryPoint ) && 
                 memcmp( handle->entry.dmiAnchor, ANCHOR_DMI, 5 ) == 0  ) { // if the DMI anchor matches (sanity check before assuming this is the SMBIOS entry point)
                found = 1;
            } else { 
                // This is not a valid entry point. Reset the file to the position of the last anchor + 0x10
                lseek( handle->file, handle->entryOffset + 0x10, SEEK_SET ); // position at start of matching anchor
            }
        }
    }


    if ( found ) {
        handle->currentStructure = 0;
        handle->currentStructureOffset = handle->entry.structureTableAddress;
        return 0;
    } else {
        serr( "SMBIOS entry point not found in memory" );
        closeSMBIOS( handle );
        return NOTFOUND;
    }
}

/**
 * Locates a given SMBIOS Structure in memory using the SMBIOSHandle provided
 * @param handle A SMBIOSHandle created from openSMBIOS
 * @param index The index of the SMBIOS structure to locate (0 based)
 * @return Returns 0 in the case of success. 
 *          Returns INDEXOUTOFRANGE if the index does not exist
 *          Returns SMBIOSNOTOPEN if the SMBIOSHandle has not be opened
 */
int seekSMBIOSStructure( struct SMBIOSHandle *handle, u32 index ) {
    
    struct SMBIOSStructureHeader tmpHeader;
    struct SMBIOSStructureHeader *header;
    header = &tmpHeader;

    if ( handle->file == -1 ) {
        return SMBIOSNOTOPEN;
    }

    if ( index >= handle->entry.structureCount ) {
        return INDEXOUTOFRANGE;
    }

    if ( index >= handle->currentStructure ) {
        index -= handle->currentStructure;
    } else {
        handle->currentStructure = 0;
        handle->currentStructureOffset = handle->entry.structureTableAddress;
    }

    while ( index ) {
        char c1 = 1, c2 = 1;
        int expended = 0;

        lseek( handle->file, handle->currentStructureOffset, SEEK_SET ); 
        if ( read( handle->file, header, sizeof( struct SMBIOSStructureHeader ) ) != sizeof( struct SMBIOSStructureHeader ) )
            return READERROR;

        lseek( handle->file, header->length, SEEK_CUR ); // position at start of the strings

        while ( !expended ) {
            // While two zeros are not encountered sequentially, keep reading the strings
            if ( read( handle->file, &c2, 1 ) != 1 )
                return READERROR;
            if ( c2 == 0 && c1 == 0 ) {
                expended = 1;
            } else {
                c1 = c2;
            }
        }

        handle->currentStructure++;
        handle->currentStructureOffset = lseek( handle->file, 0, SEEK_CUR );

        index--;
    }

    return 0;
}

/**
 * @param handle An SMBIOSHandle created with openSMBIOS
 * @param structure A SMBIOSStructure stucture to store the contents of the formatted area of the SMBIOS entyr
 * @param index The index of the structure to read (0 based!)
 * @return Returns 0 in the case of success.
 *          Returns INDEXOUTOFRANGE if the index does not exist
 *          Returns SMBIOSNOTOPEN if the SMBIOSHandle has not be opened
 */
int readSMBIOSStructure( struct SMBIOSHandle *handle, struct SMBIOSStructure *structure, u32 index ) { 
    
    int er;

    if ( ( er = seekSMBIOSStructure( handle, index ) ) != 0 ) { // update currentStructureOffset in handle 
        return er;
    }

    lseek( handle->file, handle->currentStructureOffset, SEEK_SET );
    if ( read( handle->file, &( structure->header ), sizeof( struct SMBIOSStructureHeader ) ) != sizeof( struct SMBIOSStructureHeader ) )
      return READERROR;
    if ( read( handle->file, structure->buffer, structure->header.length - sizeof( struct SMBIOSStructureHeader ) ) != sizeof( struct SMBIOSStructureHeader ) )
      return READERROR;

    return 0;
}

/**
 * @param handle The handle created using openSMBIOS
 * @param structureIndex The index of the structure to access ( 0 based! )
 * @param stringIndex The index of the string to measure ( 1 based! )
 * @return Returns the number of bytes a buffer requires to store a given SMBIOS structure string
 *          Returns 0 if the string cannot be accessed
 * WARNING: No check is made to ensure that the string referenced is a valid element of the structure.
 *          Ensure that the string exists in the structure type before attempting to measure it
 */ 
u16 getSMBIOSStructureStringLength( struct SMBIOSHandle *handle, u32 structureIndex, u32 stringIndex ) {
    struct SMBIOSStructure tmpStructure;
    struct SMBIOSStructure *structure = &tmpStructure;
    char c1;
    int expended;
    int er;
    u16 acc = 0;

    if ( stringIndex == 0 ) {
        return 0;
    }

    stringIndex--;

    if ( ( er = seekSMBIOSStructure( handle, structureIndex ) ) != 0 ) { // update currentStructureOffset in handle 
        return 0;
    }

    lseek( handle->file, handle->currentStructureOffset, SEEK_SET );
    
    if ( read( handle->file, &( structure->header ), sizeof( struct SMBIOSStructureHeader ) ) != sizeof( struct SMBIOSStructureHeader ) ||
         read( handle->file, structure->buffer, structure->header.length - sizeof( struct SMBIOSStructureHeader ) ) != sizeof( struct SMBIOSStructureHeader ) )
	return 0;

    // file position now at the start of the strings

    while ( stringIndex ) {
        expended = 0;
        while ( !expended ) {
            if ( read( handle->file, &c1, 1 ) != 1 )
                return 0;
            if ( c1 == 0 ) {
                expended = 1;
            }
        }
        stringIndex--;
    }

    c1 = 1;
    while ( c1 && read( handle->file, &c1, 1 ) == 1 ) {
        acc++;
    }

    return acc;
}

/**
 * @param handle The handle created using openSMBIOS
 * @param structureIndex The index of the structure to access ( 0 based! )
 * @param stringIndex The index of the string to read ( 1 based! )
 * @param buffer The buffer to fill with the contents of the string
 * @param bufferSize The number of bytes available in the buffer
 * @return Returns the number of bytes a buffer requires to store a given SMBIOS structure string
 *          Returns BUFFERTOOSMALL if the operation would result in buffer overrun
 * WARNING: No check is made to ensure that the string referenced is a valid element of the structure.
 *          Ensure that the string exists in the structure type before attempting to read it
 */ 
int readSMBIOSStructureString( struct SMBIOSHandle *handle, u32 structureIndex, u32 stringIndex, u8 *buffer, u16 bufferSize ) {
    struct SMBIOSStructure tmpStructure;
    struct SMBIOSStructure *structure = &tmpStructure;
    char c1;
    int expended;
    int er, i;

    if ( stringIndex == 0 ) {
        buffer[ 0 ] = 0;
        return 0;
    }

    stringIndex--; 

    if ( ( er = seekSMBIOSStructure( handle, structureIndex ) ) != 0 ) { // update currentStructureOffset in handle 
        return er;
    }

    lseek( handle->file, handle->currentStructureOffset, SEEK_SET );
    if ( read( handle->file, &( structure->header ), sizeof( struct SMBIOSStructureHeader ) ) != sizeof( struct SMBIOSStructureHeader ) ||
         read( handle->file, structure->buffer, structure->header.length - sizeof( struct SMBIOSStructureHeader ) ) != ( structure->header.length - sizeof( struct SMBIOSStructureHeader ) ) ) 
         return 0;
    // file position now at the start of the strings

    while ( stringIndex ) {
        expended = 0;
        while ( !expended ) {
            if ( read( handle->file, &c1, 1 ) != 1 )
                return 0;
            if ( c1 == 0 ) {
                expended = 1;
            }
        }
        stringIndex--;
    }

    if ( read( handle->file, buffer, bufferSize ) != bufferSize ) 
        return 0;
    for ( i = 0; buffer[ i ] && i < bufferSize ; i++);

    if ( i < bufferSize ) {
        return 0;
    }

    return BUFFERTOOSMALL;
}

void displayHelp( char **argv ) {
    printf( "Syntax:\n" );
    printf( "%s < [--memory <file>] [--type <0|1|2>] < -p | -n | -a | -v > >\n", argv[ 0 ] );

    printf( "Options:\n" );
    printf( "%-25s%s\n", "-a or --all", "Print product name and version" );
    printf( "%-25s%s\n", "-p or --productversion", "Print 'Version' from structure" );
    printf( "%-25s%s\n", "-n or --productname", "Print 'Product Name'" );
    printf( "%-25s%s\n", "-m or --memory", "Specify a memory dev node other than /dev/mem" );
    printf( "%-25s%s\n", "-t or --type", "Specify structure type to use" );
    printf( "%-25s%s\n", "-v or --verbose", "Print formatted system information" );
    printf( "%-25s%s\n", "-V or --version", "Print program version" );
    printf( "%-25s%s\n", "--help", "Display this help text" );
    printf( "Examples:\n" );
    printf( "%s --help\n", argv[ 0 ] );
    printf( "%s --all\n", argv[ 0 ] );
    printf( "%s -p -n\n", argv[ 0 ] );
    printf( "%s -p -n --type 2\n", argv[ 0 ] );
    printf( "%s -productversion --memory /dev/mem2\n", argv[ 0 ] );
}

void displayProgramVersion() {
    printf( "%s\n",  versionString );
}

int generalError = 0;

void printVerboseString( char *title, struct SMBIOSHandle *handle, int structureIndex, int stringIndex, char *buffer, u16 size ) {
    
    if ( title ) printf( "%s: ", title );
    
    if ( stringIndex ) {
        readSMBIOSStructureString( handle, structureIndex, stringIndex, (u8*)buffer, size );
        if ( buffer[ 0 ] == 0 ) {
            serr( "Invalid SMBIOS string" );
            generalError = 1;
        } else {
            printf( "%s", buffer );
        }
    } else printf( "Not provided" );

    printf( "\n" );
}


int printProductName = 0, printProductVersion = 0, printAll = 0, 
    printHelp = 0, printProgramVersion = 0, printVerbose = 0, type = 1,
    autoDetectProductName=0;

int main( int argc, char **argv ) {

	struct SMBIOSHandle handle;
    struct SMBIOSStructure structure;
    char buffer[ 100 ];
    int structureIndex;
    int found = 0;
    int er, i;
    int optionIndex;

    programName = argv[ 0 ];
    if ( argc == 1 ) { // if no arguments are given, assume autodetection
        autoDetectProductName = 1;
        printProductName = 1;
    }

    memdev = memdevDefault;

    while ( 1 ) {
        static struct option long_options[] = {
            { "productversion", no_argument, 0, 'p' },
            { "productname", no_argument, 0, 'n' },
            { "memory", required_argument, 0, 'm' },
            { "all", no_argument, 0, 'a' },
            { "help", no_argument, &printHelp, 1 },
            { "version", no_argument, 0, 'V' },
            { "verbose", no_argument, 0, 'v' }, 
            { "type", required_argument, 0, 't' }
        };

        if ( ( found = getopt_long( argc, argv, "pnm:t:aVv", long_options, &optionIndex ) ) == -1 ) {
            break;
        }

        switch ( found ) {
        case 'v':  // print verbose output
            printVerbose = 1;
            break;
        case 'V':  // print program version
            printProgramVersion = 1;
            break;
        case 'p':  // print product version
            printProductVersion = 1;
            break;
        case 'n':  // print name
            printProductName = 1;
            break;
        case 'm':  // set memory dev node
            memdev = optarg;
            break;
        case 'a':  // print all system information
            printAll = 1;
            break;
        case 't':
            if ( (*optarg) != '1' && (*optarg) != '2' && (*optarg) != '0' ) {
                serr( "System information can only be pulled from type 1 or 2 structures" );
                exit( 1 );
            }
            type = (*optarg) - '0';
        }
    
    }

    if ( printProgramVersion ) {
        displayProgramVersion();
        return 0;
    }

    if ( printHelp ) {
        displayHelp( argv );
        return 0;
    }

    if ( printAll ) {
        printProductName = 1;
        printProductVersion = 1;
    }

    if ( ( type == 0 ) && ( printVerbose == 0 && printProductVersion == 0 ) ) {
        serr( "verbose or productversion must be used with type 0 information" );
        return 1;
    }

    if ( printVerbose ) {
        printAll = 0;
        printProductName = 0;
        printProductVersion = 0;
    }

    if ( ( printAll + printProductName + printProductVersion + printVerbose ) == 0 ) {
        serr( "-p -n -a or -v must be specified" );
        return 1;
    }

reopen:
    er = openSMBIOS( memdev, &handle );
    if ( er ) { 
        return 1;
    }
	
    found = 0;

    // Search for a desired structure type
    for ( structureIndex = 0; ( structureIndex < 0xffff) && (found == 0); structureIndex++ ) {
        readSMBIOSStructure( &handle, &structure, structureIndex ); // read next structure
        if ( structure.header.type == type ) { // If this is a system information structure
            found = 1;
            break;
        }
    }

    if ( found) {
        u8 productNameString = 0, versionString = 0;

        if ( type == 0 ) {
            struct SMBIOSStructure_BIOS *system;
        
            system = (struct SMBIOSStructure_BIOS *)&structure; // interpret the structure as a Type 0 entry
            if ( printVerbose ) {
                printf( "SMBIOS BIOS Information\n" );
                printf( "Structure Index: 0x%x\n", structureIndex );
                printf( "Header Information\n" );
                printf( " Type: %d\n",  (unsigned int)system->header.type );
                printf( " Header + Formatted Structure Length: %d bytes\n",  (unsigned int)system->header.length );
                printf( " Unique Handle: %d\n", system->header.handle );
                printf( "Formatted\n" );

                printf( " Vendor String Index: %d\n", (unsigned int)system->vendorString );
                printf( " Version String Index: %d\n", (unsigned int)system->versionString );
                printf( " BIOS Segement: 0x%lx\n", (unsigned long)system->biosSegment );
                printf( " Release Date String Index: %d\n", (unsigned int)system->releaseDateString );
                printf( " ROM Size: %d (%lu bytes)\n", (unsigned int)system->romSize, (unsigned long)( 64 * 1024 * ( system->romSize + 1 ) ) );
                printf( " BIOS Characteristics: 0x%Lx\n", system->biosCharacteristics );

                printf( "Structure Strings\n" );
                printVerboseString( " Vendor", &handle, structureIndex, system->vendorString, buffer, 100 );
                printVerboseString( " Version", &handle, structureIndex, system->versionString, buffer, 100 );
                printVerboseString( " Release Date", &handle, structureIndex, system->releaseDateString, buffer, 100 );
            }

            if ( printProductVersion ) {
                versionString = system->versionString;
            }
        }

        if ( type == 1 ) {
            struct SMBIOSStructure_System *system;
            
            system = (struct SMBIOSStructure_System *)&structure; // interpret the structure as a Type 1 entry
            if ( printVerbose ) {
                printf( "SMBIOS System Information\n" );
                printf( "Structure Index: 0x%x\n", structureIndex );
                printf( "Header Information\n" );
                printf( " Type: %d\n",  (unsigned int)system->header.type );
                printf( " Header + Formatted Structure Length: %d bytes\n",  (unsigned int)system->header.length );
                printf( " Unique Handle: %d\n", system->header.handle );
                printf( "Formatted\n" );
                printf( " Manufacturer String Index: %d\n", (unsigned int)system->manufacturerString );
                printf( " Product String Index: %d\n", (unsigned int)system->productNameString );
                printf( " Version String Index: %d\n", (unsigned int)system->versionString );
                printf( " Serial Number String Index: %d\n",  (unsigned int)system->serialNumberString );
                printf( " UUID: " );
                for ( i = 0; i < 16; i++ ) printf( "%x",  (unsigned int)system->uuid[ i ] );
                printf( "\n" );
                printf( " Wakeup Type: 0x%x\n",  (unsigned int)system->wakeupType );
                printf( "Structure Strings\n" );
                printVerboseString( " Manufacturer", &handle, structureIndex, system->manufacturerString, buffer, 100 );
                printVerboseString( " Product Name", &handle, structureIndex, system->productNameString, buffer, 100 );
                printVerboseString( " Version", &handle, structureIndex, system->versionString, buffer, 100 );
                printVerboseString( " Serial Number", &handle, structureIndex, system->serialNumberString, buffer, 100 );
            }
            productNameString = system->productNameString;
            versionString = system->versionString;

            if ( autoDetectProductName ) {
                readSMBIOSStructureString( &handle, structureIndex, productNameString, (u8*)buffer, 100 );
                if ( strlen( buffer ) != 7 ) { // if the entry does not appear to be a 7 digit IBM code
                    type=2;
                    closeSMBIOS( &handle );
                    goto reopen;
                }
            }
        }

        if ( type == 2 ) {
            struct SMBIOSStructure_BaseBoard *system;
            
            system = (struct SMBIOSStructure_BaseBoard *)&structure; // interpret the structure as a Type 2 entry
            if ( printVerbose ) {
                printf( "SMBIOS Base Board Information\n" );
                printf( "Structure Index: 0x%x\n", structureIndex );
                printf( "Header Information\n" );
                printf( " Type: %d\n",  (unsigned int)system->header.type );
                printf( " Header + Formatted Structure Length: %d bytes\n",  (unsigned int)system->header.length );
                printf( " Unique Handle: %d\n", system->header.handle );
                printf( "Formatted\n" );
                printf( " Manufacturer String Index: %d\n", (unsigned int)system->manufacturerString );
                printf( " Product String Index: %d\n", (unsigned int)system->productNameString );
                printf( " Version String Index: %d\n", (unsigned int)system->versionString );
                printf( " Serial Number String Index: %d\n",  (unsigned int)system->serialNumberString );
                printf( "Structure Strings\n" );
                printVerboseString( " Manufacturer", &handle, structureIndex, system->manufacturerString, buffer, 100 );
                printVerboseString( " Product Name", &handle, structureIndex, system->productNameString, buffer, 100 );
                printVerboseString( " Version", &handle, structureIndex, system->versionString, buffer, 100 );
                printVerboseString( " Serial Number", &handle, structureIndex, system->serialNumberString, buffer, 100 );
            }
            productNameString = system->productNameString;
            versionString = system->versionString;
            
            if ( autoDetectProductName ) {
                readSMBIOSStructureString( &handle, structureIndex, productNameString, (u8*)buffer, 100 );
                if ( strlen( buffer ) != 7 ) { // if the entry does not appear to be a 7 digit IBM code
                    productNameString = 0;
                    serr( "Unable to find IBM 7 digit product name in SMBIOS" );
                    generalError = 1;
                    printProductName = 0;
                }
            }
        }


        if ( printProductName ) {
            if ( productNameString == 0 ) { // if the product name is not provided in the SMBIOS structure
                serr( "Product Name not provided in SMBIOS" );
                generalError = 1;
            } else {
                readSMBIOSStructureString( &handle, structureIndex, productNameString, (u8*)buffer, 100 );
                if ( buffer[ 0 ] == 0 ) {
                    serr( "Invalid (empty) string in Product Name" );
                    generalError = 1;
                } else {
                    printf( "%s", buffer );
                }
            }
        }

        if ( printProductVersion ) {
            if ( versionString == 0 ) { // if the version is not provided in the SMBIOS structure
                serr( "Version not provided in SMBIOS" );
                generalError = 1;
            } else {
                readSMBIOSStructureString( &handle, structureIndex, versionString, (u8*)buffer, 100 );
                if ( buffer[ 0 ] == 0 ) {
                    generalError = 1;
                    serr( "Invalid (empty) string in Version" );
                } else {
                    if ( printProductName ) {
                        printf( " " );
                    }
                    printf( "%s", buffer );
                }
            }
        }
        
    }

    closeSMBIOS( &handle );
    printf( "\n" );
	return (!found) | generalError;
}
