#!/usr/bin/perl

#
#==========================================================================
# FILE        : posASWatch.pl
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : (c) 2012 Novell, Inc. All rights reserved.
#
# Author      : Ondrej Holecek <oholecek@suse.com>, 2012
#
# DESCRIPTION : Daemon for watching Admin server availability.
#               Triggers LDAP Sync-replication
#
#==========================================================================

use warnings;
use strict;
use Getopt::Long;

our $VERSION = '11.2';
our $APPNAME = 'posASWatch';

#SLEPOS modules
use POS::Syslog;
use POS::LdapControl;
use POS::LDAP;
use POS::Config;
use POS::Pid;
use POSIX qw/SIGINT SIGTERM SA_RESTART/;

log_parse_argv();

sub usage
{
    print STDERR "posASWatch v$VERSION
    
    Usage:  $APPNAME [OPTIONS]
            $APPNAME -h|--help\n"
} ## end sub usage

sub help
{
    usage print STDERR "
    Description:
        Checks for AdminServer availability and then starts/stops posleases2ldap
        and LDAP SyncRepl services.
      
    Options:
        -d, --nodaemon              do not daemonize script
        -o, --noloop                do just one iteration, no looping
        -V, --version               print the version of script being used
        -h, --help                  displays this help
        --usage                     display brief usage message\n";
    exit;
} ## end sub help

#global variables
my $nodaemon = 0;
my $noloop   = 0;
my $config;
my $ldapControl;

#parsing commandline params
my $optres =
  GetOptions('help|h' => \&help, 'nodaemon|d' => \$nodaemon, 'noloop|o' => \$noloop, 'version|V' => \&version);

unless ($optres) {
    usage();
    exit(1);
}

################################################################################
################################################################################

sub initVariables
{
    eval { $config = POS::Config->new(); };
    if ($@) {
        log_crit_print('SLEPOS configuration not found! Make sure to run initialization scripts first');
        exit(255);
    }
    $ldapControl = new POS::LdapControl('CONFIG' => $config);
} ## end sub initVariables

sub end
{
    #just set noloop to 1, this will 'ensure' (well except SIGKILL) we do not quit during some critical op.
    log_debug_print('ending signal catched');
    $noloop = 1;
} ## end sub end

my $sa = POSIX::SigAction->new(\&end);
# set TERM&INT handlers to restart functions after interruption
$sa->flags(SA_RESTART);
$sa->safe(1);
unless (POSIX::sigaction(SIGINT, $sa)) {
    log_crit_printf('Cannot set SIGINT handler! (%s)', $!);
    exit(-1);
}
unless (POSIX::sigaction(SIGTERM, $sa)) {
    log_crit_printf('Cannot set SIGTERM handler! (%s)', $!);
    exit(-1);
}

initVariables();

#Test if it is plain BS and is configured
if ($config->{'POS_ADMIN_DEPLOYED'}) {
    log_error_print('This is only plain BranchServer service. Stopping service!');
    exit(255);
}
if (not $config->{'POS_BRANCH_DEPLOYED'}) {
    log_error_print('BranchServer is not configured. Configure BS and try againg.');
    exit(255);
}

#check if we should fork
unless ($nodaemon) {
    use POSIX qw(setsid);
    log_debug_print('daemon mode enabled, daemonizing..');
    chdir '/';
    open STDIN,  '<',  '/dev/null';
    open STDOUT, '>>', "$config->{ASWATCH_LOGFILE}";
    open STDERR, '>>', "$config->{ASWATCH_LOGFILE}";
    defined(my $pid = fork()) or die 'Unable to fork: $!';
    exit(0) if ($pid);
    setsid() or die 'Unable to start new session: $!';
    umask 0;
    log_debug_printf('daemon spawned with pid %s', $$);
}
my $pidfile = POS::Pid->new('PIDFILE' => $config->{'ASWATCH_PID'}, 'TIMEOUT' => 1);
if ($pidfile->{'ERROR'}) {
    log_error_printf('PID file creation problem: %s', $pidfile->{'ERROR_MSG'});
    exit(255);
}


my $online = 0;
# last online timestamp files
my $lastOnlineOLD = "/etc/SLEPOS/lastOnline";
my $lastOnline    = $config->{'LASTONLINEFILE'};
#main loop
while (1) {
    #check AS is online
    if ($ldapControl->checkOnline()) {
        # check LDAP AS is reachable
        log_debug_print('AS switched ONLINE') unless $online;
        $online = 1;
    }
    else {
        log_debug_print('AS switched OFFLINE') if $online;
        $online = 0;
    }
    if ($online) {
        if ($ldapControl->getSyncState() == 1) {
            #syncrepl is not running, start replication
            $ldapControl->StartSync();
        }
    }
    else {
        if ($ldapControl->getSyncState() == 0) {
            # we lost the connection but sync is running, stop it
            $ldapControl->StopSync();
            # check if lastOnline file is present
            if (not -f "$lastOnline") {
                if (-f "$lastOnlineOLD") {
                    rename "$lastOnlineOLD", "$lastOnline";
                }
                else {
                    my $ldap = POS::LDAP->new('CONFIG' => $config);
                    my $res = $ldap->search(
                        scope  => 'base',
                        base   => $config->{'ALL_LDAPBASE'},
                        filter => '(objectClass=*)',
                        attrs  => ['contextCSN']
                    );
                    my $lastfileFH;
                    open $lastfileFH, '>', '$lastOnline';
                    print $lastfileFH $res->pop_entry()->get_value('contextCSN');
                    close $lastfileFH;
                }
            }
            if (not $ldapControl->setDBreadonly(0)) {
                log_warn_print('setting DB to read/write mode failed!');
            }
        }
        #always ensure provider is running
        $ldapControl->startProvider();
    }
    for (my $i = 0; $i < $config->{'ASWATCH_CHECKTIME'}; $i++) {
        #do not wait full ASWATCH_CHECKTIME if we are told to quit
        last if ($noloop);
        #wait for next check
        sleep 1;
    }
    #second last to exit while loop
    last if ($noloop);
}

if (defined $pidfile) {
    $pidfile->delete() if $$ == $pidfile->get_pid() && $$ == $pidfile->get_pidf();
    log_info_printf("posASWatch daemon ended");
}
