#!/bin/bash
#
#==========================================================================
# FILE        : posInitAdminserver.sh
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : (c) 2011 Novell inc, USA. All rights reserved
#
# Author      : Ondřej Holeček <oholecek@suse.cz>, 2011
#             : Pavel Němec <pnemec@suse.cz>, 2007
#             : Nadeem Ahmad Nazeer <nnazeer@novell.com>, 2005
#
# DESCRIPTION : The script is used to initialize the Admin server's rsync and openLDAP database.
#==========================================================================

appName=`basename ${0}`
appVersion="11.2.4"

#include defaults
. /usr/lib/SLEPOS/defaults

SLEPOSPATH="/etc/SLEPOS"
TEMPLATEPATH="${SLEPOSPATH}/template"
CONFTEMPLATE="${TEMPLATEPATH}/adminserver.conf.template"
CONFFILE="${SLEPOSPATH}/adminserver.conf"

#include posIniCommon.sh support
. /usr/lib/SLEPOS/posInitCommon.sh

LDAPPATH="/etc/openldap"
SYNC_CERTPATH=$POS_LOCAL_SYNC_DIRECTORY/certs

# when changed this POS::Config.pm has to be changed as well
OLDBRANCHCONF=$POSCONFPATH/branchserver.conf.old
LDAPDB=/var/lib/ldap

COMPANYNAME=""
COUNTRY=""
PASSWORD="a"
ENABLE_SSL="y"
ENABLE_SM=y

SSLKEYSIZE=1024
SECUREPORT=636
UNSECUREPORT=389
COMMUNICATION="SSL on port ${SECUREPORT}"

#force regenerating SSL certificates
REGENERATE="no"

# preserve old ldap database and migrate to new one. If this option is true
# it will change workflow
UPGRADE="no"

ERROR=0 #state of error
ERR_LDAPSTOP=1 #ldap fail to stop
ERR_LDAPSTART=2 #ldap fail start after configuration is written
ERR_LDAPUPDATE=3
ERR_LDAPLOGIN=4
ERR_SSL_CERTNOTCLEAN=6
ERR_SSL_CREATE=7
ERR_FILE_NOT_FOUND=8

# constants
readonly LDAP_NONE=0
readonly LDAP_FILES=1
readonly LDAP_LDAP=2

#########################
### Functions         ###
#########################

function usage(){
    cat << EOF_USAGE
${appName} v${appVersion}

Usage:  ${appName} [OPTIONS]
        ${appName} -h|--help
EOF_USAGE
}


function help(){
    usage
    cat << EOF_HELP

Description:
    Configure and initialize branch server services.

Options:
 --regenerate               force regenerate SSL certificates (old certificates
                            are deleted from its locations. This option have
                            sense only when using SSL. Default is not to regenerate.
 -r, --reinitialize=<FILE>   reconfigure admin server with provided configuration
                            file.
 -n, --noninteractive       perform unattended installation ( -r need to be specified )
 --ldap-backup              import LDAP database from provided file (XML or LDIF) instead of default
 -V, --version              print the version of script being used
 -h, --help                 displays this help
 --usage                    display brief usage message
EOF_HELP
}

function getLdapEntryAttribute()
{
    local LOGINDN="$2"
    local LOGINPW="$3"
    local BASEDN="$4"
    local ATTRIBUTE="$5"
    local FILTER="$6"

    if [ -z "$FILTER" ]; then
        FILTER='(objectClass=*)'
    fi

    local LDAPPARAM="-x -LLL -H ${BASEURI}://${LDAPHOST} -D $LOGINDN -w $LOGINPW -b $BASEDN -s sub $FILTER $ATTRIBUTE"
    local RES=`ldapsearch $LDAPPARAM | sed -ne "s@$ATTRIBUTE: \([^\$]*\)@\1@p"`
    eval "$1=\"$RES\""
}

#######################################################################################################################
## SSL helper functions ##
##########################
# copy YaST CAManagement default and point them to SLEPOS repository
# we don't want to use system repository as we delete whole CA on every reinitialization and don't want to mess with system
function initSSLrepository()
{
    # verify template exists
    local systemSSLtemplate="/var/lib/CAM/openssl.cnf.tmpl"
    if [ ! -f "$systemSSLtemplate" ]; then
        err "[ERROR] Cannot find system SSL defaults, check YaST CaManagement installation"
    fi
    sed -e "s@dir=/var/lib/CAM/@dir=$CAPATH@g" "$systemSSLtemplate" > "$CAPATH"/openssl.cnf.tmpl
}

# generating certificate which will serve for Certification Authority
function createCA()
{
    ## Generate a new CA private key if one does not already exist
    echo "Generating root SLEPOS CA..."
    perl -I "/usr/share/YaST2/modules/" << EOFPERL
use strict;
use YaPI::CaManagement;

my \$data = {
    'caName'   => 'SLEPOS_CA',
    'caPasswd' => "$PASSWORD",
    'type'     => 'plain',
    'repository'       => "$CAPATH"
};
my \$res = YaPI::CaManagement->ReadCA(\$data);
if (defined \$res) {
    print "[ERROR] SLEPOS CA already exists!\n";
    exit 1;
}
\$data = {
    'caName'           => 'SLEPOS_CA',
    'keyPasswd'        => "$PASSWORD",
    'commonName'       => "cn=${HOSTNAME},${LDAPBASE}",
    'keyLength'        => $SSLKEYSIZE,
    'days'             => '3650',
    'countryName'      => "`echo $COUNTRY | tr '[a-z]' '[A-Z]'`",
    'organizationName' => "$COMPANYNAME",
    'basicConstraints' => 'critical,CA:true',
    'nsCertType'       => 'objsign,email,server',
    'repository'       => "$CAPATH"
};

\$res = YaPI::CaManagement->AddRootCA(\$data);
if (not defined \$res) {
    print "[ERROR] Failed to create SLEPOS CA\n";
    exit 2;
}
\$data = {
    'caName'          => 'SLEPOS_CA',
    'caPasswd'        => "$PASSWORD",
    'exportFormat'    => 'PEM_CERT',
    'repository'      => "$CAPATH",
    'destinationFile' => "$CAPATH/ca.crt"
};
\$res = YaPI::CaManagement->ExportCA(\$data);
if (not defined \$res) {
    print "[ERROR] Failed to export SLEPOS CA\n";
    exit 3;
}
exit 0;
EOFPERL
    local result=$?
    if [ $result -gt 0 ]; then
        err "[ERROR] Failed to initialize SLEPOS CA"
    fi
    # link to YaST CAMgm dir so CA can be also managed from YaST
    ln -s "${CAPATH}/SLEPOS_CA" "/var/lib/CAM/SLEPOS_CA"
}

createServerCert()
{
    echo "Generating server certificate for secure communication..."
    perl -I "/usr/share/YaST2/modules/" << EOFPERL
use strict;
use YaPI::CaManagement;

my \$data = {
    'caName'           => 'SLEPOS_CA',
    'caPasswd'         => "$PASSWORD",
    'keyPasswd'        => "$PASSWORD",
    'certType'         => 'server',
    'commonName'       => "$FQDN",
    'keyLength'        => $SSLKEYSIZE,
    'days'             => '3649',
    'countryName'      => "`echo $COUNTRY | tr '[a-z]' '[A-Z]'`",
    'organizationName' => "$COMPANYNAME",
    'nsComment'        => 'SLEPOS Administration Server certificate',
    'basicConstraints' => 'critical,CA:false',
    'extendedKeyUsage' => 'serverAuth,clientAuth,msSGC,nsSGC',
    'nsCertType'       => 'server',
    'repository'       => "$CAPATH"
};
my \$res = YaPI::CaManagement->AddCertificate(\$data);
if ( not defined \$res ) {
    print "[ERROR] Failed to generate server certificate\n";
    exit 1;
}
\$data = {
    'caName'          => 'SLEPOS_CA',
    'caPasswd'        => "$PASSWORD",
    'keyPasswd'       => "$PASSWORD",
    'exportFormat'    => 'PEM_CERT',
    'certificate'     => "\$res",
    'destinationFile' => "${CERTPATH}/server.crt",
    'repository'      => "$CAPATH"
};
\$res = YaPI::CaManagement->ExportCertificate(\$data);
if ( not defined \$res ) {
    print "[ERROR] Failed to export server certificate\n";
    exit 2;
}
\$data->{'exportFormat'}    = 'PEM_CERT_KEY';
\$data->{'destinationFile'} = "${CERTPATH}/server.key";
\$res = YaPI::CaManagement->ExportCertificate(\$data);
if ( not defined \$res ) {
    print "[ERROR] Failed to export server certificate key\n";
    exit 3;
}
exit 0;
EOFPERL
    local result=$?
    if [ $result -gt 0 ]; then
        err "[ERROR] Error during issuing server certificate"
    fi
}

#
# this will check if SSL certificates are complete (all files)
# return
#    0  - if all files are on right place
#    1 - if SSL certificates are not correctly generated
function checkSSLComplete(){
    # check if files exists and verify server against CA
    if [ -d "${CAPATH}/SLEPOS_CA" ]; then
        perl -I "/usr/share/YaST2/modules/" << EOFPERL
use strict;
use YaPI::CaManagement;
# check existing certificate
my \$data = {
    'caName'           => 'SLEPOS_CA',
    'caPasswd'         => "$PASSWORD",
    'repository'       => "$CAPATH"
};
my \$res = YaPI::CaManagement->ReadCertificateList(\$data);
if ( not defined \$res ) {
    exit 1;
}
my \$cert;
foreach (@\$res) {
    if (\$_->{'commonName'} eq "$FQDN") {
        \$cert = \$_->{'certificate'};
        last;
    }
}
# verify it
\$data = {
    'caName'          => 'SLEPOS_CA',
    'caPasswd'        => "$PASSWORD",
    'certificate'     => \$cert,
    'repository'      => "$CAPATH"
};
#\$res = YaPI::CaManagement->Verify(\$data);
#if ( not defined \$res ) {
#    exit 2;
#}
# check if is correct one is exported
\$data = {
    'caName'          => 'SLEPOS_CA',
    'caPasswd'        => "$PASSWORD",
    'certificate'     => \$cert,
    'type'            => 'parsed',
    'repository'      => "$CAPATH"
};
\$res = YaPI::CaManagement->ReadCertificate(\$data);
if ( not defined \$res ) {
    exit 3;
}
my \$output = "";
my \$accepting = 0;
open(FILE, "${CERTPATH}/server.crt") or exit 4;
while(<FILE>){
    \$accepting = 1 if (/-----BEGIN CERTIFICATE-----/);
    \$output .= \$_ if (\$accepting);
    last if (/-----END CERTIFICATE-----/);
}
close(FILE);
chomp \$output;
if ( \$res->{'BODY'} ne \$output) {
    exit 5;
}
exit 0;
EOFPERL
        local result=$?
        if [ $result -eq 0 ]; then
            return 0
        fi
    fi
    return 1
}

#
# this will remove all SSL certificates (except SYNC_CERTPATH)
# return nothing
function clearSSLrepository(){
    echo "Clearing old SSL repository..."
    pushd "$CAPATH" > /dev/null
    if [ -d "SLEPOS_CA" ]; then
        rm -rf "SLEPOS_CA" &> /dev/null
    fi
    popd > /dev/null
    pushd "$CERTPATH" > /dev/null
    rm ca.crt server.crt &> /dev/null
    popd > /dev/null
}

#######################################################################################################################
## LDAP helper functions ##
###########################

# backup LDAP database and LDAP configuration
# return 0 if successful. name of file is set in LDAP_BACKUPFILE and CONF_BACKUPFILE
# TODO: rewrite this to be more fail proof, check for space, proper target directory

function backupLdapConfiguration()
{
    echo -n "Backuping LDAP configuration... "
    CONF_BACKUPFILE="${HOME}/SLEPOS_LDAPCFG.$(date +'%Y%m%d-%T').bak.tgz"
    tar -zcS -f "${CONF_BACKUPFILE}" -C "/" "${LDAPPATH:1}/"
    echo "LDAP Configuration backup file ${CONF_BACKUPFILE}"
}

function backupLdapDatabase(){
    echo -n "Backuping LDAP database (${LDAPDB})... "
    #this is not serious if file not exist
    if [ ! -f "${LDAPDB}/id2entry.bdb" ]; then
        echo "Database appear to be empty, skipping backup!"
        LDAP_BACKUPFILE=""
    else
        LDAP_BACKUPFILE="${HOME}/SLEPOS_LDAPDB.$(date +'%Y%m%d-%T').bak.tgz"
        tar -zcS -f "${LDAP_BACKUPFILE}" -C "/" "${LDAPDB:1}/"
        echo "LDAP Database backup file ${LDAP_BACKUPFILE}"
    fi
}

function cleanLdapConfiguration()
{
    echo "Cleaning LDAP configuration... "
    # remove old slapd.d content
    if [ -d "${LDAPPATH}/slapd.d" ]; then
        rm -rf "${LDAPPATH}/slapd.d/"*
    fi
    rm -f "${LDAPPATH}/slapd.conf"
}

# removes old LDAP database and LDAB configuration
function cleanLdapDatabase(){
    echo "Cleaning LDAP database (${LDAPDB})"
    if [ -n "$LDAPDB" -a "${LDAPDB}x" != "/x" -a \( -d "$LDAPDB" -o -L "$LDAPDB" \) ]; then
       #get temp file name to be sure not to overwrite some other DB_CONFIGs
       TMP=`mktemp`
       #backup DB_CONFIG
       mv "${LDAPDB}/DB_CONFIG" "$TMP"
       #delete DB
       rm -rf "${LDAPDB}/"*
       #restore DB_CONFIG
       mv "$TMP" "${LDAPDB}/DB_CONFIG"
    fi
}

# get LDAP DB suffix, SSL usage, cn=config password and LDAP address to use (default is at localhost)
function getLdapDetails(){
    local PASSWORD1

    # read company name
    askValue COMPANYNAME "Please enter your organization/company name [$COMPANYNAME]: "
    # read country
    while true; do
        askValue COUNTRY "Please enter the 2 letter abbreviation of your country [$COUNTRY]: "
        if [ ${#COUNTRY} == 2 ]; then
          break
        else
            echo "Country abbreviation has to be exactly two letters"
        fi
    done
    # ask for password
    while true; do
        local PASSWORD1
        askPassword PASSWORD "Please choose the LDAP administrator password"
        askPassword PASSWORD1 "Please enter the password again"
        if [ "$PASSWORD" != "$PASSWORD1" ];    then
            echo "The passwords you entered do not match, please try again!"
        elif [ -z "$PASSWORD" ]; then
            echo "Password can't be empty!"
        else
            PASSWORDLDAP=`slappasswd -s "${PASSWORD}"`;
            if [ $? -ne 0 ]; then
                 echo "[ERROR] Unable to encrypt chosen password"
            else
                break
            fi
        fi
    done

    # ask user for SSL
    if readYesNo "Enable secure connection (SSL) to LDAP service?" "$ENABLE_SSL"; then
        ENABLE_SSL="y"
        BASEURI="ldaps"
    else
        ENABLE_SSL="n"
        BASEURI="ldap"
    fi

    # ask for LDAP
    while true; do
        askPassword LDAP_CONFIG_PASSWORD \
            "Please enter LDAP configuration database password, keep empty to recreate LDAP database from scratch"
        # test config database and ldap availability
        if [ -z "$LDAP_CONFIG_PASSWORD" ]; then
            # no password given -> warn and recreate recreate
            if readYesNo "Recreate LDAP database? All data in LDAP database will be destroyed!"; then
                # leave LDAP_CONFIG_PASSWORD empty to recreate database
                break
            fi
        else
            if ! checkLdap; then
                echo "[ERROR] Unable to contact LDAP server!"
            else
                break
            fi
        fi
        if ! readYesNo "Retry configuration?" "y"; then
            err "Configuration aborted"
        fi
    done
}

# checkLdap assumes filled LDAPHOST (either remote or local) and filled LDAP_CONFIG_PASSWORD and checks if LDAP is reachable via cn=admin,cn=config
function checkLdap(){
    echo -n "Checking LDAP server for cn=config database..."
    if ! ldapwhoami -x -H "${BASEURI}://${LDAPHOST}" -D "cn=admin,cn=config" -w "$LDAP_CONFIG_PASSWORD"; then
        return 1;
    fi
}

# prepareLocalLdap creates new LDAP database on local machine and enable cn=config database
function createLocalLdap(){
    stopService ldap
    backupLdapConfiguration
    cleanLdapConfiguration
    backupLdapDatabase
    cleanLdapDatabase
    echo "Creating new LDAP configuration..."
    if [ "$ENABLE_SSL" == "y" ]; then
        processTemplate "${TEMPLATEPATH}/slapd.conf.template" "${SLEPOSPATH}/slapd.conf" \
                        @LDAPBASE@              "$LDAPBASE"                              \
                        @LDAPCONFIGPASSWORD@    "$LDAP_CONFIG_PASSWORD"                  \
                        "##TLS##"               ""
        updateSysconfigFile "OPENLDAP_START_LDAPS" "yes" /etc/sysconfig/openldap single
    else
        processTemplate "${TEMPLATEPATH}/slapd.conf.template" "${SLEPOSPATH}/slapd.conf" \
                        @LDAPBASE@              "$LDAPBASE"                              \
                        @LDAPCONFIGPASSWORD@    "$LDAP_CONFIG_PASSWORD"                  \
                        "##TLS##"               "#"
        updateSysconfigFile "OPENLDAP_START_LDAPS" "no" /etc/sysconfig/openldap single
    fi
    # transform static configuration to slapd.d, note slaptest essentially fail as actual DB does not exists at the moment
    slaptest -f "${SLEPOSPATH}/slapd.conf" -F "${LDAPPATH}/slapd.d" >/dev/null 2>&1
    # set config backend to ldap
    updateSysconfigFile "OPENLDAP_CONFIG_BACKEND" "ldap" /etc/sysconfig/openldap single
    #Always listen on standard interface, needed for both unencrypted and StartTLS
    updateSysconfigFile "OPENLDAP_START_LDAP" "yes" /etc/sysconfig/openldap single
    echo "Restarting LDAP service..."
    if ! startService ldap; then
        err "[CRIT] Initial LDAP DB start failed! See syslog for errors and/or consult SUSE support"
    fi
    chkconfig ldap on
}

# preparePosLdap assumes filled LDAPHOST and LDAP_CONFIG_PASSWORD and creates SLEPOS database in LDAP 
function preparePosLdap(){
    if ! checkLdap; then
        # ldap connection failed
         err "[ERROR] Unable to contact LDAP server!"
    fi
    echo "Creating SLEPOS database..."
    local LDAP_PARAMS="-x -H ${BASEURI}://${LDAPHOST} -D cn=admin,cn=config -w $LDAP_CONFIG_PASSWORD"
    #check if any database with LDAPBASE exists
    local DBDN=`ldapsearch $LDAP_PARAMS -b cn=config -LLL -s sub "(olcSuffix=${LDAPBASE})" dn | cut -d' ' -f2`
    if [ -n "$DBDN" ]; then
        # it does, get LDAP DB path
        LDAPDB=`ldapsearch $LDAP_PARAMS -b $DBDN -LLL -s base "(olcSuffix=${LDAPBASE})" olcDbDirectory \
            | sed -ne 's@olcDbDirectory: \([^$]*\)@\1@p'`
        # and remove it
        ldapdelete $LDAP_PARAMS -r $DBDN
        # shutdown ldap during data delete
        stopService ldap
        # backup
        backupLdapDatabase
        # delete database if exists
        cleanLdapDatabase
        # start ldap
        startService ldap
    fi
    #test if LDAP DB does contain any database and generate new DB dir if so
    while test -f "${LDAPDB}/id2entry.bdb"; do
        LDAPDB="${LDAPDB}-1"
    done
    #test if LDAPDB exists, if not create it. May happen when default is used so we are creating new for us
    # accept symlinks in case of HA
    if [ ! -d "$LDAPDB" -a ! -L "$LDAPDB" ]; then
        mkdir -p "$LDAPDB"
        chown ldap:ldap "$LDAPDB"
        chmod go-rwx "$LDAPDB"
    fi
    #test if DB_CONFIG exists, if not create default one
    if [ ! -f "${LDAPDB}/DB_CONFIG" ]; then
        cat > "${LDAPDB}/DB_CONFIG" <<EODBCONFIG
set_cachesize 0 15000000 1
set_lg_regionmax 262144
set_lg_bsize 2097152
set_lk_max_locks 30000
set_lk_max_objects 30000
set_flags DB_LOG_AUTOREMOVE
EODBCONFIG
    fi
    #test SSL usage for olcSecurity
    if [ "$ENABLE_SSL" == "y" ]; then
        SECURITY="olcSecurity: ssf=128"
    fi
    #now create SLEPOS database
    ldapadd $LDAP_PARAMS >/dev/null <<EOLDAPADD
dn: olcDatabase=hdb,cn=config
objectClass: olcDatabaseConfig
objectClass: olcHdbConfig
olcDatabase: hdb
olcDbDirectory: $LDAPDB
olcSuffix: $LDAPBASE
olcAccess: {0}to dn.subtree="cn=global,$LDAPBASE" by dn.children="$LDAPBASE" read by anonymous none
olcAccess: {1}to dn.subtree="cn=standards,$LDAPBASE" by dn.children="$LDAPBASE" read by anonymous none
olcAccess: {2}to dn.base="$LDAPBASE" by dn.children="$LDAPBASE" read by anonymous none
olcAccess: {3}to filter="(objectClass=organizationalUnit)" by set.exact="user/-* & this" read by anonymous none
olcRootDN: cn=admin,$LDAPBASE
olcRootPW: $PASSWORD
olcSyncUseSubentry: FALSE
olcMonitoring: TRUE
olcDbCacheSize: 10000
olcDbCheckpoint: 1024 5
olcDbIDLcacheSize: 30000
olcDbIndex: objectClass eq
olcDbIndex: modifyTimestamp eq
olcDbIndex: entryUUID eq
olcDbIndex: entryCSN eq
olcDbIndex: cn eq
olcDbIndex: scServiceName eq
olcDbIndex: o eq
olcDbIndex: ou eq
olcDbIndex: macAddress eq
olcDbIndex: scCashRegisterName eq
olcDbIndex: ipHostNumber eq
olcDbIndex: scImageName eq
olcDbIndex: scImageFile eq
olcDbIndex: scRoleName eq
$SECURITY
EOLDAPADD

    #get updated DBDN - openLDAP automatically prepend db number to avoid confusion
    DBDN=`ldapsearch $LDAP_PARAMS -b cn=config -LLL -s sub "(olcSuffix=${LDAPBASE})" dn | cut -d' ' -f2`
    if [ -z "$DBDN" ]; then
        err "Creating of SLEPOS ldap database failed! Check syslog for errors"
    fi
    ldapadd $LDAP_PARAMS >/dev/null <<EOLDAPADD
dn: olcOverlay={0}syncprov,$DBDN
objectClass: olcOverlayConfig
objectClass: olcSyncProvConfig
olcOverlay: {0}syncprov
olcSpCheckpoint: 100 5
olcSpSessionlog: 100
EOLDAPADD
}

#######################################################################################################################
## Main                     ##
##############################
#parsing commandline params
# We need TEMP as the `eval set --' would nuke the return value of getopt.
TEMP=`getopt -o r::nhV -l reinitialize::,regenerate,noninteractive,ldap-backup:,help,version,usage \
      -n "${appName}" -- "$@"`

if [ $? != 0 ] ; then usage 1>&2; exit 254; fi

# Note the quotes around `$TEMP': they are essential!
eval set -- "$TEMP"


while [ $# -gt  0 ]; do
    case $1 in
        --regenerate)
            REGENERATE="yes"
            ;;
        -r|--reinitialize)
            OLDCONFIGFILE=$2
            if [ -f "$OLDCONFIGFILE" ]; then
                shift
                UPGRADE="yes"
            elif [ -f "$CONFFILE" ]; then
                #source old config for defaults
                OLDCONFIGFILE="$CONFFILE"
                UPGRADE="yes"
            else
                err $ERR_GENERAL "[ERROR] Configuration file not found. Configuration halted."
            fi
            ;;
        -n| --noninteractive)
            export UNATTENDED="yes"
            ;;
        --ldap-backup)
            LDAP_FILE=$2
            if [ -f "$LDAP_FILE" ]; then
                shift
                USE_LDAP_BACKUP=1
            else
                err $ERR_FILE_NOT_FOUND "[ERROR] Import file not found. Configuration halted."
            fi
            ;;
        -h|--help)
            #just show help and exit
            help
            exit 0
            ;;
        -V|--version)
            echo "${appName} ${appVersion}"
            exit 0
            ;;
        --usage)
            usage
            exit 0
            ;;
    esac
    shift
done

echo "------------------------------------------------------------------------------"
echo "                   AdminServer LDAP Configuration                             "
echo "------------------------------------------------------------------------------"

HOSTNAME=`hostname`
# use domainname instead of 'hostname --domain' - bnc#596584
DOMAINNAME=`dnsdomainname 2>/dev/null`
if [ -z "$DOMAINNAME" ]; then
    FQDN="$HOSTNAME"
else
    FQDN="$HOSTNAME.$DOMAINNAME"
fi
# init external LDAP support
LDAPHOST="$FQDN"
# check if we are resolvable by FQDN?
if ! host "$FQDN" 2>&1 > /dev/null; then
    # no, so add ourselves into /etc/hosts
    # but check if $HOSTNAME isn't already there
    if [ "$HOSTNAME" == "$FQDN" ]; then
        echo -e "127.0.0.1\t$HOSTNAME" >> /etc/hosts
    else
        if [ -z "`grep $HOSTNAME /etc/hosts | tr -s [:space:]`" ]; then
            # note, just add the line
            echo -e "127.0.0.1\t$HOSTNAME\t$FQDN" >> /etc/hosts
        else
           # hostname is there, update the line, backup to be safe
           backupConf cp "/etc/hosts"
           sed -i 's@\(^[0-9.]*\)[ \t]*'"$HOSTNAME"'[ \t]*$@\1'"\t$HOSTNAME\t$FQDN"'@' "/etc/hosts"
        fi
    fi
fi

if [ "$UNATTENDED" == "yes" -a "$UPGRADE" != "yes" ]; then
    echo -e "[ERROR] Non-interactive installation mode selected but reinitialization (-r) not selected!.\nExiting..."
    exit -1
fi

if [ $UPGRADE = "yes" ]; then
    echo "Upgrading admin server installation"
    #source old config
    . "$OLDCONFIGFILE"
    #parse old config
    COMPANYNAME=`echo $ALL_LDAPBASE | sed -n 's@.*o=\([^,]*\).*@\1@p'`
    COUNTRY=`echo $ALL_LDAPBASE | sed -n 's@.*c=\([^$]*\).*@\1@p'`
    #try to decrypt old password
    decodePassword PASSWORD $POS_ADMIN_PASSWORD
    decodePassword LDAP_CONFIG_PASSWORD $LDAP_CONFIG_PASSWORD
    URI=`echo $ALL_LDAPHOST | sed 's@\(ldaps\{0,1\}\).*$@\1@'`
    if [ "${URI}x" == "ldapx" ]; then
            ENABLE_SSL="n"
    elif [ "${URI}x" == "ldapsx" ]; then
            ENABLE_SSL="y"
    else
        if [ -n "$ALL_LDAPSSL" -a "$ALL_LDAPSSL" -eq 1 ]; then
            ENABLE_SSL="y"
        else
            ENABLE_SSL="n"
        fi
    fi
    #backup old conf
    backupConf mv "$CONFFILE"
    # try to query SM address
    SMHOST=`ldapsearch -x -LLL -H "$ALL_LDAPHOST" -D "$POS_ADMIN_DN" -w "$PASSWORD" -b "cn=standards,$ALL_LDAPBASE" '(objectClass=scHardware)' 'scSUSEManager' 2>/dev/null | sed -n -e 's/scSUSEManager: \(.*\)$/\1/p'`
    if [ -n "$SMHOST" ]; then
        ENABLE_SM="y";
        decodePassword SMPASS $SMPASS
    else
        ENABLE_SM="n";
    fi
fi

if [ "$UNATTENDED" != "yes" ]; then
    getLdapDetails
    # ask about SUSE Manager
    if readYesNo "Enable SUSE Manager integration" "$ENABLE_SM"; then
        askValue SMHOST "Please enter SUSE Manager host address [$SMHOST]: " nocheck
        askValue SMUSER "Please enter SUSE Manager user name [$SMUSER]: "
        askPassword SMPASS "Please enter SUSE Manager password"
        ENABLE_SM="y";
    else
        ENABLE_SM="n";
        SMHOST="";
    fi
fi

if [ "$ENABLE_SSL" = "y" ]; then
    COMMUNICATION="SSL on port $SECUREPORT"
    port=$SECUREPORT
    BASEURI="ldaps"
else
    COMMUNICATION="TCP/IP on port $UNSECUREPORT"
    port=$UNSECUREDPORT
    BASEURI="ldap"
fi

LDAPBASE="o=${COMPANYNAME},$([ ${#COUNTRY} -gt 2 ] && echo 'd')c=${COUNTRY}"
echo "------------------------------------------------------------------------------"
echo "     LDAP root        :  $LDAPBASE"
echo "     SLEPOS account   :  cn=admin,$LDAPBASE"
echo "     Communication    :  ${COMMUNICATION}"
if [ "$ENABLE_SM" != "n" ]; then
    echo "     SUSE Manager     :  $SMHOST"
    echo "     SUSE Manager user:  $SMUSER"
fi
echo "------------------------------------------------------------------------------"
echo "Make sure port ${port} is open in firewall. "
echo "Firewall can be set up using command: yast2 firewall services add tcpport=$port zone=EXT"
if [ "$UNATTENDED" != "yes" ]; then
    if readYesNo "Continue with configuration?" "y"; then
        echo "Performing configuration..."
    else
        err $ERR_GENERAL "[ERROR] configuration aborted"
    fi
fi

# remove any certificates in the SYNC_CERPATH directory
# do it everytime so Branch Servers are not confused if they found certificates
rm -f "${SYNC_CERTPATH}"/*

if [ $ENABLE_SSL = "y" ] ; then
    if [ -z "$LDAP_CONFIG_PASSWORD" -o "$LDAP_CONFIG_PASSWORD" == "$PASSWORD" ]; then
        ## use SLEPOS CA
        #check if ssl is ok (complete)
        if [ $REGENERATE = "yes" ] || ! checkSSLComplete; then
            initSSLrepository
            # we will regenearte certificates
            clearSSLrepository
            createCA
            createServerCert
            if ! checkSSLComplete; then
                err $ERR_SSL_CREATE "[ERROR] generating SSL keys failed"
            fi
        fi
        CAFILE="${CAPATH}/ca.crt"
    else
        getLdapEntryAttribute CAFILE "cn=admin,cn=config" "$LDAP_CONFIG_PASSWORD" "cn=config" "olcTLSCACertificateFile"
    fi
    # link the public key to the rsync/cert directory
    ln -s "$CAFILE" "${SYNC_CERTPATH}/ca.crt"
    # link public key to keys/cert directory
    [ -e "${CERTPATH}/ca.crt" ] && rm -f "${CERTPATH}/ca.crt"
    ln -s "$CAFILE" "${CERTPATH}/ca.crt"
    # update using of SLEPOS CA for LDAP
    cat > "${LDAPPATH}/ldap.conf" <<EOT
TLS_CACERT $CAFILE
TLS_REQCERT demand
EOT
    if [ "$LDAP_CONFIG_PASSWORD" == "$PASSWORD" ]; then
        # restart LDAP to load new certificate as we will not be recreating LDAP config
        startService ldap
    fi
else
    # delete ldap.conf file if it exists
    [ -f "${LDAPPATH}/ldap.conf" ] && rm "${LDAPPATH}/ldap.conf"
fi

echo "Modifying the LDAP service configuration..."
if [ "$LDAPHOST" == "$FQDN" -a -z "$LDAP_CONFIG_PASSWORD" ]; then
    LDAP_CONFIG_PASSWORD="$PASSWORD"
    createLocalLdap
fi
preparePosLdap

# encrypt LDAP DB password
POSADMINDN="cn=admin,${LDAPBASE}"
encryptPassword       # takes PASSWORD, generated ENCRYPTEDPASSWORD
POSADMINPW="$ENCRYPTEDPASSWORD"
# encrypt LDAP Config password
PASSWORD="$LDAP_CONFIG_PASSWORD"
encryptPassword "no-key-reset"     # takes PASSWORD, generated ENCRYPTEDPASSWORD
LDAP_CONFIG_PASSWORD="$ENCRYPTEDPASSWORD"
# encrypt SM password
if [ "$ENABLE_SM" = "y" ]; then
    PASSWORD="$SMPASS"
    encryptPassword "no-key-reset" #use the same salt as for LDAP password
    SMPASS="$ENCRYPTEDPASSWORD"
fi

processTemplate "$CONFTEMPLATE"      "$CONFFILE"                 \
                @ADMINSERVERNAME@    "${BASEURI}://${LDAPHOST}"  \
                @POSADMINDN@         "$POSADMINDN"               \
                @POSADMINPASSWORD@   "$POSADMINPW"               \
                @LDAPCONFIGPASSWORD@ "$LDAP_CONFIG_PASSWORD"     \
                @LDAPBASETEMPLATE@   "$LDAPBASE"

if [ "$ENABLE_SM" = "y" ]; then
    processTemplate -i "$CONFFILE"          \
                        @SMUSER@  "$SMUSER" \
                        @SMPASS@  "$SMPASS"
else
    processTemplate -i "$CONFFILE"   \
                        @SMUSER@  "" \
                        @SMPASS@  ""
fi

#import base ldap structure
echo "Importing LDAP directory information..."
if [ -z "$LDAP_FILE" ]; then
    LDAP_FILE="$(mktemp).xml"
    processTemplate "${TEMPLATEPATH}/XML/basic.xml" "$LDAP_FILE"    \
                    @LDAPBASE@              "$LDAPBASE"             \
                    @COMPANYNAME@           "$COMPANYNAME"          \
                    @TFTP_DEFAULT_INITRD@   "$TFTP_DEFAULT_INITRD"  \
                    @TFTP_DEFAULT_KERNEL@   "$TFTP_DEFAULT_KERNEL"  \
                    @POS_DHCP_BOOT_REMOTE@  "$POS_DHCP_BOOT_REMOTE" \
                    @POS_DHCP_BOOT_LOCAL@   "$POS_DHCP_BOOT_LOCAL"  \
                    @SM_HOST@               "$SMHOST"

fi
if ! posAdmin --import --file "$LDAP_FILE"; then
    err $ERR_LDAPUPDATE "[ERROR] LDAP import failed, check ${LDAP_FILE} and try configuration again."
fi
if [ -z "$USE_LDAP_BACKUP" -o "$USE_LDAP_BACKUP" != "1" ]; then
    rm -f "$LDAP_FILE"
fi

# switch rsync on
echo "Enabling and starting rsync service..."
backupConf mv "/etc/rsyncd.conf"
ln -sf /etc/SLEPOS/rsync/rsyncd.conf /etc/rsyncd.conf
chkconfig rsyncd on
startService rsyncd

warnNotReplaced "$CONFFILE" 

echo "------------------------------------------------------------------------------------"
echo "                      AdminServer configuration successful."
echo "------------------------------------------------------------------------------------"
exit 0
