#!/usr/bin/perl -w
use strict;
use Net::LDAP::LDIF;
use Net::LDAP::Util qw(canonical_dn);
use Getopt::Long;
use Date::Parse;
use Data::Dumper;

my %workstations;
my %locations;
my %locations_ws;
my $min_timestamp = 0;

sub last_image
{
	my ($entry) = @_;
	my @notifiedImage = $entry->get_value('scNotifiedImage');
	my $last_ts = $entry->get_value('scLastBootTime');
	my $last_image;
	
	$last_ts = 0 unless defined $last_ts;
		
	foreach my $rec (@notifiedImage) {
		my ($image, $ts, $hash) = split(/;/, $rec);
		$last_image = $image unless defined $last_image; 
		
		if (defined $ts && $ts > $last_ts) {
			$last_ts = $ts;
			$last_image = $image;
		}
	}
	$last_image = '' unless defined $last_image;
	$last_image =~ s/.*\///;
	return $last_ts, $last_image;
}

sub last_timestamp
{
	my ($entry) = @_;
	my ($last_ts, $last_image) = last_image($entry);
	return $last_ts;
}

sub read_ldif
{
	my ($file) = @_;
	
	my $ldif;
	
	if ($file) {
		$ldif = Net::LDAP::LDIF->new($file, "r", onerror => 'undef' );
	} else {
		$ldif = Net::LDAP::LDIF->new(\*STDIN, "r", onerror => 'undef' );
	}

	while( not $ldif->eof ( ) ) {
		my $entry = $ldif->read_entry ( );
		if ( $ldif->error ( ) ) {
			print "Error msg: ", $ldif->error ( ), "\n";
			print "Error lines:\n", $ldif->error_lines ( ), "\n";
			exit(1);
		}
		
		my $dn = $entry->dn();
		my @object_class = $entry->get_value('objectClass');
		
		if ($object_class[0] =~ /scWorkstation/) {
			if ($min_timestamp == 0 || $min_timestamp < last_timestamp($entry)) {
				$workstations{$dn} = $entry;
			}

		}
		elsif ($object_class[0] =~ /scLocation/) {
			$locations{$dn} = $entry;
			$locations_ws{$dn} = [];
		}
	}
	$ldif->done ( );
}

sub sort_ws
{
	foreach my $dn (keys %workstations) {
		my $entry = $workstations{$dn};
		
		while ($dn && !defined $locations{$dn}) {
			$dn =~ s/^[^,]*//;
			$dn =~ s/^,//;
		}
		
		if ($dn) {
			$locations_ws{$dn} = [] unless defined $locations_ws{$dn};
			push @{$locations_ws{$dn}}, $entry;
		}
	}
}


sub report_workstations
{
	foreach my $branch_dn (sort {canonical_dn($a, reverse => 1) cmp canonical_dn($b, reverse => 1)} keys %locations) {
		print "\n";
		print "===============================================================================\n";
		print "\n";
		print "Branch / Location: $branch_dn\n";
		print "\n";
		printf "%-10s %-16s %-25s %s\n", "cn", "IP", "last boot time", "image";
		print "\n";
		
		foreach my $entry (sort {$a->get_value('cn') cmp $b->get_value('cn')} @{$locations_ws{$branch_dn}}) {
#			print Dumper($dn);
			my @mac = $entry->get_value('macAddress');
			my $ip  = $entry->get_value('ipHostNumber');
		
			my ($last_ts, $last_image) = last_image($entry);
			my $last_boot = localtime($last_ts);
			$last_boot = '-' if $last_ts == 0;
			#join(' ', @mac)
			printf "%-10s %-16s %-25s %s\n", $entry->get_value('cn'), $ip, $last_boot, $last_image;
		}
	}
}

sub report_locations
{
	printf "%-60s %s\n", "Branch / Location DN", "number of ws";
	foreach my $dn (sort {canonical_dn($a, reverse => 1) cmp canonical_dn($b, reverse => 1)} keys %locations) {
		my $num_ws = @{$locations_ws{$dn}};
		printf "%-60s %s\n", $dn, $num_ws;
	}
}

sub report_stats
{
	my $num_ws = keys %workstations;
	my $num_loc = keys %locations;

	print "\n";
	print "Number of workstations: $num_ws\n";
	print "Number of branches:     $num_loc\n";
}

###################################################################################

my @tmp     = split('/', $0);
my $appname = pop @tmp;

my $report_ws = 0;
my $report_loc = 0;
my $report_stats = 0;
my $date_str = undef;

my $optres  = GetOptions(       # returns true on sucess
    'help|h' => \&help,
    'workstations|w' => \$report_ws,
    'locations|l'  => \$report_loc,
    'statistics|s'  => \$report_stats,
    'since=s' => \$date_str
);  

sub help
{
    print STDERR "Description:\n";
    print STDERR "  SLEPOS LDIF statistics.\n";
    print STDERR "\n";
    usage();
    exit(0);
} ## end sub help

sub usage
{
    print STDERR "Usage:\n";
    print STDERR "  $appname [-w|--workstations] [-l|--locations] [-s|--statistics] [--since <date>] [<file>]\n";
    print STDERR "  $appname -h|--help\n";
    print STDERR "\n";
    print STDERR "Options:\n";
    print STDERR "  -h, --help                  displays this help   \n";
    print STDERR "  -w, --workstations          workstations report   \n";
    print STDERR "  -l, --locations             branches / locations report   \n";
    print STDERR "  -s, --statistics            overall statistics report   \n";
    print STDERR "  --since <date>              consider only workstations booted since <date>\n";
    
    print STDERR "  <file>                      ldif file, use stdin if missing\n";
    print STDERR "\n";
} ## end sub usage
unless ($optres) {
    print STDERR "Error parsing parameters\n\n";
    usage();
    exit(1);
}

$report_stats = 1 unless ($report_ws || $report_loc);

$min_timestamp = str2time($date_str) if defined $date_str;

read_ldif($ARGV[0]);
sort_ws();

report_workstations() if $report_ws;
report_locations() if $report_loc;
report_stats() if $report_stats;

exit 0;

