#!/bin/bash 


. /usr/share/SLEPOS/KIWILinuxRC-lite.sh
runHook init

IFS_ORIG=$IFS

CLEANUP=
SERVER=tftp
SERVERTYPE=tftp
SOURCECONFIG=
DESTCONFIG=config.default
DESTDIR=
FILES=

usage () {
    cat << EOF_USAGE

Creates a directory structure for off-line boot on SLEPOS service partition.

usage:

posSyncSrvPart [--server <host|path> ] [--server-type <tftp|ftp|local>] 
               [--source-config <config>] [--dest-config <config>]
               [--cleanup] [--dest-dir <dir>]

--source-config      name of config file on branch server
                     default: config.<mac> with fallback to config.default

--dest-config        name of config file on service partition
                     default: config.default

--server             branch server hostname
                     default: tftp

--server-type        tftp|ftp|local
                     default: tftp

--dest-dir           path to service partition
                     default: /srv/SLEPOS if it is mounted

--cleanup            delete all files on service partition that are not referenced
                     from config file

EOF_USAGE
}

TEMP=`getopt -o isgq -l cleanup,server:,server-type:,source-config:,dest-config:,dest-dir: \
    -n "$0" -- "$@"`

if [ $? != 0 ] ; then usage 1>&2; exit 1; fi

# Note the quotes around `$TEMP': they are essential! 
eval set -- "$TEMP"

while true; do
    case $1 in
      --cleanup)
          CLEANUP=y
          shift
          ;;
      --server)
          SERVER=$2
          shift 2
          ;;

      --server-type)
          SERVERTYPE=$2
          shift 2
          ;;

      --source-config)
          SOURCECONFIG=$2
          shift 2
          ;;

      --dest-config)
          DESTCONFIG=$2
          shift 2
          ;;

      --dest-dir)
          DESTDIR=$2
          shift 2
          ;;

    

      --) shift ; break ;;

      *)
          echo "Internal error!"
          exit 254
          ;;
    esac
done

if [ $# -gt 0 ]; then
    echo "Odd addition arguments '$*'\n".
fi

if [ -z "$DESTDIR" ] && mount |grep " /srv/SLEPOS " ; then
	DESTDIR=/srv/SLEPOS
fi

if [ -z "$DESTDIR" ]; then
	echo "Dest dir not specified. Could not fall back to service partition in /srv/SLEPOS." 
	exit 1
elif [ ! -d "$DESTDIR" ]; then
	echo "Dest dir '$DESTDIR' does not exist"
	exit 1
fi

mkdir -p "$DESTDIR/KIWI"
mkdir -p "$DESTDIR/image"
imageBlkSize=8192

if [ -n "$SOURCECONFIG" ] ; then
	rm -f "$DESTDIR/KIWI/$DESTCONFIG"
	if ! fetchFile "KIWI/$SOURCECONFIG" "$DESTDIR/KIWI/$DESTCONFIG.tmp" ; then
		echo "Can't download '$SOURCECONFIG'"
		exit 1
	fi
else
	for suffix in `ifconfig |grep HWaddr| sed -e "s/.*HWaddr //"` default ; do
		rm -f "$DESTDIR/KIWI/$DESTCONFIG.tmp"
		if fetchFile "KIWI/config.$suffix" "$DESTDIR/KIWI/$DESTCONFIG.tmp" ; then
			break
		fi
	done
	if [ ! -s "$DESTDIR/KIWI/$DESTCONFIG.tmp" ] ; then
		echo "Can't download config file"
		exit 1
	fi
fi

mv -f "$DESTDIR/KIWI/$DESTCONFIG.tmp" "$DESTDIR/KIWI/$DESTCONFIG"
FILES="$DESTDIR/KIWI/$DESTCONFIG"
importFile <"$DESTDIR/KIWI/$DESTCONFIG"

count=0
IFS="," ; for i in $IMAGE;do
	count=$(($count + 1))
	field=0
	IFS=";" ; for n in $i;do
	case $field in
		0) imageDevice=$n ; field=1 ;;
		1) imageName=$n   ; field=2 ;;
		2) imageVersion=$n; field=3 ;;
		3) imageServer=$n ; field=4 ;;
		4) imageBlkSize=$n; field=5 ;;
		5) imageZipped=$n ;
	esac
	done
	imagePath="image/$imageName-$imageVersion"

	IFS=$IFS_ORIG
	
	md5path=$imagePath.md5
	fetchFile $md5path "$DESTDIR/$md5path.tmp" uncomp
	read sum1 blocks blocksize zblocks zblocksize <"$DESTDIR/$md5path.tmp"
	if [ "$imageZipped" = "compressed" ] ; then
	        imagePath=$imagePath.gz
	fi
	if [ -f "$DESTDIR/$imagePath" ]; then
		if [ "$imageZipped" = "compressed" ] ; then
			md5=`gzip -d -c "$DESTDIR/$imagePath" |md5sum | cut -f 1 -d ' '`
		else
			md5=`md5sum "$DESTDIR/$imagePath" | cut -f 1 -d ' '`
		fi
	fi
	if [ "x$sum1" != "x$md5" ]; then
		Echo "Fetching $imagePath"
		if ! fetchFile $imagePath "$DESTDIR/$imagePath" uncomp ; then
			Echo "Downloading $imagePath failed"
			exit 1
		fi
	fi
	mv -f "$DESTDIR/$md5path.tmp" "$DESTDIR/$md5path"
	FILES="$FILES
$DESTDIR/$md5path
$DESTDIR/$imagePath"
done

installServiceKernel "$DESTDIR"
FILES="$FILES
$DESTDIR/boot/$POS_KERNEL
$DESTDIR/boot/$POS_INITRD
$DESTDIR/boot/$POS_KERNEL.md5
$DESTDIR/boot/${POS_INITRD%.gz}.md5"

IFS="," ; for i in $CONF;do
	unset configHash
	field=0
	IFS=";"; for n in $i;do
	case $field in
		0) configSource=$n ; field=1 ;;
		1) configDest=$n   ; field=2 ;;
		2) configServer=$n ; field=3 ;;
		3) configBlkSize=$n; field=4 ;;
		4) configHash=$n   ; field=5
	esac
	done
	mkdir -p "`dirname \"$DESTDIR/$configSource\"`"
	if [ ! -f "$DESTDIR/$configSource" -o \
	     "x$configHash" != x`md5sum "$DESTDIR/$configSource" 2>/dev/null| cut -f 1 -d ' '` ]; then
	        Echo "Fetching $configSource"
		fetchFile $configSource "$DESTDIR/$configSource"
	fi
	FILES="$FILES
$DESTDIR$configSource"
done

#files used by grub:
FILES="$FILES
$DESTDIR/boot/grub
$DESTDIR/boot/linux.vmx
$DESTDIR/boot/initrd.vmx"

if [ -n "$CLEANUP" ]; then
	find "$DESTDIR/boot" "$DESTDIR/image" "$DESTDIR/KIWI" -type f -print | grep -v -F "$FILES" | \
	        while read f; do rm -f "$f"; done
fi

exit 0
