#!/usr/bin/perl -w

#
# Checks if two syncrepl instances (provider and consumer) are in sync
# by comparing the contextcsn Attribute.
# (Currently this tool assumes a single master setup. In multimaster
# setups contextcsn in mulitvalued and must be compared based on the
# server ID)
#

use Getopt::Long;
use Net::LDAP;
use POS::Config;
use POS::Syslog;

use strict;

log_parse_argv();

my $config    = POS::Config->new();
my $binddn    = $config->{'POS_ADMIN_DN'};
my $bindpw    = $config->{'POS_ADMIN_PASSWORD'};
my $provCsn   = "";
my $consCsn   = "";
my $provider  = "";
my $consumer  = "";
my $contextDn = "";
my $wait      = 0;
# LDAP connections
my $provLdap;
my $consLdap;

if (!GetOptions("provider=s" => \$provider, "consumer=s" => \$consumer, "base=s" => \$contextDn, "wait=i" => \$wait)) {
    print_usage();
    exit -1;
}

if (!$provider) {
    print STDERR "Missing required Argument --provider\n";
    print_usage();
    exit -1;
}
if (!$consumer) {
    print STDERR "Missing required Argument --consumer\n";
    print_usage();
    exit -1;
}
if (!$contextDn) {
    print STDERR "Missing required Argument --base\n";
    print_usage();
    exit -1;
}

# Provider, is assumed stable. Only one connection attemp
$provLdap = Net::LDAP->new($provider);
if (!$provLdap) {
    die("LDAP connection to provider failed");
}
my $msg = $provLdap->bind($binddn, password => $bindpw);
if ($msg->code()) {
    print STDERR "Error authenticating to provider LDAP server:\n";
    print STDERR "    " . $msg->error() . "\n";
    exit -1;
}
# Consumer, is assumed unstable, during resync connection can be lost
connectConsumer();
#autoflush
$| = 1;
my $interval = 3;
print "Waiting for Syncreplication to complete ";
if ($wait > 0) {
    print "(Maximum wait time $wait seconds. Checking every $interval seconds)";
}
print "\n";
my $insync = 0;
for (my $i = 0; $i <= $wait; $i += $interval) {
    $provCsn = readContextCsn($provLdap, $contextDn);
    $consCsn = readContextCsn($consLdap, $contextDn);

    if (!defined $provCsn) {
        print STDERR "\nCould not read contextCsn from provider";
        last;
    }
    if (!defined $consCsn) {
        # "contextCsn" not yet populated?
    }
    elsif ($consCsn eq $provCsn) {
        $insync = 1;
        last;
    }
    print ".";

    sleep($interval);
}

print "\n";

if ($insync) {
    print "Server are in sync now.\n";
    exit 0;
}
else {
    if ($wait > 0) {
        print "Maximum waittime exceeded. ";
    }
    print "Servers are in not sync.\n";
    exit 1;
}

sub connectConsumer
{
    # Consumer, connection may not be successful 100% time
    $consLdap = Net::LDAP->new($consumer);
    if (!$consLdap) {
        print STDERR "LDAP connection to consumer failed. May happen during resyncing. Skipping cycle\n";
        undef $consLdap;
        return;
    }
    # authentication is critical.
    my $msg = $consLdap->bind("cn=admin," . $binddn, password => $bindpw);
    if ($msg->code()) {
        print STDERR "Error authenticating to consumer LDAP server:\n";
        print STDERR "    " . $msg->error() . "\n";
        exit -1;
    }
} ## end sub connectConsumer

sub readContextCsn
{
    my ($ldap, $base) = @_;
    if (not defined $ldap) {
        connectConsumer();
        return;
    }
    my $msg = $ldap->search(base => $base, scope => "base", filter => "(objectclass=*)", attrs => ["contextcsn"]);
    if (!$msg->code()) {
        my $entry  = $msg->entry(0);
        my @ctxCsn = $entry->get_value("contextcsn");
        if (scalar(@ctxCsn) == 1) {
            return $ctxCsn[0];
        }
        elsif (scalar(@ctxCsn) > 1) {
            print STDERR "multimaster setup not supported\n";
            return;
        }
        else {
            return;
        }

    }
    else {
        # trigger reconnect, LDAP could be restarted. Should happen only to consumer, provider is considered stable
        connectConsumer();
    }
    return;
} ## end sub readContextCsn

sub print_usage
{
    print STDERR "usage:\n";
    print STDERR "  --provider=<ldap-uri of the provider>\n";
    print STDERR "  --consumer=<ldap-uri of the consumer>\n";
    print STDERR "  --base=<context/base DN>\n";
    print STDERR "  --wait=<number of seconds to wait for provider and consumer coming to sync>\n";
    log_print_usage();
} ## end sub print_usage

